# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::CatGen.pm
#  Created	: 21 September 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::CatGen;
require 5.005_62;
use strict;
use warnings;
use stl;
use ETL::Pequel3::Type::Document;
# ----------------------------------------------------------------------------------------------------
# DOCUMENT TYPE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::MainProgram; 
	use base qw(ETL::Pequel3::Type::Document::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'doc_type', default => 'Pequel'),
		Class::STL::ClassMembers::DataMember->new(name => 'doc_title', default => 'Pequel Type Catalogue'),
		Class::STL::ClassMembers::DataMember->new(name => 'doc_version', default => '3'),
		Class::STL::ClassMembers::DataMember->new(name => 'doc_email', default => 'http://sourceforge.net/projects/pequel'),
		Class::STL::ClassMembers::DataMember->new(name => 'doc_author', default => ''),
		Class::STL::ClassMembers::DataMember->new(name => 'doc_filename', default => 'pequel_catalogue.pod'),
		Class::STL::ClassMembers::DataMember->new(name => 'show_xsd', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'document_name', default => 'pequel_catalogue'),
		Class::STL::ClassMembers::DataMember->new(name => 'show_copyright', default => 1);
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->chapters()->add(
			$self->catalogue()->cat_chapters()->section->new(),
			$self->catalogue()->cat_chapters()->aggregate->new(),
		);
		foreach ($self->catalogue()->macro_categories()->to_array()) 
		{
			$self->chapters()->add(
				$self->catalogue()->cat_chapters()->macro->new(
					macro_category => $_->name(),
					chapter_name => "@{[ $_->name() ]}_macros")
			);
		}
		$self->chapters()->add(
			$self->catalogue()->cat_chapters()->pequel_type->new(),
			$self->catalogue()->cat_chapters()->date_type->new(),
			$self->catalogue()->cat_chapters()->table->new(),
			$self->catalogue()->cat_chapters()->input_dataset->new(),
			$self->catalogue()->cat_chapters()->output_dataset->new(),
			$self->catalogue()->cat_chapters()->input_datasource->new(),
			$self->catalogue()->cat_chapters()->output_datasource->new(),
			$self->catalogue()->cat_chapters()->configuration_property->new(),
			$self->catalogue()->cat_chapters()->command_line_option->new(),
#?			$self->catalogue()->cat_chapters()->document->new(),
#?			$self->catalogue()->cat_chapters()->document_chapter->new(),
#?			$self->catalogue()->cat_chapters()->generator->new(),
#?			$self->catalogue()->cat_chapters()->generator_segment->new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
# DOCUMENT TYPE CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Catalogue;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'cat_documents'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'document_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::CatGen::Abstract');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::CatGen::MainProgram->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
# CHAPTER TYPE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Section; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'section');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach (map($_->new(), $self->catalogue()->sections()->to_array())) 
		{
			$self->user_prog()->code("@{[ $_->section_name() ]} -- I<@{[ $_->construction_type() ]}>");
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
		$self->user_prog()->code();
		foreach (map($_->new(), $self->catalogue()->sections()->to_array())) 
		{
			$self->user_prog()->head2($_->section_name());
			if (defined($_->description()) && $_->description() ne '') {
				$self->user_prog()->code($_->description());
				$self->user_prog()->code();
				$self->user_prog()->code();
			}

			# Section Construction Type
			$self->user_prog()->code("F<Section Construction Type>");
			$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
			$self->user_prog()->over();
			$self->user_prog()->code($_->construction_type());
			$self->user_prog()->code();
			$self->user_prog()->back();

			# Section Properties:
			$self->user_prog()->code("F<Section Properties>");
			$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
			$self->user_prog()->over();
			if ($_->attributes()->size() == 0) {
				$self->user_prog()->code("I<N/A>");
				$self->user_prog()->code();
			}
			foreach my $attr ($_->attributes()->to_array()) {
				$self->user_prog()->code("B<@{[ $attr->name() ]}>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->over();
				if (defined($attr->description()) && $attr->description() ne '') {
					$self->user_prog()->code("I<@{[ $attr->description() ]}>");
					$self->user_prog()->code();
				}
				$self->user_prog()->code($attr->required() ? "-- required" : "-- optional");
				$self->user_prog()->code();
				$self->user_prog()->code("-- default: B<@{[ $attr->default() ]}>") if (defined($attr->default() ));
				$self->user_prog()->code() if (defined($attr->default() ));
				if (defined($attr->validate_catalogue())) 
				{ 
					my $n = $self->POM()->catalogue->exists($attr->validate_catalogue())
						->target_mem_name();
					$self->user_prog()->code("-- validate: I<@{[ $attr->validate_catalogue() ]}>");
					$self->user_prog()->over();
					my $names;
					my $pipe='';
					foreach ($self->POM()->catalogue->exists($attr->validate_catalogue())->to_array())
				   	{
						$names .= ($pipe . $_->$n());
						$pipe=' | ';
						if (length($names) >= 60) {
							$self->user_prog()->code("B<@{[ $names ]}>");
							$names='';
							$pipe='| ';
						}
					}
					$self->user_prog()->code("B<@{[ $names ]}>") if (length($names));
					$self->user_prog()->back();
				}	
				elsif (defined($attr->validate())) 
				{
					$self->user_prog()->code("-- validate: B<@{[ $attr->validate() ]}>");
					$self->user_prog()->code();
				}
				$self->user_prog()->code();
				$self->user_prog()->back();
			}
			$self->user_prog()->back();

			# Item attributes
			$self->user_prog()->code("F<Item Attributes>");
			$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
			$self->user_prog()->over();
			if ($_->items()->factory()->attributes()->to_array() == 0) {
				$self->user_prog()->code("I<N/A>");
				$self->user_prog()->code();
			}
			# attributes contains ETL::Pequel3::Type::Properties::Abstract type elements;
			foreach my $attr ($_->items()->factory()->attributes()->to_array()) 
			{
				$self->user_prog()->code("B<@{[ $attr->name() ]}>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->over();
				if (defined($attr->description()) && $attr->description() ne '') {
					$self->user_prog()->code("I<@{[ $attr->description() ]}>");
					$self->user_prog()->code();
				}
				$self->user_prog()->code($attr->required() ? "-- required" : "-- optional");
				$self->user_prog()->code();
				$self->user_prog()->code("-- default: B<@{[ $attr->default() ]}>") if (defined($attr->default() ));
				$self->user_prog()->code() if (defined($attr->default() ));
				if (defined($attr->validate_catalogue())) 
				{ 
					my $n = $self->POM()->catalogue->exists($attr->validate_catalogue())
						->target_mem_name();
					$self->user_prog()->code("-- validate: I<@{[ $attr->validate_catalogue() ]}>");
					$self->user_prog()->over();
					my $names;
					my $pipe='';
					foreach ($self->POM()->catalogue->exists($attr->validate_catalogue())->to_array())
				   	{
						$names .= ($pipe . $_->$n());
						$pipe=' | ';
						if (length($names) >= 60) {
							$self->user_prog()->code("B<@{[ $names ]}>");
							$names='';
							$pipe='| ';
						}
					}
					$self->user_prog()->code("B<@{[ $names ]}>") if (length($names));
					$self->user_prog()->back();
				}	
				elsif (defined($attr->validate())) 
				{
					$self->user_prog()->code("-- validate: B<@{[ $attr->validate() ]}>");
				}
				$self->user_prog()->code() if (defined($attr->validate() ));
				$self->user_prog()->code();
				$self->user_prog()->back();
			}
			$self->user_prog()->back();
			#$self->user_prog()->code("F<Related Section Types>");
			# if can('on') then
			# section_on_input
			# section_on_output

			# Pequel XSD Schema:
			if ($self->doc_ref()->show_xsd()) {
				my $xsd = $self->catalogue()->xml_schemas()->exists($self->properties()->xml_schema())
					->new(POM => $self->POM());
				$_->xml_schema($xsd, $xsd->schema_root());
				$self->user_prog()->code("F<pequel xsd extract>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->begin();
				$self->user_prog()->code(' ' . $xsd->to_string());
				$self->user_prog()->end();
			}
#?			$self->user_prog()->page(); #if ($self->doc_ref()->show_xsd() && $self->doc_ref()->target_fmt() eq 'pdf');
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Aggregate; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'aggregate');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->aggregates()->to_array()) 
		{
			$self->user_prog()->head2($_->name());
			$self->user_prog()->code($_->description());
			$self->user_prog()->code(); #? if ($self->doc_ref()->target_fmt() ne 'pdf');
			# Pequel XSD Schema:
			if ($self->doc_ref()->show_xsd()) {
				my $xsd = $self->catalogue()->xml_schemas()->exists($self->properties()->xml_schema())
					->new(POM => $self->POM());
				$_->xml_schema($xsd, $xsd->schema_root());
				$self->user_prog()->code("F<pequel xsd extract>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->begin();
				$self->user_prog()->code(' ' . $xsd->to_string());
				$self->user_prog()->end();
			}
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Table; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'table');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach (reverse $self->catalogue()->table_types()->to_array()) 
		{
			$self->user_prog()->head2($_->name());
			$self->user_prog()->code($_->description());
			$self->user_prog()->code(); #? if ($self->doc_ref()->target_fmt() ne 'pdf');
			# Pequel XSD Schema:
			if ($self->doc_ref()->show_xsd()) {
				my $xsd = $self->catalogue()->xml_schemas()->exists($self->properties()->xml_schema())
					->new(POM => $self->POM());
				$_->xml_schema($xsd, $xsd->schema_root());
				$self->user_prog()->code("F<pequel xsd extract>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->begin();
				$self->user_prog()->code(' ' . $xsd->to_string());
				$self->user_prog()->end();
			}
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Macro; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers qw( macro_category ),
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'macro');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach (grep($_->category() eq $self->macro_category(), $self->catalogue()->macros()->to_array()))
		{
			$self->user_prog()->code("I<@{[ $_->usage() ]}>");
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
		$self->user_prog()->code();
		foreach (grep($_->category() eq $self->macro_category(), $self->catalogue()->macros()->to_array()))
		{
			$self->user_prog()->head2($_->name());
			$self->user_prog()->code();
			if (defined($_->usage())) {
				$self->user_prog()->code("B<@{[ $_->usage() ]}>");
				$self->user_prog()->code();
			}
			$self->user_prog()->code("@{[ $_->description() ]}") if (defined($_->description()));
			$self->user_prog()->code();
			$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::PequelType; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'pequel_type');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->pequel_types()->to_array())
		{
			$self->user_prog()->head2($_->name());
			$self->user_prog()->code("I<@{[ $_->description() ]}>");
			$self->user_prog()->code();
			$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
			foreach my $k (sort keys %{$_->members()}) 
			{
				next if ($k =~ /^(data|description|name|data_type)$/);
				my $attr = ${$_->members()}{$k};
				$self->user_prog()->code("B<@{[ $k ]}>");
				$self->user_prog()->code();
				if (${$attr}[0] ne '') {
					$self->user_prog()->code("-- default: B<@{[ ${$attr}[0] ]}>");
					$self->user_prog()->code();
				}
				if (${$attr}[1] ne '') {
					$self->user_prog()->code("-- validate: B<@{[ ${$attr}[1] ]}>");
					$self->user_prog()->code();
				}
				if ($k eq 'validate_catalogue' && defined($_->validate_catalogue()))
				{ 
					my $n = $self->POM()->catalogue->exists(${$attr}[0])->target_mem_name();
					$self->user_prog()->code("-- validate: I<@{[ ${$attr}[0] ]}>");
					$self->user_prog()->over();
					my $names='';
					my $pipe='';
					foreach ($self->POM()->catalogue->exists(${$attr}[0])->to_array())
				   	{
						$names .= ($pipe . $_->$n());
						$pipe=' | ';
						if (length($names) >= 60) {
							$self->user_prog()->code("B<@{[ $names ]}>");
							$names='';
							$pipe='| ';
						}
					}
					$self->user_prog()->code("B<@{[ $names ]}>") if (length($names));
					$self->user_prog()->back();
				}	
			}
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::DateType; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'date_type');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->date_types()->to_array())
		{
			$self->user_prog()->head2($_->name());
			$self->user_prog()->code();
			$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
			if (defined($_->regex())) {
				$self->user_prog()->code("-- regex: B<@{[ $_->regex() ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->delimiter())) {
				$self->user_prog()->code("-- delimiter: B<@{[ $_->delimiter() ]}>");
				$self->user_prog()->code();
			}
			$self->user_prog()->code("-- d.pos: B<@{[ $_->d()->pos() ]}>;  d.len: B<@{[ $_->d()->len() ]}>");
			$self->user_prog()->code();
			$self->user_prog()->code("-- m.pos: B<@{[ $_->m()->pos() ]}>;  m.len: B<@{[ $_->m()->len() ]}>");
			$self->user_prog()->code();
			$self->user_prog()->code("-- y.pos: B<@{[ $_->y()->pos() ]}>;  y.len: B<@{[ $_->y()->len() ]}>");
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::InputDataset; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'input_dataset');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->input_datasets()->to_array())
		{
			$self->user_prog()->head2($_->name());
			if (defined($_->description()) && $_->description() ne '') {
				$self->user_prog()->code($_->description());
				$self->user_prog()->code();
				$self->user_prog()->code();
			}
			foreach my $k (sort keys %{$_->members()}) 
			{
				my $attr = ${$_->members()}{$k};
				next unless UNIVERSAL::isa(${$attr}[2], 'ETL::Pequel3::Type::DataMember::User');
				$self->user_prog()->code("B<@{[ $k ]}>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->over();
				if (${$attr}[0] ne '') {
					$self->user_prog()->code("-- default: B<@{[ ${$attr}[0] ]}>");
					$self->user_prog()->code();
				}
				if (${$attr}[1] ne '') {
					$self->user_prog()->code("-- validate: B<@{[ ${$attr}[1] ]}>");
					$self->user_prog()->code();
				}
				$self->user_prog()->back();
			}
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::OutputDataset; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'output_dataset');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->output_datasets()->to_array())
		{
			$self->user_prog()->head2($_->name());
			if (defined($_->description()) && $_->description() ne '') {
				$self->user_prog()->code($_->description());
				$self->user_prog()->code();
				$self->user_prog()->code();
			}
			foreach my $k (sort keys %{$_->members()}) 
			{
				my $attr = ${$_->members()}{$k};
				next unless UNIVERSAL::isa(${$attr}[2], 'ETL::Pequel3::Type::DataMember::User');
				$self->user_prog()->code("B<@{[ $k ]}>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->over();
				if (${$attr}[0] ne '') {
					$self->user_prog()->code("-- default: B<@{[ ${$attr}[0] ]}>");
					$self->user_prog()->code();
				}
				if (${$attr}[1] ne '') {
					$self->user_prog()->code("-- validate: B<@{[ ${$attr}[1] ]}>");
					$self->user_prog()->code();
				}
				$self->user_prog()->back();
			}
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::InputDatasource; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'input_datasource');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->input_datasources()->to_array())
		{
			$self->user_prog()->head2($_->datasource_name());
			if (defined($_->description()) && $_->description() ne '') {
				$self->user_prog()->code($_->description());
				$self->user_prog()->code();
				$self->user_prog()->code();
			}
			foreach my $k (sort keys %{$_->members()}) 
			{
				my $attr = ${$_->members()}{$k};
				next unless UNIVERSAL::isa(${$attr}[2], 'ETL::Pequel3::Type::DataMember::User');
				$self->user_prog()->code("B<@{[ $k ]}>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->over();
				if (${$attr}[0] ne '') {
					$self->user_prog()->code("-- default: B<@{[ ${$attr}[0] ]}>");
					$self->user_prog()->code();
				}
				if (${$attr}[1] ne '') {
					$self->user_prog()->code("-- validate: B<@{[ ${$attr}[1] ]}>");
					$self->user_prog()->code();
				}
				$self->user_prog()->back();
			}
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::OutputDatasource; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'output_datasource');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->output_datasources()->to_array())
		{
			$self->user_prog()->head2($_->datasource_name());
			if (defined($_->description()) && $_->description() ne '') {
				$self->user_prog()->code($_->description());
				$self->user_prog()->code();
				$self->user_prog()->code();
			}
			foreach my $k (sort keys %{$_->members()}) 
			{
				my $attr = ${$_->members()}{$k};
				next unless UNIVERSAL::isa(${$attr}[2], 'ETL::Pequel3::Type::DataMember::User');
				$self->user_prog()->code("B<@{[ $k ]}>");
				$self->user_prog()->code() if ($self->doc_ref()->target_fmt() ne 'pdf');
				$self->user_prog()->over();
				if (${$attr}[0] ne '') {
					$self->user_prog()->code("-- default: B<@{[ ${$attr}[0] ]}>");
					$self->user_prog()->code();
				}
				if (${$attr}[1] ne '') {
					$self->user_prog()->code("-- validate: B<@{[ ${$attr}[1] ]}>");
					$self->user_prog()->code();
				}
				$self->user_prog()->back();
			}
			$self->user_prog()->code();
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Property; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'configuration_property');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach (grep($_->isa('ETL::Pequel3::Type::Properties::Configuration::Abstract'), 
				$self->catalogue()->properties()->to_array()))
		{
			$self->user_prog()->head2($_->name());
			if (defined($_->description())) {
				$self->user_prog()->code("I<@{[ $_->description() ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->default())) {
				$self->user_prog()->code("-- default: B<@{[ $_->default() ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->validate())) {
				$self->user_prog()->code("-- validate: B<@{[ $_->validate() ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->validate_catalogue()))
			{ 
				my $n = $self->POM()->catalogue->exists($_->validate_catalogue())->target_mem_name();
				$self->user_prog()->code("-- validate: I<@{[ $_->validate_catalogue() ]}>");
				$self->user_prog()->over();
				my $names='';
				my $pipe='';
				foreach ($self->POM()->catalogue->exists($_->validate_catalogue())->to_array())
				{
					$names .= ($pipe . $_->$n());
					$pipe=' | ';
					if (length($names) >= 60) {
						$self->user_prog()->code("B<@{[ $names ]}>");
						$names='';
						$pipe='| ';
					}
				}
				$self->user_prog()->code("B<@{[ $names ]}>") if (length($names));
				$self->user_prog()->back();
			}	
			if (defined($_->global()) && $_->global() != 0) {
				$self->user_prog()->code("-- global");
				$self->user_prog()->code();
			}
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::CmdOption; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'command_line_option');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		foreach ($self->catalogue()->options()->to_array())
		{
			$self->user_prog()->head2("--@{[ $_->long_name() ]}");
			if (defined($_->description())) {
				$self->user_prog()->code("I<@{[ $_->description() ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->short_name())) {
				$self->user_prog()->code("-- short name: B<--@{[ $_->short_name() ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->default())) {
				$self->user_prog()->code("-- default: B<@{[ $_->default() ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->flag())) {
				$self->user_prog()->code("-- flag: B<@{[ $_->flag() ? 'YES' : 'NO' ]}>");
				$self->user_prog()->code();
			}
			if (defined($_->prop_name())) {
				$self->user_prog()->code("-- property name: B<@{[ $_->prop_name() ]}>");
				$self->user_prog()->code();
			}
		}
		$self->user_prog()->code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Document; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'document');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::DocChapter; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'document_chapter');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Generator; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'generator');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Segment; 
	use base qw(ETL::Pequel3::Type::Document::Chapter::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'chapter_name', default => 'generator_segment');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
	}
}
# ----------------------------------------------------------------------------------------------------
# CHAPTER TYPE CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::CatGen::Chapter::Catalogue;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'cat_chapters'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'chapter_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::Document::Chapter::Abstract');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::CatGen::Chapter::Section->_new(),
			ETL::Pequel3::CatGen::Chapter::Aggregate->_new(),
			ETL::Pequel3::CatGen::Chapter::Macro->_new(),
			ETL::Pequel3::CatGen::Chapter::PequelType->_new(),
			ETL::Pequel3::CatGen::Chapter::DateType->_new(),
			ETL::Pequel3::CatGen::Chapter::Table->_new(),
			ETL::Pequel3::CatGen::Chapter::InputDataset->_new(),
			ETL::Pequel3::CatGen::Chapter::OutputDataset->_new(),
			ETL::Pequel3::CatGen::Chapter::InputDatasource->_new(),
			ETL::Pequel3::CatGen::Chapter::OutputDatasource->_new(),
			ETL::Pequel3::CatGen::Chapter::Property->_new(),
			ETL::Pequel3::CatGen::Chapter::CmdOption->_new(),
			ETL::Pequel3::CatGen::Chapter::Document->_new(),
			ETL::Pequel3::CatGen::Chapter::DocChapter->_new(),
			ETL::Pequel3::CatGen::Chapter::Generator->_new(),
			ETL::Pequel3::CatGen::Chapter::Segment->_new(),
#>			ETL::Pequel3::CatGen::Chapter::UserExtension->_new(), # list all extenstions
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
1;
