# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Generator.pm
#  Created	: 19 May 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Generator;
require 5.005_62;
use strict;
use warnings;
# ----------------------------------------------------------------------------------------------------
# GENERATOR TYPE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Abstract; 
	use base qw(Class::STL::Element);
	use stl qw(list);
	use Class::STL::ClassMembers qw( err catalogue properties generator_name main segments user_prog caller_id ),
		Class::STL::ClassMembers::DataMember->new(name => 'show_info', default => 1);
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		use ETL::Pequel3::Error;
		$self->err(ETL::Pequel3::Error->new());
		use ETL::Pequel3::Type::Properties;
		$self->properties(ETL::Pequel3::Type::Properties->new());
		use ETL::Pequel3::Catalogue;
		$self->catalogue(ETL::Pequel3::Catalogue->new());

		$self->segments(list(element_type => 'ETL::Pequel3::Generator::Segment::Abstract'));
		$self->err()->user_error(10501, "Attribute @{[ ref($self) ]}::main is undefined!")
			unless (defined($self->main()));
		$self->err()->user_error(10500, "Attribute @{[ ref($self) ]}::main->config is undefined!")
			unless (defined($self->main()->config()));
		$self->caller_id(do { my $n=$self->main()->pequel_name();$n=~s/\W+/_/g;$n }); #???

		use ETL::Pequel3::CodeStyler;
		$self->user_prog(ETL::Pequel3::CodeStyler::Program::Perl->new(
			@_, # contains 'main', other
			program_name => 
			(
				$self->properties()->exec_dir() 
				. '/' 
				. (split(/[\/]+/, $self->main()->script_filename()))[-1]
				. ".code"
			),
			print_bookmarks => $self->err()->trace_level() >= 5,
			tab_size => $self->main()->config()->tab_size(),
			tab_type => $self->main()->config()->tab_type(),
			divider_char => $self->main()->config()->divider_char(),
			divider_length => $self->main()->config()->divider_length(),
		))
		unless (defined($self->user_prog()));
		return $self;
	}
	sub build
	{
		my $self = shift;
		$self->err()->diag(1, "[@{[ $self->main()->pequel_name() ]}] generate...");
		map($_->prepare(), $self->segments()->to_array());
		$self->user_prog()->prepare();
#>		TODO:use $self-os()->getpath(...)
		mkdir("@{[ $self->properties()->exec_dir() ]}");
		$self->user_prog()->save() if ($self->err()->trace_level() >= 5);
		$self->err->trace_msg(10, "\n++++++\n@{[ $self->user_prog()->display() ]}\n+++++\n");
	}
	sub exec
	{
		my $self = shift;
		$self->user_prog()->run();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::MainProgram; 
	use base qw(ETL::Pequel3::Generator::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'generator_name', default => 'main_program');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->segments()->push_back(
			$self->catalogue()->code_segments()->startup()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->user_options()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->init_global()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->open_global()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->call_load_table()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->call_execute()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->close_global()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->divider()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->main_package()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->load_table()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->sub_programs()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->end_of_program()->new(@_, user_prog => $self->user_prog() ),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::SubProgram; 
	use base qw(ETL::Pequel3::Generator::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'generator_name', default => 'sub_program');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->segments()->push_back(
			$self->catalogue()->code_segments()->main_package()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->sub_programs()->new(@_, user_prog => $self->user_prog() ),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::TableLoader; 
	use base qw(ETL::Pequel3::Generator::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'generator_name', default => 'table_loader');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->segments()->push_back(
			$self->catalogue()->code_segments()->startup()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->user_options()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->header_info()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->load_table()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->sub_programs()->new(@_, user_prog => $self->user_prog() ),
			$self->catalogue()->code_segments()->end_of_program()->new(@_, user_prog => $self->user_prog() ),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
# GENERATOR TYPE CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Catalogue;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'generators'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'generator_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Generator::Abstract');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Generator::MainProgram->_new(),
			ETL::Pequel3::Generator::SubProgram->_new(),
			ETL::Pequel3::Generator::TableLoader->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
# SEGMENT TYPE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Abstract; 
	use base qw(Class::STL::Element);
	use Class::STL::ClassMembers 
		qw( err properties catalogue package_name user_prog main segment_name segments ),
		Class::STL::ClassMembers::DataMember->new(name => 'show_info', default => 1);
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		use ETL::Pequel3::Error;
		$self->err(ETL::Pequel3::Error->new()); # Singleton
		use ETL::Pequel3::Type::Properties;
		$self->properties(ETL::Pequel3::Type::Properties->new()); # Singleton
		use ETL::Pequel3::Catalogue;
		$self->catalogue(ETL::Pequel3::Catalogue->new());

		$self->err()->user_error(10502, "Class @{[ ref($self) ]}::user_prog is undefined!")
			unless (defined($self->user_prog()));
		use stl qw(list);
		$self->segments(list(element_type => 'ETL::Pequel3::Generator::Segment::Abstract'));
		$self->package_name("Pequel::@{[ do { my $n=$self->main()->pequel_name();$n=~s/\W+/_/g;$n } ]}")
			unless (defined($self->package_name()) || !defined($self->main()));
	}
	sub program_info
	{
		my $self = shift;
		my $extra = shift;
		return unless ($self->show_info());
		$self->user_prog()->comment("");
		$self->user_prog()->comment(" ++++ @{[ join(' ', map(ucfirst($_), split(/_/, $self->segment_name()))) ]}@{[ defined($extra) ? qq/ ($extra)/ : '' ]} Segment ++++");
		$self->user_prog()->comment("");
	}
	sub prepare_start
	{
		my $self = shift;
		$self->err()->diag(20, "...@{[ $_->segment_name() ]}...");
		map($_->prepare_start(), $self->segments()->to_array());
	}
	sub prepare_finish
	{
		my $self = shift;
		map($_->prepare_finish(), CORE::reverse($self->segments()->to_array()));
	}
	sub prepare
	{
		my $self = shift;
		$self->prepare_start();
		$self->prepare_finish();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::SubPrograms; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'sub_programs');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare
	{
		my $self = shift;
		$self->program_info();
#?		foreach ($self->main()->all_pequels()->to_array()) {
		foreach (grep($_->duplicate() == 0, $self->main()->user_pequels()->to_array())) {
			my $sub = ETL::Pequel3::Generator::SubProgram->new(main => $_);
			$sub->build();
			$self->user_prog()->add($sub->user_prog());
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::MainPackage; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'main_package');
	use Class::STL::ClassMembers::Constructor; 
	sub new_extra
	{
		my $self = shift;
		$self->segments()->push_back(
			$self->catalogue()->code_segments()->package()->new(@_),
			$self->catalogue()->code_segments()->header_info()->new(@_),
			$self->catalogue()->code_segments()->sub_execute_body()->new(@_),
		);
	}
}
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Generator::Segment::SubExecuteBody; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'sub_execute_body');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->segments()->push_back(
			$self->catalogue()->code_segments()->sub_execute_def()->new(@_),
			$self->catalogue()->code_segments()->declaration()->new(@_),
			$self->catalogue()->code_segments()->init_local()->new(@_),
			$self->catalogue()->code_segments()->open_local()->new(@_),
			$self->catalogue()->code_segments()->finish()->new(@_),
			$self->catalogue()->code_segments()->main()->new(@_),
		);
	}
}	
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Generator::Segment::SubExecuteDef; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'sub_execute_def');
	use Class::STL::ClassMembers::Constructor;
	sub prepare_start
	{
		my $self = shift;
		$self->user_prog()->code("sub execute");
		$self->user_prog()->open_block();
	}
	sub prepare_finish
	{
		my $self = shift;
		$self->user_prog()->close_block();
	}
}	
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Startup; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'startup');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->user_prog()->indent_off();
		$self->user_prog()->comment("!$^X");
		$self->user_prog()->indent_on();
		$self->user_prog()->divider();
		$self->user_prog()->comment(" vim: syntax=perl ts=@{[ 
			$self->main()->config()->tab_type() eq 'hard' 
				? '1' :  $self->main()->config()->tab_size() 
			]} sw=@{[ $self->main()->config()->tab_size() ]}");
		$self->user_prog()->divider();
		$self->user_prog()->comment(" Generated By: Pequel3 @{[ $self->properties()->version() ]}");
		$self->user_prog()->comment("             : http://sourceforge.net/projects/pequel/");
		$self->user_prog()->comment(" Created On  : @{[ '' . localtime() ]}");
		$self->user_prog()->comment(" Perl Version: @{[ $^X . ' ' . sprintf(qq{%vd}, $^V) . ' on ' . $^O ]}");
		$self->user_prog()->comment(" For         : @{[ getlogin() || '' ]}");
		$self->user_prog()->divider();
		$self->user_prog()->code("use strict;");
		$self->user_prog()->code("use warnings;");
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Divider; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'divider');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->user_prog()->divider();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::EndOfProgram; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'end_of_program');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->user_prog()->divider();
		$self->user_prog()->comment(" ++++ END OF PROGRAM -- @{[ $self->main()->script_filename() ]} ++++");
		$self->user_prog()->divider();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::CallLoadTable; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'call_load_table');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		map
		(
			$_->code_start($self->user_prog()),
			grep(defined($_), $self->main()->user_tables()->to_array())
		);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::CallExecute; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'call_execute');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->user_prog()->code("&@{[ $self->package_name() ]}::execute();");
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Package; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'package');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->user_prog()->open_block();
		$self->user_prog()->code("package @{[ $self->package_name() ]};");
	}
	sub prepare_finish
	{
		my $self = shift;
		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::HeaderInfo; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'header_info');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->user_prog()->divider();
		$self->user_prog()->comment(" Script Name : @{[ $self->main()->pequel_name() ]}");
		$self->user_prog()->divider();
		$self->user_prog()->comment(" Configuration:");
		foreach 
		(
			grep
			(
				$_->user_program_info() != 0 && defined($_->value()),
					grep($_->global(), $self->properties()->to_array()), 
					$self->main()->config()->to_array()
			)
		) 
		{
			$self->user_prog()->comment
			(
				" "
				. $_->name()
				. " -- "
				. ($_->value() =~ /^(0|1)$/ ? ($_->value() == 0 ? 'NO' : 'YES') : $_->value())
				. " -- "
				. $_->description()
				. ($_->global() ? ' -- (global)' : '')
			);
		}
		$self->user_prog()->divider();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Declaration;  # all local
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'declaration');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		foreach 
		(
			grep
			(
				defined($_),
				(
					$self->main()->input()->input_fields(),
					$self->main()->output()->output_fields(),
					map($_->input_fields(), $self->main()->user_joins()->to_array()),
					map($_->fields(), $self->main()->user_tables()->to_array()),
					map($_->output_fields(), 
						map($_->reason_codes_list()->to_array(), 
							$self->main()->user_sections()->reject_record_on_input()->to_array())),
					map($_->output_fields(), 
						map($_->reason_codes_list()->to_array(), 
							$self->main()->user_sections()->reject_record_on_output()->to_array())),
				)
			)
		)
		{
			$_->code_decl($self->user_prog());
		}
		$self->user_prog()->code("use constant _VERBOSE => int @{[ $self->main()->config()->verbose() ]};")
			if ($self->main()->config()->verbose());

		$self->user_prog()->code("use constant @{[ $self->main()->input()->input_dataset()->numflds_vname() ]} => int "
		   . int(grep(!defined($_->calc_exp()), $self->main()->input()->input_fields()->to_array())) 
		   . ";");

		foreach my $j ($self->main()->user_joins()->to_array()) 
		{
			$self->user_prog()->code("use constant @{[ $j->input_dataset()->numflds_vname() ]} => int "
			   . int(grep(!defined($_->calc_exp()), $j->input_fields()->to_array())) 
			   . ";");
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Generator::Segment::InitGlobal;
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'init_global');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		# Only wan't to the ds_var code from datasource which is relevant to the open command:
		map($_->datasource()->_code_init($self->user_prog()), $self->main()->shared_datasets()->to_array());
	}
}	
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::OpenGlobal; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'open_global');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		foreach ($self->main()->shared_datasets()->to_array())
		{
			$self->user_prog()->comment("");
			$self->user_prog()->comment(" ---> @{[ 
					ucfirst($_->datasource()->io_name()) 
				]}: @{[ $_->datasource()->fdname() ]} (@{[ 
				$_->name() ]}:@{[ $_->datasource()->datasource_name() ]})");
			$self->user_prog()->comment("");
			$_->code_open($self->user_prog());
		}
		foreach 
		(
			grep
			(
				$_->can('print_header') && $_->print_header(),
				(
					$self->main()->shared_datasets()->to_array(), 
				)
			)
		) 
		{
			$self->catalogue()->code_segments()->print_output_header()->new(
				main => $self->main(), user_prog => $self->user_prog(), dataset => $_)
				->prepare();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::CloseGlobal;
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'close_global');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		map($_->code_close($self->user_prog()), reverse $self->main()->shared_datasets()->to_array());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::InitLocal; #-->Initialisation
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'init_local');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		my %__shared_ds;
		$self->program_info();

		if ($self->main->macro_use_list()->size() > 1) {
			stl::stable_sort($self->main->macro_use_list()->begin(), $self->main->macro_use_list()->end());
			stl::unique($self->main->macro_use_list()->begin(), $self->main->macro_use_list()->end());
		}
		foreach ($self->main->macro_use_list()->to_array()) {
			$_->code_init_global($self->user_prog());
			#TODO:code_init()
		}

		foreach 
		(
			map($_->aggregate_type(), grep(defined($_->aggregate_type()), $self->main()->output()->output_fields()->to_array())),
			grep
			(
				defined($_) && (!$_->can('shared') || !$_->shared() || ++$__shared_ds{$_->arr_vname()} <= 1),
				(
#<					$self->main()->user_datasets()->to_array() # !! contains repeated ds's
					map($_->input_dataset(), $self->main()->user_joins()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->copy_record_on_input()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->copy_record_on_output()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->divert_record_on_input()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->divert_record_on_output()->to_array()),
					map($_->dataset(), $self->main()->user_sections()->reject_record_on_input()->to_array()),
					map($_->dataset(), $self->main()->user_sections()->reject_record_on_output()->to_array()),
					$self->main()->input()->input_dataset(),
					$self->main()->output()->transfer_dataset(),
					map($_->transfer_dataset(), $self->main()->user_joins()->to_array()),
					$self->main()->output()->output_dataset(),
				)
			)
		)
		{
			$_->code_init($self->user_prog()); # arr_vname
		}
		if ($self->main()->user_sections()->group_by()->size()) {
			$self->user_prog()->comment("");
			$self->user_prog()->comment(" ---> Break control grouping fields");
			$self->user_prog()->comment("");
			foreach ($self->main()->group_by()->input_fields()->to_array()) {
				$self->user_prog()->code("my \$key@{[ $_->tname() ]};");
				$self->user_prog()->code("my \$previous_key@{[ $_->tname() ]} = undef;");
			}
			if ($self->main()->config()->hash()) {
				$self->user_prog()->comment("");
				$self->user_prog()->comment(" ---> Configuration >> hash <<");
				$self->user_prog()->comment("");
				$self->user_prog()->code("my \$_HKEY;");
			}
		}
		if ($self->main()->user_sections()->dedup_on_input()->size()) {
			foreach ($self->main()->dedup_on_input()->fields()->to_array()) {
				$self->user_prog()->code("my \$previous_dedup@{[ $_->input_field()->tname() ]} = undef;");
			}
		}
		if ($self->main()->user_sections()->dedup_on_output()->size()) {
			foreach ($self->main()->dedup_on_output()->fields()->to_array()) {
				$self->user_prog()->code("my \$previous_dedup@{[ $_->output_field()->tname() ]} = undef;");
			}
		}
		if ($self->main()->config()->verbose() && $self->main()->config()->statistics()) {
			$self->user_prog()->comment("");
			$self->user_prog()->comment(" ---> Configuration >> statistics <<");
			$self->user_prog()->comment("");
			$self->user_prog()->code("use Benchmark;");
			$self->user_prog()->code("my \$benchmark_start = new Benchmark;");
		}

		my $r=0;
		map
		(
			$self->user_prog()->code("my \$_IRVAL_@{[ ++$r ]}; # Input repeater $r value"),
			$self->main()->repeat_record_on_input()->fields()->to_array()
		)
		if ($self->main()->user_sections()->repeat_record_on_input()->size());

		$r=0;
		map
		(
			$self->user_prog()->code("my \$_ORVAL_@{[ ++$r ]}; # Output repeater $r value"),
			$self->main()->repeat_record_on_output()->fields()->to_array()
		)
		if ($self->main()->user_sections()->repeat_record_on_output()->size());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::OpenLocal; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'open_local');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		foreach (grep(!$_->shared(), $self->_get_datasets()))
		{
			$self->user_prog()->comment("");
			$self->user_prog()->comment(" ---> @{[ 
				ucfirst($_->datasource()->io_name()) 
				]}: @{[ $_->datasource()->fdname() ]} (@{[ 
				$_->name() ]}:@{[ $_->datasource()->datasource_name() ]})");
			$self->user_prog()->comment("");
			$_->code_open($self->user_prog()); 
		}
		# Output Header
		foreach 
		(
			grep
			(
				defined($_) && !$_->shared() && $_->print_header(),
				(
					$self->main()->output()->output_dataset(),
					map($_->output_dataset(), $self->main()->user_sections()->divert_record_on_output()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->copy_record_on_output()->to_array()),
					map($_->dataset(), $self->main()->user_sections()->reject_record_on_output()->to_array()),
				)
			)
		) 
		{
			$self->catalogue()->code_segments()->print_output_header()->new(
				main => $self->main(), user_prog => $self->user_prog(), dataset => $_)
				->prepare();
		}
		# Input Header
		foreach 
		(
			grep
			(
				defined($_) && !$_->shared() && $_->print_header(),
				(
					map($_->output_dataset(), $self->main()->user_sections()->divert_record_on_input()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->copy_record_on_input()->to_array()),
					map($_->dataset(), $self->main()->user_sections()->reject_record_on_input()->to_array()),
				)
			)
		) 
		{
			$self->catalogue()->code_segments()->print_input_header()->new(
				main => $self->main(), user_prog => $self->user_prog(), dataset => $_)
				->prepare();
		}
	}
	sub prepare_finish
	{
		my $self = shift;
		$self->program_info();
		foreach (reverse grep(!$_->shared(), $self->_get_datasets())) {
			$_->code_close($self->user_prog());
		}
	}
	sub _get_datasets
	{
		my $self = shift;
		return (
			grep
			(
				defined($_),
				(
					map($_->input_dataset(), $self->main()->user_joins()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->copy_record_on_input()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->copy_record_on_output()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->divert_record_on_input()->to_array()),
					map($_->output_dataset(), $self->main()->user_sections()->divert_record_on_output()->to_array()),
					map($_->dataset(), $self->main()->user_sections()->reject_record_on_input()->to_array()),
					map($_->dataset(), $self->main()->user_sections()->reject_record_on_output()->to_array()),
					$self->main()->input()->input_dataset(),
					$self->main()->output()->output_dataset()
				)
			)
		);	
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Main; #--> Input
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'main');
	use Class::STL::ClassMembers::Constructor; 
	sub new_extra
	{
		my $self = shift;
	
		$self->segments()->push_back($self->catalogue()->code_segments()->main_while_loop()->new(@_));

		$self->segments()->push_back($self->catalogue()->code_segments()->input_record_counter()->new(@_));

		$self->segments()->push_back($self->catalogue()->code_segments()->input_join()->new(@_))
			if ($self->main()->user_joins()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->input_repeater()->new(@_))
			if ($self->main()->user_sections()->repeat_record_on_input()->size());
	
		# TODO: this should be right after main_while_loop
#<		$self->segments()->push_back($self->catalogue()->code_segments()->input_record_counter()->new(@_));
	
		$self->segments()->push_back($self->catalogue()->code_segments()->input_record_limit()->new(@_))
			if ($self->main()->config()->input_record_limit());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->derived_input_fields()->new(@_));
	
		$self->segments()->push_back($self->catalogue()->code_segments()->field_process_input()->new(@_))
			if ($self->main()->user_sections()->field_process_on_input()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->input_filter()->new(@_))
			if ($self->main()->user_sections()->filter_on_input()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->reject_input_record()->new(@_))
			if ($self->main()->user_sections()->reject_record_on_input()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->divert_input_record()->new(@_))
			if ($self->main()->user_sections()->divert_record_on_input()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->copy_input_record()->new(@_))
			if ($self->main()->user_sections()->copy_record_on_input()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->grouping_break_control()->new(@_))
			if ($self->main()->user_sections()->group_by()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->dedup_on_input()->new(@_))
			if ($self->main()->user_sections()->group_by()->size()); #< && !$self->main()->config()->hash());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->main_body()->new(@_));
	
		$self->segments()->push_back($self->catalogue()->code_segments()->prepare_transfer_fields()->new(@_))
			if ($self->main()->output()->transfer_fields()->size());
	
		$self->segments()->push_back($self->catalogue()->code_segments()->output()->new(@_))
			if ($self->main()->user_sections()->group_by()->size() == 0);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::MainWhileLoop; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'main_while_loop');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->catalogue()->code_segments()->verbose_msg()->new(
			target_name => $self->main()->pequel_name(),
			user_prog => $self->user_prog(), 
			msg_text => "Start",
		)->prepare() if ($self->main()->config()->verbose()); 
		$self->main()->input()->input_dataset()->code_prepare($self->user_prog()); 
		$self->user_prog()->newline_off();

		$self->user_prog()->code("while (");
		$self->main()->input()->input_dataset()->code_read($self->user_prog(), 
		   grep(!defined($_->calc_exp()), $self->main()->input()->input_fields()->to_array()));
		$self->user_prog()->newline_on();
		$self->user_prog()->code(")");
		$self->user_prog()->open_block(); # main while loop
			$self->main()->input()->input_dataset()->code_after_read($self->user_prog(),
			   grep(!defined($_->calc_exp()), $self->main()->input()->input_fields()->to_array())); #TODO: remove -- now done in parse
			$self->catalogue()->code_segments()->input_verify()->new(
				main => $self->main(), 
				user_prog => $self->user_prog(), 
				dataset => $self->main()->input()->input_dataset(),
				input_fields => $self->main()->input()->input_fields(), 
			)->prepare()
				if ($self->main()->input()->input_dataset()->verify());
	}
	sub prepare_finish
	{
		my $self = shift;
		$self->user_prog()->close_block(); # main while loop

		if ($self->main()->user_sections()->group_by()->size()) {
			my %ts_set;
			if (!$self->main()->config()->hash()) {
				$ts_set{$self->main()->output()->output_dataset()->arr_vname()}++ 
					if ($self->main()->output()->output_fields()->size());
				map($ts_set{$_->dataset()->arr_vname()}++, $self->main()->output()->transfer_fields()->to_array());
				if (keys(%ts_set)) {
					$self->user_prog()->code("if (" . join(' || ', map("\@$_", keys(%ts_set))) . ")");
					$self->user_prog()->open_block();
				}
			}
			$self->catalogue()->code_segments()->output()->new(
				main => $self->main(), user_prog => $self->user_prog())->prepare();
			$self->user_prog()->close_block() if (!$self->main()->config()->hash() && keys(%ts_set));
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::InputJoin; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'input_join');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		my $r=0;
		foreach my $j ($self->main()->user_joins()->to_array()) 
		{
			$self->user_prog()->comment(" +++ Input Join (@{[ $j->input_dataset()->datasource()->fdname() ]}) Segment +++");
			$self->user_prog()->code("if");
			$self->user_prog()->open_block("(");
				$self->user_prog()->code("!defined(@{[ $j->key_field_ref()->getvar() ]})");
				$self->user_prog()->code("|| "
					. $j->join_field_ref()->getvar()
					. ($j->key_type() =~ /^(numeric|amount)$/ ? ' > ' : ' gt ')
					. $j->key_field_ref()->getvar() 
				);
			$self->user_prog()->close_block();
			$self->user_prog()->open_block();
				$j->input_dataset()->code_prepare($self->user_prog());
				$self->user_prog()->newline_off();
				$self->user_prog()->code("while (");
				$j->input_dataset()->code_read($self->user_prog(), 
				   grep(!defined($_->calc_exp()), $j->input_fields()->to_array()));
				$self->user_prog()->newline_on();
				$self->user_prog()->code(")");
				$self->user_prog()->open_block();
					$j->input_dataset()->code_after_read($self->user_prog(),
					   grep(!defined($_->calc_exp()), $j->input_fields()->to_array())); #TODO: remove -- now done in parse
					$self->user_prog()->code("++\$@{[ $j->input_dataset()->count_vname() ]};")
						if ($self->main()->config()->verbose());
					$self->catalogue()->code_segments()->input_verify()->new(
						main => $self->main(), 
						user_prog => $self->user_prog(), 
						dataset => $j->input_dataset(),
						input_fields => $j->input_fields(), 
					)->prepare()
						if ($j->input_dataset()->verify());
					$self->user_prog()->code("last if (@{[ $j->join_field_ref()->getvar() ]} le @{[ $j->key_field_ref()->getvar() ]});");
				$self->user_prog()->close_block();
			$self->user_prog()->close_block();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::InputVerify; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers qw( input_fields dataset ),
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'input_verify');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info("@{[ $self->dataset()->datasource()->fdname() ]} --> @{[ 
			uc($self->dataset()->verify()) ]}");
		$self->user_prog()->code("my \$fn=0;");
		$self->user_prog()->code("if");
		$self->user_prog()->open_block("(");
			$self->user_prog()->code("int(\@@{[ $self->dataset()->arr_vname() ]}) < @{[ 
				$self->main()->config()->show_synonyms() 
					? $self->dataset()->numflds_vname()
					: int(grep(!defined($_->calc_exp()), $self->input_fields()->to_array())) 
				]}");
			map
			(
				do {
					$self->user_prog()->code("|| (++\$fn!=@{[ $_->field_number() ]})");
					$self->user_prog()->over();
						$self->user_prog()->code("|| !defined(@{[ $_->getvar() ]})");
						$self->user_prog()->code("|| @{[ $_->getvar() ]} !~ /@{[ 
							$_->pequel_type()->validate() ]}/ # @{[ $_->pequel_type()->name() ]}"),
					$self->user_prog()->back();
				},
				grep(!defined($_->calc_exp()), $self->input_fields()->to_array())
			);
		$self->user_prog()->close_block();
		$self->user_prog()->open_block();
			if ($self->dataset()->verify() eq 'abort' || $self->main()->config()->verbose()) {
				$self->user_prog()->code("print STDERR \"Input record \", \$@{[ 
					$self->dataset()->count_vname() ]}, \" mismatch for @{[ $self->dataset()->datasource()->fdname() ]}->\@{[");
				$self->user_prog()->open_block("(");
					$self->user_prog()->code("'<num-fields>',");
					map
					(
						$self->user_prog()->code("'@{[ $_->name() ]}',"),
						grep(!defined($_->calc_exp()), $self->input_fields()->to_array())
					);
				$self->user_prog()->close_block();
				$self->user_prog()->code("[\$fn]");
				$self->user_prog()->code("]}!\\n\";");
			}
			$self->dataset()->verify() eq 'abort' 
				? $self->user_prog()->code("die \"Process aborted.\";")
				: $self->user_prog()->code("next;"); # verify == 'skip'
		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::OutputVerify; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers qw( output_fields dataset ),
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'output_verify');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
#?		return unless ($self->main()->divert_output()->size());
#>		$self->program_info("@{[ $self->dataset()->datasource()->fdname() ]} --> @{[ uc($self->dataset()->verify()) ]}");
#>		$self->user_prog()->code("my \$fn=0;");
#>		$self->user_prog()->code("if");
#>		$self->user_prog()->open_block("(");
#>			$self->user_prog()->code("0");
#>			map
#>			(
#>				$self->user_prog()->code("|| (++\$fn!=@{[ $_->field_number() ]}) || !defined(@{[ $_->getvar() ]}) || @{[ $_->getvar() ]} !~ /@{[ $_->pequel_type()->validate() ]}/ # @{[ $_->pequel_type()->name() ]}"),
#>				grep(!defined($_->calc_exp()), $self->output_fields()->to_array())
#>			);
#>		$self->user_prog()->close_block();
#>		$self->user_prog()->open_block();
#>			if ($self->dataset()->verify() eq 'abort' || $self->main()->configuration()->verbose()) {
#>				$self->user_prog()->code("print STDERR \"Output record \", \$@{[ $self->dataset()->count_vname() ]}, \" mismatch for @{[ $self->dataset()->datasource()->fdname() ]}->\@{[");
#>				$self->user_prog()->over();
#>				$self->user_prog()->code("('<num-fields>', " . join(', ', map(qq{'$_'}, map($_->name(), grep(!defined($_->calc_exp()), $self->output_fields()->to_array())))) . ")[\$fn]");
#>				$self->user_prog()->back();
#>				$self->user_prog()->code("]}!\\n\";");
#>			}
#>			$self->dataset()->verify() eq 'abort' 
#>				? $self->user_prog()->code("die \"Process aborted.\";")
#>				: $self->user_prog()->code("next;"); # verify == 'skip'
#>		$self->user_prog()->close_block();
#>	
#>		$self->user_prog()->code("else");
#>		$self->user_prog()->open_block();
	}
	sub prepare_finish
	{
#?		my $self = shift;
#?		return unless ($self->main()->divert_output()->size());
#?		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::InputRecordCounter; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'input_record_counter');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->user_prog()->code("++\$@{[ $self->main()->input()->input_dataset()->count_vname() ]};")
			if ($self->main()->config()->verbose() || $self->main()->config()->input_record_limit());

		$self->catalogue()->code_segments()->verbose_msg()->new(
			target_name => $self->main()->pequel_name(),
			user_prog => $self->user_prog(), 
			msg_text => "\$@{[ $self->main()->input()->input_dataset()->count_vname() ]} records.", 
			cond => (
				"if (\$@{[ $self->main()->input()->input_dataset()->count_vname() ]} % "
				. ( $self->properties()->show_synonyms() 
					? '_VERBOSE' : $self->main()->config()->verbose() )
				. " == 0)"
			)
		)->prepare() if ($self->main()->config()->verbose()); 
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::InputRepeater; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'input_repeater');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->repeat_record_on_input()->size());
		$self->program_info();
		my $r=0;
		foreach ($self->main()->repeat_record_on_input()->fields()->to_array()) {
			$self->user_prog()->code("foreach \$_IRVAL_@{[ ++$r ]} (@{[ $_->exp_raw() ]})");
			$self->user_prog()->code("# Input repeater $r value (@{[ $_->exp() ]})");
			$self->user_prog()->open_block();
		}
	}
	sub prepare_finish
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->repeat_record_on_input()->size());
		foreach ($self->main()->repeat_record_on_input()->fields()->to_array()) {
			$self->user_prog()->close_block();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::PrintOutputHeader;
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers qw( dataset ),
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'print_output_header');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->dataset()->code_lock($self->user_prog()) if ($self->dataset()->shared());
		$self->dataset()->code_write_header($self->user_prog(),
				grep
				(
					$_->name() !~ /^_/, 
					$self->main()->output()->transfer_fields()->to_array(), 
					$self->main()->output()->output_fields()->to_array()
				)
			);
		$self->dataset()->code_unlock($self->user_prog()) if ($self->dataset()->shared());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::PrintInputHeader;
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers qw( dataset ),
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'print_input_header');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->dataset()->code_lock($self->user_prog()) if ($self->dataset()->shared());
		$self->dataset()->code_write_header($self->user_prog(),
				grep
				(
					$_->name() !~ /^_/, 
					$self->main()->input()->input_fields()->to_array()
				)
			);
		$self->dataset()->code_unlock($self->user_prog()) if ($self->dataset()->shared());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::DerivedInputFields; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'derived_input_fields');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		# Calculate Derived Input Fields:
		# if the field is in use only (by output-field, other input-field, group-by, transfer, etc
		# ...not quite though because some input derived fields may only be used conditionally!
		foreach (grep(defined($_->calc_exp()), $self->main()->input()->input_fields()->to_array())) {
			$self->user_prog()->code("@{[ $_->getvar() ]} @{[ defined($_->operator()) ? $_->operator() : '=' ]} # @{[ $_->calc_exp() ]}");
			$self->user_prog()->over();
				$self->user_prog()->code($_->calc_exp_raw());
				$self->user_prog()->code(";");
			$self->user_prog()->back();
		}
		foreach my $j ($self->main()->user_joins()->to_array()) {
			foreach (grep(defined($_->calc_exp()), $j->input_fields()->to_array())) {
				$self->user_prog()->code("@{[ $_->getvar() ]} @{[ defined($_->operator()) ? $_->operator() : '=' ]} # @{[ $_->calc_exp() ]}");
				$self->user_prog()->over();
					$self->user_prog()->code($_->calc_exp_raw());
					$self->user_prog()->code(";");
				$self->user_prog()->back();
			}
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::RejectInputRecord; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'reject_input_record');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->reject_record_on_input()->size());
		$self->program_info();
		my $join='if';
		foreach my $ri ($self->main()->user_sections()->reject_record_on_input()->to_array()) {
			foreach my $r ($ri->reason_codes_list()->to_array()) {
				#$r->reason_code() ...
				#$r->condition_exp_raw() ...
				$self->user_prog()->comment(" +++ Reason Code (@{[ $r->reason_code() ]}) +++");
				$self->user_prog()->newline_off();
				$self->user_prog()->code("$join (");
				$self->user_prog()->code($r->condition_exp_raw());
				$self->user_prog()->newline_on();
				$self->user_prog()->code(")");
				$self->user_prog()->open_block();

				foreach (grep(defined($_->calc_exp()), $r->output_fields()->to_array())) {
					$self->user_prog()->code("@{[ $_->getvar() ]} = # @{[ $_->calc_exp() ]}");
					$self->user_prog()->over();
						$self->user_prog()->code($_->calc_exp_raw());
						$self->user_prog()->code(";");
					$self->user_prog()->back();
				}
				$ri->dataset()->code_lock($self->user_prog()) if ($ri->dataset()->shared());
				$ri->dataset()->code_write($self->user_prog(),
						grep
						(
							$_->name() !~ /^_/, 
							$r->output_fields()->to_array(),
							$self->main()->input()->input_fields()->to_array()
						)
					);
				$ri->dataset()->code_unlock($self->user_prog()) if ($ri->dataset()->shared());

				$self->user_prog()->code('next;');
				$self->user_prog()->close_block();
				$join='elsif';
			}
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::DivertInputRecord; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'divert_input_record');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->divert_record_on_input()->size());
		$self->program_info();
		my $join='if';
		foreach ($self->main()->user_sections()->divert_record_on_input()->to_array()) {
			$self->user_prog()->newline_off();
			$self->user_prog()->code("$join (");
			$self->user_prog()->code($_->condition_exp_raw());
			$self->user_prog()->newline_on();
			$self->user_prog()->code(")");
			$self->user_prog()->open_block();
			$self->catalogue()->code_segments()->print_input()->new(
				main => $self->main(), user_prog => $self->user_prog(), dataset => $_->output_dataset())
				->prepare();
			$self->user_prog()->code('next;');
			$self->user_prog()->close_block();
			$join='elsif'; # maybe not
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::CopyInputRecord; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'copy_input_record');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->copy_record_on_input()->size());
		$self->program_info();
		foreach ($self->main()->user_sections()->copy_record_on_input()->to_array()) {
			if (defined($_->condition_exp())) {
				$self->user_prog()->newline_off();
				$self->user_prog()->code("if (");
				$self->user_prog()->code($_->condition_exp_raw());
				$self->user_prog()->newline_on();
				$self->user_prog()->code(")");
				$self->user_prog()->open_block();
			}
			$self->catalogue()->code_segments()->print_input()->new(
				main => $self->main(), user_prog => $self->user_prog(), dataset => $_->output_dataset())
				->prepare();
			$self->user_prog()->close_block() if (defined($_->condition_exp()));
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::PrintInput; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers qw( dataset ),
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'print_input');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->dataset()->code_lock($self->user_prog()) if ($self->dataset()->shared());
		$self->dataset()->code_write($self->user_prog(),
				grep
				(
					$_->name() !~ /^_/, 
					$self->main()->input()->input_fields()->to_array()
				)
			);
		$self->dataset()->code_unlock($self->user_prog()) if ($self->dataset()->shared());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::FieldProcessInput; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'field_process_input');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->field_process_on_input()->size());
		$self->program_info();
		foreach ($self->main()->field_process_on_input()->fields()->to_array())
		{
			$self->user_prog()->code("@{[ $_->input_field()->getvar() ]} @{[ defined($_->operator()) ? $_->operator() : '=' ]} # @{[ $_->calc_exp() ]}");
			$self->user_prog()->over();
				$self->user_prog()->code($_->calc_exp_raw());
				$self->user_prog()->code(";");
			$self->user_prog()->back();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::InputFilter; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'input_filter');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->filter_on_input()->size());
		$self->program_info();
		$self->user_prog()->code("next unless");
		$self->user_prog()->open_block("(");
		my $join = "";
		foreach ($self->main()->filter_on_input()->filter_expressions()->to_array()) {
			$self->user_prog()->newline_off();
			$self->user_prog()->code("$join");
			$self->user_prog()->newline_on();
			$self->user_prog()->code($_->exp_raw());
#> Configurable join operator
			$join = "|| ";
		}
		$self->user_prog()->close_block();
		$self->user_prog()->code(";");
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::DedupOnInput; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'dedup_on_input');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->dedup_on_input()->size());
		$self->program_info();
		$self->user_prog()->code("next if");
		$self->user_prog()->open_block("(");
		my $join = "";
		foreach ($self->main()->dedup_on_input()->fields()->to_array()) 
		{
			$self->user_prog()->code($join . "defined(\$previous_dedup@{[ $_->input_field()->tname() ]})");
			$self->user_prog()->code("&& \$previous_dedup@{[ $_->input_field()->tname() ]}"
				. ($_->pequel_type()->name() =~ /^(numeric|amount)$/ ? ' == ' : ' eq ')
				. $_->input_field()->getvar());
			$join = "&& ";
		}
		$self->user_prog()->close_block();
		$self->user_prog()->code(";");
		map
		(
			$self->user_prog()->code("\$previous_dedup@{[ $_->input_field()->tname() ]} = @{[ $_->input_field()->getvar() ]};"),
			$self->main()->dedup_on_input()->fields()->to_array()
		);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::InputRecordLimit; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'input_record_limit');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->user_prog()->code("last if (\$@{[ 
			$self->main()->input()->input_dataset()->count_vname() 
		]} == @{[ 
			$self->main()->config()->input_record_limit()+1 
		]});");
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::MainBody; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'main_body');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		my @calculated_output_flds;	#C3;
		my %conditional_output_flds; #C2;
		foreach ($self->main()->output()->output_fields()->to_array())
		{
			defined($_->calc_exp())
				? push(@calculated_output_flds, $_)
				: push(@{$conditional_output_flds{"@{[ defined($_->condition_exp()) ? $_->condition_exp() : '' ]}"}}, $_);
		}	
#?		$self->scope()->push_back ....
		foreach my $cond (sort keys %conditional_output_flds) # for each condition...
		{
			if ($cond ne '')
			{
				my $obj = ${$conditional_output_flds{$cond}}[0];
				$self->user_prog()->newline_off();
				$self->user_prog()->code("if (");
				$self->user_prog()->code($obj->condition_exp_raw());
				$self->user_prog()->newline_on();
				$self->user_prog()->code(")");
				$self->user_prog()->open_block();
#?				$self->scope()->push_back ....
			}
			map($self->code_output($_), @{$conditional_output_flds{$cond}}); # for each output-field inside condition...
#?			$self->scope()->pop_back ....if ($cond ne '');
			$self->user_prog()->close_block() if ($cond ne '');
		}
		map($self->code_output($_), @calculated_output_flds);
	}
	sub code_output
	{        	
		my $self = shift;
		my $output_field = shift;
		if (defined($output_field->aggregate_type()))
		{
			$output_field->aggregate_type()->code_output($self->user_prog());
		}
		elsif (!defined($output_field->calc_exp()))
		{
			$self->user_prog()->code("@{[ $output_field->getvar() ]} = @{[ $output_field->input_field()->getvar() ]};");
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::PrepareTransferFields; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'prepare_transfer_fields');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->user_prog()->comment(" ---> Input Fields (@{[ $self->main()->config()->transfer() ]})");
		foreach
		(
			grep
			(
				$_->input_field()->dataset() == $self->main()->input()->input_dataset(),
				$self->main()->output()->transfer_fields()->to_array()
			)
		)
		{
			if (defined($_->aggregate_type()))
			{
				$_->aggregate_type()->code_output($self->user_prog());
			}
			elsif (!defined($_->calc_exp()))
			{
#?				if ($self->main()->group_by_fields()->size() == 0 && $_->getvar() eq $_->input_field()->getvar() 
				$self->user_prog()->code("@{[ $_->getvar() ]} = @{[ $_->input_field()->getvar() ]}; # assigned");
			}
		}

		foreach my $j ($self->main()->user_joins()->to_array()) 
		{
			$self->user_prog()->comment(" ---> Input Join Fields (@{[ $j->input_dataset()->datasource()->fdname() ]})");
			$self->user_prog()->code(
				"if ("
				. $j->key_field_ref()->getvar() 
			    . ($j->key_type() =~ /^(numeric|amount)$/ ? ' == ' : ' eq ')
				. $j->join_field_ref()->getvar()
				. ")"
			);
			$self->user_prog()->open_block();
			foreach
			(
				grep
				(
					$_->input_field()->dataset() == $j->input_dataset(),
					$self->main()->output()->transfer_fields()->to_array()
				)
			)
			{
				if (defined($_->aggregate_type()))
				{
					$_->aggregate_type()->code_output($self->user_prog());
				}
				elsif (!defined($_->calc_exp()))
				{
					$self->user_prog()->code("@{[ $_->getvar() ]} = @{[ $_->input_field()->getvar() ]}; # assigned");
				}
			}
			$self->user_prog()->close_block();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::GroupingBreakControl; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'grouping_break_control');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		return unless ($self->main()->user_sections()->group_by()->size());
		# TODO:code_break_before()...

		foreach ($self->main()->group_by()->input_fields()->to_array()) {
			$self->user_prog()->newline_off();
			$self->user_prog()->code("\$key@{[ $_->tname() ]} = @{[ $_->input_field()->getvar() ]}");
			$self->user_prog()->code($_->pequel_type()->name() =~ /^(numeric|amount)$/ ? '+0' : '');
			$self->user_prog()->newline_on();
			$self->user_prog()->code(";");
		}
		
		if ($self->main()->config()->hash()) 
		{
			$self->user_prog()->code("\$_HKEY = "
				. join(" . '|' . ", map("\$key@{[ $_->tname() ]}", $self->main()->group_by()->input_fields()->to_array()))
				. ";"
			);
		}
		else
		{
			$self->user_prog()->code("if ("
				. join(" || ", map("!defined(\$previous_key@{[ $_->tname() ]})", $self->main()->group_by()->input_fields()->to_array()))
				. ")"
			);
			$self->user_prog()->open_block();
				map
				(
					$self->user_prog()->code("\$previous_key@{[ $_->tname() ]} = \$key@{[ $_->tname() ]};"),
					$self->main()->group_by()->input_fields()->to_array()
				);
			$self->user_prog()->close_block();

			$self->user_prog()->code("elsif ("
				. join
					(
						" || ", 
						map
						(
							"\$previous_key@{[ $_->tname() ]} @{[ $_->pequel_type()->name() =~ /^(numeric|amount)$/ ? '!=' : 'ne' ]} \$key@{[ $_->tname() ]}",
							$self->main()->group_by()->input_fields()->to_array()
						)
					)
				. ")"
			);
			$self->user_prog()->open_block();
#?				$self->code_zero_nulls if ($self->properties()->nonulls());
				$self->catalogue()->code_segments()->output()->new(
					main => $self->main(), user_prog => $self->user_prog())->prepare();
				$self->catalogue()->code_segments()->reset()->new(
					main => $self->main(), user_prog => $self->user_prog())->prepare();
			$self->user_prog()->close_block();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Reset; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'reset');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		map
		(
			$self->user_prog()->code("\$previous_key@{[ $_->tname() ]} = \$key@{[ $_->tname() ]};"),
			$self->main()->group_by()->input_fields()->to_array()
		) 
		if ($self->main()->user_sections()->group_by()->size());

		map
		(
			$self->user_prog()->code("\$previous_dedup@{[ $_->input_field()->tname() ]} = undef;"),
			$self->main()->dedup_on_input()->fields()->to_array()
		)
		if ($self->main()->user_sections()->dedup_on_input()->size());

		map
		(
			$self->user_prog()->code("\$previous_dedup@{[ $_->output_field()->tname() ]} = undef;"),
			$self->main()->dedup_on_output()->fields()->to_array()
		)
		if ($self->main()->user_sections()->dedup_on_output()->size());

		foreach 
		(
			grep
			(
				defined($_),
				(
					map($_->aggregate_type(), 
						grep(defined($_->aggregate_type()), $self->main()->output()->output_fields()->to_array())),
					$self->main()->output()->output_dataset(),
					$self->main()->output()->transfer_dataset(),
					map($_->transfer_dataset(), $self->main()->user_joins()->to_array()),
				)
			)
		)
		{
			$_->code_reset($self->user_prog());
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Output; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'output');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;

		$self->segments()->push_back($self->catalogue()->code_segments()->hash_output()->new(@_))
			if ($self->main()->config()->hash());

		$self->segments()->push_back($self->catalogue()->code_segments()->output_final()->new(@_))
			if ($self->main()->output()->output_fields()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->output_repeater()->new(@_))
			if ($self->main()->user_sections()->repeat_record_on_output()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->output_filter()->new(@_))
			if ($self->main()->user_sections()->filter_on_output()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->dedup_on_output()->new(@_))
			if ($self->main()->user_sections()->dedup_on_output()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->field_process_output()->new(@_))
			if ($self->main()->user_sections()->field_process_on_output()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->reject_output_record()->new(@_))
			if ($self->main()->user_sections()->reject_record_on_output()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->divert_output_record()->new(@_))
			if ($self->main()->user_sections()->divert_record_on_output()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->copy_output_record()->new(@_))
			if ($self->main()->user_sections()->copy_record_on_output()->size());

		$self->segments()->push_back($self->catalogue()->code_segments()->print_output()->new(
				@_, dataset => $self->main()->output()->output_dataset()));
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::HashOutput; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'hash_output');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->group_by()->size());
		return unless ($self->main()->config()->hash());
		$self->program_info();
		my $numeric = do {
			my $n=1; 
			foreach ($self->main()->group_by()->input_fields()->to_array()) 
			{ $n=0 unless ($_->pequel_type()->name() =~ /^(numeric|amount)$/); }
			$n;
		};
		# In case there are no output fields but has transfer fields, then get key list from latter...
		$self->user_prog()->code("foreach \$_HKEY (sort@{[ 
			$numeric ? ' {$a <=> $b} ' : ' ' 
		]}keys \%@{[ 
			$self->main()->output()->transfer_fields()->size() 
				? $self->main()->output()->transfer_fields()->front()->dataset()->arr_vname() 
				: $self->main()->output()->output_dataset()->arr_vname() 
		]})");
		$self->user_prog()->open_block();
	}
	sub prepare_finish
	{
		my $self = shift;
		return unless ($self->main()->config()->hash());
		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::OutputFinal; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'output_final');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
#> TODO:Exclude any oflds that reference &orval()'s

#?		if ($self->main()->configuration()->no_nulls()) {
#?			foreach ($self->main()->output()->transfer_fields()->to_array()) {
#?				$self->user_prog()->code("@{[ $_->getvar() ]} = '' unless (defined(@{[ $_->getvar() ]}));");
#?			}
#?		}
		foreach ($self->main()->output()->output_fields()->to_array())
		{
			if (defined($_->calc_exp())) 
			{	
				next if ($_->calc_exp() =~/\&orval\b/); # TODO:need to go deeper into reference list
				$self->user_prog()->code("@{[ $_->getvar() ]} = # @{[ $_->calc_exp() ]}");
				$self->user_prog()->over();
					$self->user_prog()->code($_->calc_exp_raw());
					$self->user_prog()->code(";");
				$self->user_prog()->back();
			}
			elsif (defined($_->aggregate_type()))
			{
				$_->aggregate_type()->code_output_final($self->user_prog());
			}
			next unless ($self->main()->config()->no_nulls());
			$self->user_prog()->code("@{[ $_->getvar() ]} = @{[ $_->pequel_type()->name =~ /^(numeric|amount)$/ ? '0' : q/''/ ]}");
			$self->user_prog()->over();
				$self->user_prog()->code("if");
				$self->user_prog()->open_block("(");
					$self->user_prog()->code("!defined(@{[ $_->getvar() ]})");
					$self->user_prog()->code("|| @{[ $_->getvar() ]} eq ''");
				$self->user_prog()->close_block();
				$self->user_prog()->code(";# no_nulls");
			$self->user_prog()->back();
		}
		foreach (grep($_->name() !~ /^_/, $self->main()->output()->transfer_fields()->to_array())) {
			$self->user_prog()->code("@{[ $_->getvar() ]} = '' unless (defined(@{[ $_->getvar() ]}));");
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::OutputRepeater; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'output_repeater');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->repeat_record_on_output()->size());
		$self->program_info();
		my $r=0;
		foreach ($self->main()->repeat_record_on_output()->fields()->to_array()) {
			$self->user_prog()->code("foreach \$_ORVAL_@{[ ++$r ]} (@{[ $_->exp_raw() ]})");
			$self->user_prog()->code("# Output repeater $r value (@{[ $_->exp() ]})");
			$self->user_prog()->open_block();
		}
		#> Need to output final any oflds that reference &orval()
		#> ensure that repeat-record exp does not reference field where calc_exp references &orval()
		foreach ($self->main()->output()->output_fields()->to_array())
		{
			next unless (defined($_->calc_exp()) && $_->calc_exp() =~/\&orval\b/); # TODO:need to go deeper into reference list
			$self->user_prog()->code("@{[ $_->getvar() ]} = # @{[ $_->calc_exp() ]}");
			$self->user_prog()->over();
			$self->user_prog()->code($_->calc_exp_raw());
			$self->user_prog()->code(";");
			$self->user_prog()->back();
			if ($_->pequel_type()->name() =~ /(numeric|amount)/ && $_->name() !~ /^_/ 
				&& $self->main()->config()->no_nulls()) 
			{
				$self->user_prog()->code("@{[ $_->getvar() ]} = 0");
				$self->user_prog()->code("if");
				$self->user_prog()->open_block("(");
					$self->user_prog()->code("!defined(@{[ $_->getvar() ]})");
					$self->user_prog()->code("|| @{[ $_->getvar() ]} eq ''");
				$self->user_prog()->close_block();
				$self->user_prog()->code("; # no_nulls");
			}
		}
	}
	sub prepare_finish
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->repeat_record_on_output()->size());
		foreach ($self->main()->repeat_record_on_output()->fields()->to_array()) {
			$self->user_prog()->close_block();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::OutputFilter; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'output_filter');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->filter_on_output()->size());
		$self->program_info();
		$self->user_prog()->code("if");
		$self->user_prog()->open_block("(");
			my $join = "";
			foreach ($self->main()->filter_on_output()->filter_expressions()->to_array()) {
				$self->user_prog()->newline_off();
				$self->user_prog()->code("$join");
				$self->user_prog()->newline_on();
				$self->user_prog()->code($_->exp_raw());
#> Configurable join operator
				$join = "&& ";
			}
		$self->user_prog()->close_block();
		$self->user_prog()->open_block();
	}
	sub prepare_finish
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->filter_on_output()->size());
		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::DedupOnOutput; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'dedup_on_output');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->dedup_on_output()->size());
		$self->program_info();
		$self->user_prog()->code("if");
		$self->user_prog()->open_block("(");
			my $join = "";
			foreach ($self->main()->dedup_on_output()->fields()->to_array()) {
				$self->user_prog()->code($join) unless ($join eq '');
				$self->user_prog()->open_block("(");
					$self->user_prog()->code("!defined(\$previous_dedup@{[ $_->output_field()->tname() ]})");
					$self->user_prog()->code("|| \$previous_dedup@{[ $_->output_field()->tname() ]}"
						. ($_->output_field()->pequel_type()->name() =~ /^(numeric|amount)$/ ? ' != ' : ' ne ')
						. $_->output_field()->getvar());
				$self->user_prog()->close_block();
				$join = "&& ";
			}
		$self->user_prog()->close_block();
		$self->user_prog()->open_block();
		map
		(
			$self->user_prog()->code("\$previous_dedup@{[ $_->output_field()->tname() ]} = @{[ $_->output_field()->getvar() ]};"),
			$self->main()->dedup_on_output()->fields()->to_array()
		);
	}
	sub prepare_finish
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->dedup_on_output()->size());
		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::FieldProcessOutput; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'field_process_output');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->field_process_on_output()->size());
		$self->program_info();
		foreach ($self->main()->field_process_on_output()->fields()->to_array())
		{
			$self->user_prog()->code("@{[ $_->output_field()->getvar() ]} @{[ 
				defined($_->operator()) ? $_->operator() : '=' ]} # @{[ $_->calc_exp() ]}");
			$self->user_prog()->over();
				$self->user_prog()->code($_->calc_exp_raw());
				$self->user_prog()->code(";");
			$self->user_prog()->back();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::RejectOutputRecord; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'reject_output_record');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->reject_record_on_output()->size());
		$self->program_info();
		my $join='if';
		foreach my $ri ($self->main()->user_sections()->reject_record_on_output()->to_array()) {
			foreach my $r ($ri->reason_codes_list()->to_array()) {
				$self->user_prog()->comment(" +++ Reason Code (@{[ $r->reason_code() ]}) +++");
				$self->user_prog()->newline_off();
				$self->user_prog()->code("$join (");
				$self->user_prog()->code($r->condition_exp_raw());
				$self->user_prog()->newline_on();
				$self->user_prog()->code(")");
				$self->user_prog()->open_block();

				foreach (grep(defined($_->calc_exp()), $r->output_fields()->to_array())) {
					$self->user_prog()->code("@{[ $_->getvar() ]} = # @{[ $_->calc_exp() ]}");
					$self->user_prog()->over();
						$self->user_prog()->code($_->calc_exp_raw());
						$self->user_prog()->code(";");
					$self->user_prog()->back();
				}
				$ri->dataset()->code_lock($self->user_prog()) if ($ri->dataset()->shared());
				$ri->dataset()->code_write($self->user_prog(),
						grep
						(
							$_->name() !~ /^_/, 
							$r->output_fields()->to_array(),
							$self->main()->output()->transfer_fields()->to_array(), 
							$self->main()->output()->output_fields()->to_array()
						)
					);
				$ri->dataset()->code_unlock($self->user_prog()) if ($ri->dataset()->shared());

				$self->user_prog()->close_block();
				$join='elsif';
			}
		}
		$self->user_prog()->code("else");
		$self->user_prog()->open_block();
	}
	sub prepare_finish
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->reject_record_on_output()->size());
		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::DivertOutputRecord; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'divert_output_record');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->divert_record_on_output()->size());
		$self->program_info();
		my $join='if';
		foreach ($self->main()->user_sections()->divert_record_on_output()->to_array()) {
			$self->user_prog()->newline_off();
			$self->user_prog()->code("$join (");
			$self->user_prog()->code($_->condition_exp_raw());
			$self->user_prog()->newline_on();
			$self->user_prog()->code(")");
			$self->user_prog()->open_block();
			$self->catalogue()->code_segments()->print_output()->new(
				main => $self->main(), user_prog => $self->user_prog(), dataset => $_->output_dataset())
				->prepare_start();
			$self->user_prog()->close_block();
			$join='elsif';
		}
		$self->user_prog()->code("else");
		$self->user_prog()->open_block();
	}
	sub prepare_finish
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->divert_record_on_output()->size());
		$self->user_prog()->close_block();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::CopyOutputRecord; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'copy_output_record');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return unless ($self->main()->user_sections()->copy_record_on_output()->size());
		$self->program_info();
		foreach ($self->main()->user_sections()->copy_record_on_output()->to_array()) {
			if (defined($_->condition_exp())) {
				$self->user_prog()->newline_off();
				$self->user_prog()->code("if (");
				$self->user_prog()->code($_->condition_exp_raw());
				$self->user_prog()->newline_on();
				$self->user_prog()->code(")");
				$self->user_prog()->open_block();
			}
			$self->catalogue()->code_segments()->print_output()->new(
				main => $self->main(), user_prog => $self->user_prog(), dataset => $_->output_dataset())
				->prepare_start();
			$self->user_prog()->close_block() if (defined($_->condition_exp()));
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::PrintOutput; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers qw( dataset ),
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'print_output');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		$self->dataset()->code_lock($self->user_prog()) if ($self->dataset()->shared());
		$self->dataset()->code_write($self->user_prog(),
				grep
				(
					$_->name() !~ /^_/, 
					$self->main()->output()->transfer_fields()->to_array(), 
					$self->main()->output()->output_fields()->to_array()
				)
			);
		$self->dataset()->code_unlock($self->user_prog()) if ($self->dataset()->shared());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Finish; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'finish');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_finish
	{
		my $self = shift;
		$self->program_info();
		$self->catalogue()->code_segments()->verbose_msg()->new(
			target_name => $self->main()->pequel_name(),
			user_prog => $self->user_prog(), 
			msg_text => "\$@{[ $self->main()->input()->input_dataset()->count_vname() ]} records."
		)->prepare() if ($self->main()->config()->verbose()); 
		if ($self->main()->config()->verbose() && $self->main()->config()->statistics()) 
		{
			$self->user_prog()->code("my \$benchmark_end = new Benchmark;");
			$self->user_prog()->code("my \$benchmark_timediff = timediff(\$benchmark_start, \$benchmark_end);");
			$self->catalogue()->code_segments()->verbose_msg()->new(
				target_name => $self->main()->pequel_name(),
				user_prog => $self->user_prog(), 
				msg_text => "Code statistics: \@{[timestr(\$benchmark_timediff)]}"
			)->prepare();
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::LoadTable; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'load_table');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		$self->program_info();
		foreach ($self->main()->user_tables()->to_array()) {
			my $f = "@{[ $_->package_name() ]}::@{[ $_->load_function_name() ]}";
			undef(&{$f}); # TODO: probably does nothing!
		}
		map
		(
			$_->code_load_table($self->user_prog()), 
			$self->main()->user_tables()->to_array()
		);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::VerboseMsg; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers qw( msg_text cond target_name ),
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'verbose_msg');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		my $self = shift;
		return if ($self->properties()->silent() != 0);
		my $msg_text = shift || $self->msg_text();
		my $cond = shift || $self->cond();
		$self->user_prog()->code("print STDERR '[@{[ $self->target_name() ]} ' . localtime()");
		$self->user_prog()->over();
			$self->user_prog()->code(". \"] $msg_text\\n\"");
			$self->user_prog()->code("$cond") if (defined($cond));
			$self->user_prog()->code(";");
		$self->user_prog()->back();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::UserOptions; 
	use base qw(ETL::Pequel3::Generator::Segment::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'segment_name', default => 'user_options');
	use Class::STL::ClassMembers::Constructor; 
	sub prepare_start
	{
		use ETL::Pequel3::GetOpt;
		my $self = shift;
		my $user = ETL::Pequel3::GetOpt->new()->user();
		$self->user_prog()->divider();
		$self->program_info();
		$self->user_prog()->code("our \%g_USER_OPTION = #long_name | short_name => [ default, flag, description, alt_name ]");
		$self->user_prog()->open_block("(");
		my $comment='';
		foreach 
		(
			grep($_->user_option(), $self->catalogue()->options()->to_array())
			#NOTE: User Command Option Type
		)	
		{
			$self->user_prog()->code("@{[ $_->long_name() ]} =>");
			$self->user_prog()->open_block("[");
			if ($user->exists($_->long_name())) { # get any opt that were passed in via Command->main():
				$self->user_prog()->code("'@{[ $user->exists($_->long_name())->data() ]}', ");
				$comment = 'command option long-name';
			} 
			elsif (defined($_->short_name()) && $user->exists($_->short_name())) {
				$self->user_prog()->code("'@{[ $user->exists($_->short_name())->data() ]}', ");
				$comment = 'command option short-name';
			} 
			else {
				if 
				(
					$self->main()->config()->exists($_->long_name())
					&& defined($self->main()->config()->exists($_->long_name())->value())
				) 
				{
					$self->user_prog()->code("'@{[ 
						$self->main()->config()->exists($_->long_name())->value() ]}', ");
					$comment = 'configuration long-name';
				}
				elsif 
				(
					defined($_->short_name()) && $self->main()->config()->exists($_->short_name())
						&& defined($self->main()->config()->exists($_->short_name())->value())
				)
				{
					$self->user_prog()->code("'@{[ 
							$self->main()->config()->exists($_->short_name())->value() ]}', ");
					$comment = 'configuration short-name';
				}
				elsif (defined($_->default()))
				{
					$self->user_prog()->code("'@{[ $_->default() ]}', ");
					$comment = 'command option default';
				}
				else
				{
					$self->user_prog()->code("undef, ");
					$comment = 'undefined';
				}
			}
			$self->user_prog()->code("@{[ $_->flag() ]}, ");
			$self->user_prog()->code("'@{[ defined($_->description()) ? $_->description() : q## ]}', ");
			$self->user_prog()->code("'@{[ defined($_->short_name()) ? $_->short_name() : q## ]}', ");
			$self->user_prog()->close_block(); # ] #
			$self->user_prog()->code(",");
			
			next unless (defined($_->short_name()));
			$comment='';
			$self->user_prog()->code("@{[ $_->short_name() ]} =>");
			$self->user_prog()->open_block("[");
			if ($user->exists($_->short_name())) { # get any opt that were passed in via Command->main():
				$self->user_prog()->code("'@{[ $user->exists($_->short_name())->data() ]}', ");
				$comment = 'command option long-name(2)';
			} 
			elsif ($user->exists($_->long_name())) {
				$self->user_prog()->code("'@{[ $user->exists($_->long_name())->data() ]}', ");
				$comment = 'command option short-name(2)';
			} 
			else {
				if 
				(
					$self->main()->config()->exists($_->long_name())
					&& defined($self->main()->config()->exists($_->long_name())->value())
				) 
				{
					$self->user_prog()->code("'@{[ 
						$self->main()->config()->exists($_->long_name())->value() ]}', ");
					$comment = 'configuration long-name(2)';
				}
				elsif 
				(
					defined($_->short_name()) && $self->main()->config()->exists($_->short_name())
						&& defined($self->main()->config()->exists($_->short_name())->value())
				)
				{
					$self->user_prog()->code("'@{[ 
							$self->main()->config()->exists($_->short_name())->value() ]}', ");
					$comment = 'configuration short-name(2)';
				}
				elsif (defined($_->default()))
				{
					$self->user_prog()->code("'@{[ $_->default() ]}', ");
					$comment = 'command option default(2)';
				}
				else
				{
					$self->user_prog()->code("undef, ");
					$comment = 'undefined(2)';
				}
			}
			$self->user_prog()->code("@{[ $_->flag() ]}, ");
			$self->user_prog()->code("'@{[ defined($_->description()) ? $_->description() : q## ]}', ");
			$self->user_prog()->code("'@{[ $_->long_name() ]}'");
			$self->user_prog()->close_block(); # ] #
			$self->user_prog()->code(",");
		}
		foreach 
		(
			map
			(
				$_->items()->to_array(), 
				$self->main()->user_sections()->user_option()->to_array()
				# NOTE: User Defined Options
			)
		)	
		{
			$self->user_prog()->code("@{[ $_->attributes()->long_name() ]} =>");
			$self->user_prog()->open_block("[");
			if ($user->exists($_->attributes()->long_name())) { # get any opt that were passed in via Command->main():
				$self->user_prog()->code("'@{[ $user->exists($_->attributes()->long_name())->data() ]}', ");
			} 
			elsif (defined($_->attributes()->short_name()) && $user->exists($_->attributes()->short_name())) {
				$self->user_prog()->code("'@{[ $user->exists($_->attributes()->short_name())->data() ]}', ");
			} 
			else {
				defined($_->attributes()->default())
					? $self->user_prog()->code("'@{[ $_->attributes()->default() ]}', ")
					: $self->user_prog()->code("undef, ");
			}
			$self->user_prog()->code("@{[ $_->attributes()->flag() ]}, ");
			$self->user_prog()->code("'@{[ 
				defined($_->attributes()->description()) ? $_->attributes()->description() : q## ]}', ");
			$self->user_prog()->code("'@{[ $_->attributes()->short_name() ]}'");
			$self->user_prog()->close_block(); # ] #
			$self->user_prog()->code(",");

			next unless (defined($_->attributes()->short_name()));
			$self->user_prog()->code("@{[ $_->attributes()->short_name() ]} =>");
			$self->user_prog()->open_block("[");
			if ($user->exists($_->attributes()->short_name())) { # get any opt that were passed in via Command->main():
				$self->user_prog()->code("'@{[ $user->exists($_->attributes()->short_name())->data() ]}', ");
			} 
			elsif ($user->exists($_->attributes()->long_name())) {
				$self->user_prog()->code("'@{[ $user->exists($_->attributes()->long_name())->data() ]}', ");
			} 
			else {
				defined($_->attributes()->default())
					? $self->user_prog()->code("'@{[ $_->attributes()->default() ]}', ")
					: $self->user_prog()->code("undef, ");
			}
			$self->user_prog()->code("@{[ $_->attributes()->flag() ]}, ");
			$self->user_prog()->code("'@{[ 
				defined($_->attributes()->description()) ? $_->attributes()->description() : q## ]}', ");
			$self->user_prog()->code("'@{[ $_->attributes()->long_name() ]}'");
			$self->user_prog()->close_block(); # ] #
			$self->user_prog()->code(",");
		}
		$self->user_prog()->close_block();
		$self->user_prog()->code(";");
		$self->user_prog()->comment("");
		$self->user_prog()->comment("Check environement for any PEQUEL_ variables and copy in:");
		$self->user_prog()->comment("");
		$self->user_prog()->code('for (grep(exists $ENV{"PEQUEL_@{[ uc($_) ]}"}, keys %g_USER_OPTION))');
		$self->user_prog()->open_block();
			$self->user_prog()->code('${$g_USER_OPTION{$_}}[0] = $ENV{"PEQUEL_@{[ uc($_) ]}"};');
		$self->user_prog()->close_block();
		$self->user_prog()->comment("");
		$self->user_prog()->comment('Check command-line arguments:');
		$self->user_prog()->comment("");
		$self->user_prog()->code('for (my $i=0; $i <= $#ARGV; ++$i)');
		$self->user_prog()->open_block();
			$self->user_prog()->code('if ($ARGV[$i] =~ /^\-{1,2}(.*)/ && exists $g_USER_OPTION{$1})');
			$self->user_prog()->open_block();
				$self->user_prog()->code('if ($1 =~ /^(help|usage)$/)');
				$self->user_prog()->open_block();
					$self->user_prog()->code(q#print STDERR "$0 Usage:\n";#);
					$self->user_prog()->code(q#foreach (keys %g_USER_OPTION)#);
					$self->user_prog()->open_block();
						$self->user_prog()->code(q#print STDERR sprintf("%-30s -- %s",#);
						$self->user_prog()->over();
							$self->user_prog()->code(q#"-@{[ $_ ]}@{[ defined(${$g_USER_OPTION{$_}}[3])#);
							$self->user_prog()->over();
								$self->user_prog()->code(q#? (' | -' . ${$g_USER_OPTION{$_}}[3]) : '' ]}",#);
							$self->user_prog()->back();
							$self->user_prog()->code(q#"@{[ ${$g_USER_OPTION{$_}}[2] ]}\n");#);
						$self->user_prog()->back();
					$self->user_prog()->close_block();
					$self->user_prog()->code(q#exit(0);#);
				$self->user_prog()->close_block();

				$self->user_prog()->code('${$g_USER_OPTION{$i}}[0] = ${$g_USER_OPTION{$1}}[1] == 1 ');
				$self->user_prog()->over();
					$self->user_prog()->code('? 1 : $i < $#ARGV && $ARGV[$i+1] !~ /^-/ ');
					$self->user_prog()->over();
						$self->user_prog()->code('? $ARGV[++$i] : ${$g_USER_OPTION{$1}}[0];');
					$self->user_prog()->back();
				$self->user_prog()->back();
			$self->user_prog()->close_block();
		$self->user_prog()->close_block();

		#TODO: move these to the appropriate command option's code_init_global
		$self->user_prog()->code('if (defined(${$g_USER_OPTION{diag}}[0]) && ${$g_USER_OPTION{diag}}[0] != 0)');
		$self->user_prog()->open_block();
			$self->user_prog()->code('foreach (grep(defined(${$g_USER_OPTION{$_}}[0]), keys %g_USER_OPTION))');
			$self->user_prog()->open_block();
				$self->user_prog()->code('print STDERR "option->$_:",');
				$self->user_prog()->over();
					$self->user_prog()->code('${$g_USER_OPTION{$_}}[0], "--",');
					$self->user_prog()->code('${$g_USER_OPTION{$_}}[2], "\n";');
				$self->user_prog()->back();
			$self->user_prog()->close_block();
		$self->user_prog()->close_block();
		$self->user_prog()->divider();
	}
}
# ----------------------------------------------------------------------------------------------------
# SEGMENT TYPE CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Generator::Segment::Catalogue;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'code_segments'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'segment_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Generator::Segment::Abstract');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Generator::Segment::Startup->_new(),
			ETL::Pequel3::Generator::Segment::UserOptions->_new(),
			ETL::Pequel3::Generator::Segment::CallExecute->_new(),
			ETL::Pequel3::Generator::Segment::CallLoadTable->_new(),
			ETL::Pequel3::Generator::Segment::MainPackage->_new(),
			ETL::Pequel3::Generator::Segment::LoadTable->_new(), 
			ETL::Pequel3::Generator::Segment::SubPrograms->_new(),
			ETL::Pequel3::Generator::Segment::SubExecuteBody->_new(),
			ETL::Pequel3::Generator::Segment::SubExecuteDef->_new(),
			ETL::Pequel3::Generator::Segment::Package->_new(),
			ETL::Pequel3::Generator::Segment::HeaderInfo->_new(),
			ETL::Pequel3::Generator::Segment::Declaration->_new(),
			ETL::Pequel3::Generator::Segment::InitGlobal->_new(),
			ETL::Pequel3::Generator::Segment::InitLocal->_new(),
			ETL::Pequel3::Generator::Segment::OpenGlobal->_new(),
			ETL::Pequel3::Generator::Segment::OpenLocal->_new(),
			ETL::Pequel3::Generator::Segment::CloseGlobal->_new(),
			ETL::Pequel3::Generator::Segment::Main->_new(),
			ETL::Pequel3::Generator::Segment::MainWhileLoop->_new(),
			ETL::Pequel3::Generator::Segment::InputVerify->_new(),
			ETL::Pequel3::Generator::Segment::InputJoin->_new(),
			ETL::Pequel3::Generator::Segment::InputRecordCounter->_new(),
			ETL::Pequel3::Generator::Segment::InputRepeater->_new(),
			ETL::Pequel3::Generator::Segment::PrintOutputHeader->_new(),
			ETL::Pequel3::Generator::Segment::PrintInputHeader->_new(),
			ETL::Pequel3::Generator::Segment::DerivedInputFields->_new(),
			ETL::Pequel3::Generator::Segment::RejectInputRecord->_new(),
			ETL::Pequel3::Generator::Segment::DivertInputRecord->_new(),
			ETL::Pequel3::Generator::Segment::CopyInputRecord->_new(),
			ETL::Pequel3::Generator::Segment::PrintInput->_new(),
			ETL::Pequel3::Generator::Segment::FieldProcessInput->_new(),
			ETL::Pequel3::Generator::Segment::InputFilter->_new(),
			ETL::Pequel3::Generator::Segment::DedupOnInput->_new(),
			ETL::Pequel3::Generator::Segment::InputRecordLimit->_new(),
			ETL::Pequel3::Generator::Segment::MainBody->_new(),
			ETL::Pequel3::Generator::Segment::PrepareTransferFields->_new(),
			ETL::Pequel3::Generator::Segment::GroupingBreakControl->_new(),
			ETL::Pequel3::Generator::Segment::Reset->_new(),
			ETL::Pequel3::Generator::Segment::Output->_new(),
			ETL::Pequel3::Generator::Segment::HashOutput->_new(),
			ETL::Pequel3::Generator::Segment::OutputFinal->_new(),
			ETL::Pequel3::Generator::Segment::OutputRepeater->_new(),
#<			ETL::Pequel3::Generator::Segment::OutputFinalAfterRepeat->_new(),
			ETL::Pequel3::Generator::Segment::OutputFilter->_new(),
			ETL::Pequel3::Generator::Segment::DedupOnOutput->_new(),
			ETL::Pequel3::Generator::Segment::FieldProcessOutput->_new(),
			ETL::Pequel3::Generator::Segment::RejectOutputRecord->_new(),
			ETL::Pequel3::Generator::Segment::DivertOutputRecord->_new(),
			ETL::Pequel3::Generator::Segment::CopyOutputRecord->_new(),
			ETL::Pequel3::Generator::Segment::PrintOutput->_new(),
			ETL::Pequel3::Generator::Segment::Finish->_new(),
			ETL::Pequel3::Generator::Segment::VerboseMsg->_new(),
			ETL::Pequel3::Generator::Segment::Divider->_new(),
			ETL::Pequel3::Generator::Segment::EndOfProgram->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
1;
