# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::Aggregates.pm
#  Created	: 8 May 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::Aggregates;
require 5.005_62;
use strict;
use warnings;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Abstract;
	use base qw(Class::STL::Element);
	use Class::STL::ClassMembers 
		qw( 
			name 
			err
			properties 
			pequel_types
			hash 
			configuration
			use_list
			output_field
			requires_input_field
			pequel_ref
			xref
			description
		),
		Class::STL::ClassMembers::DataMember->new(name => 'print_info', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'xs_type', default => 'NMTOKEN'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'string'),
		Class::STL::ClassMembers::DataMember->new(name => 'require_input_field', default => 1);
	use Class::STL::ClassMembers::Constructor;
	use Class::CodeStyler;
	sub new_extra
	{
		my $self = shift;
		$self->err(ETL::Pequel3::Error->new());
		$self->configuration($self->pequel_ref()->config()) if (defined($self->pequel_ref()));
		use ETL::Pequel3::Type::Properties;
		$self->properties(ETL::Pequel3::Type::Properties->new());
		use ETL::Pequel3::Type::PequelTypes;
		$self->pequel_types(ETL::Pequel3::Type::PequelTypes::Catalogue->new());
		$self->err()->user_error(10900, "Aggregate '@{[ 
			$self->name() ]}' -- required attribute 'output_field' is undefined!")
			if (!defined($self->output_field()));
		$self->err()->user_error(10900, "Aggregate '@{[ 
				$self->name() 
			]}' -- required attribute 'input_field' is undefined for output field '@{[ 
				$self->output_field()->name() 
			]}'!")
			if ($self->require_input_field() && !defined($self->output_field()->input_field()));
		$self->hash(defined($self->configuration()) && $self->configuration()->hash() == 1 ? '{$_HKEY}' : '');
#TODO	$self->hash($self->configuration()->hash() == 1 ? '{$@{[ $self->properties()->hkey_vname() ]}}' : '');
		use ETL::Pequel3::Type::Dates;
		use ETL::Pequel3::Type::FieldSet;
		$self->use_list(ETL::Pequel3::Type::FieldSet::OutputFields->new());
		$self->data($self->name());
		use ETL::Pequel3::CrossRef;
		$self->xref(ETL::Pequel3::CrossRef->new(
			target_obj => $self, 
			target_name => $self->name(), 
			pequel_ref => $self->pequel_ref(),
			@_,
		));
		return $self;
	}
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || Class::CodeStyler::Program::Perl->new();
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || Class::CodeStyler::Program::Perl->new();
		return $c;
	}
    sub code_output 
	{ 
		my $self = shift; 
		my $c = shift || Class::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub prinfo
	{
		my $self = shift; 
		return $self->print_info() ? "# ~@{[ $self->name() ]}(@{[ $self->output_field()->name() ]})" : '';
	}
	sub xml_schema
	{
		my $self = shift;
		my $xsd = shift;
		my $xml_node = shift || $xsd->schema_root();
		return if ($xsd->exists('element', $self->name(), $xml_node));
		$xsd->comment($xml_node, "Aggregate Type: @{[ $self->name() ]}"); 
		my $element = $xsd->xs_element($xml_node, 'name' => $self->name());
		if ($xsd->configuration()->xmlschema_annotation() && defined($self->description()) 
			&& $self->description() ne '') 
		{
			my $doc = $xsd->xs_documentation($xsd->xs_annotation($element), 'xml:lang' => 'en');
			$xsd->text($doc, $self->description());
		}
		my $valid_types = $xsd->xs_element($xsd->xs_all($xsd->xs_complexType($element)),
			'name', 'valid_types'
		);
		my $choice = $xsd->xs_choice($xsd->xs_complexType($valid_types));
		foreach (split(/[|]/, $self->valid_types())) {
			$xsd->xs_element($choice, 'name' => $_);
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Sum;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'sum'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Accumulate the total for all values in the group.'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date');
	use Class::STL::ClassMembers::Constructor;
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("@{[ $self->output_field()->getvar() ]} += $if");
		$c->over();
		$c->code("unless");
		$c->open_block("(");
			$c->code("!defined($if)");
			$c->code("|| $if eq ''");
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}"); #??? if ($self->properties()->no_nulls() == 1);
		$c->back();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::SumDistinct;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'sum_distinct'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Accumulate the total for unique values for field-name in the group.'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("my \%SUM_DISTINCT_KEYS_@{[ $self->output_field()->name() ]}; @{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%SUM_DISTINCT_KEYS_@{[ $self->output_field()->name() ]} = (); @{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("@{[ $self->output_field()->getvar() ]} += $if");
		$c->over();
		$c->code("if");
		$c->open_block("(");
			$c->code("$if ne ''");
			$c->code("&& ++\$SUM_DISTINCT_KEYS_$of\{qq{$if}} == 1");
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::ValuesUniq;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'values_uniq'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the list of unique values in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%VALUES_UNIQ_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%VALUES_UNIQ_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $delim = $self->properties()->default_list_delimiter();
		$c->code("@{[ $self->output_field()->getvar() ]} = ");
		$c->over();
			$c->code("join");
			$c->open_block("(");
				$c->code("qq{$delim},");
				$c->newline_off();
				$c->code("grep(length, sort keys ");
				$self->hash() eq ''
					? $c->code("\%VALUES_UNIQ_$of")
					: $c->code("\%{\$VALUES_UNIQ_$of}"); # hash
				$c->newline_on();
				$c->code(")"); #grep
			$c->close_block();
			$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		my $delim = $self->properties()->default_list_delimiter();
		if ($self->output_field()->input_field()->pequel_type()->name() eq 'array')
		{
			$c->code("map");
			$c->open_block("(");
				$c->code("\$VALUES_UNIQ_$of\{\$_}++,");
				$c->code("split(/\\s*$delim\\s*/, $if)");
			$c->close_block();
			$c->code(";@{[ $self->prinfo() ]}");
		}
		else
		{
			$c->code("\$VALUES_UNIQ_$of\{qq{$if}}++;@{[ $self->prinfo() ]}");
		}
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::ValuesAll;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'values_all'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the list of all values in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my @{[ $self->hash() eq '' ? '@' : '%' ]}VALUES_ALL_@{[ 
			$self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("@{[ $self->hash() eq '' ? '@' : '%' ]}VALUES_ALL_@{[ 
			$self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $delim = $self->properties()->default_list_delimiter();
		$c->code("@{[ $self->output_field()->getvar() ]} = ");
		$c->over();
			$c->code("join");
			$c->open_block("(");
				$c->code("qq{$delim},");
				$c->newline_off();
				$c->code("grep(length, ");
				$self->hash() eq ''
					? $c->code("\@VALUES_ALL_$of")
					: $c->code("\@{\$VALUES_ALL_$of}"); # hash
				$c->newline_on();
				$c->code(")"); #grep
			$c->close_block();
			$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		my $delim = $self->properties()->default_list_delimiter();
		$c->code("push");
		$c->open_block("(");
			$c->code($self->hash() eq '' ? "\@VALUES_ALL_$of, " : "\@{\$VALUES_ALL_$of}, ");
			$self->output_field()->input_field()->pequel_type()->name() eq 'array'
				? $c->code("split(/\\s*$delim\\s*/, $if)")
				: $c->code("qq{$if}");
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Count;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'count'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the count of records in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("@{[ $self->output_field()->getvar() ]}++");
		$c->over();
		$c->code("if (defined($if));@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Distinct;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'distinct'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%DISTINCT_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%DISTINCT_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("@{[ $self->output_field()->getvar() ]}++");
		$c->over();
		$c->code("if");
		$c->open_block("(");
			$c->code("defined($if)");
			$c->code("&& ++\$DISTINCT_$of\{qq{$if}} == 1");
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::CountDistinct;
	use base qw(ETL::Pequel3::Type::Aggregates::Distinct);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'count_distinct'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the count of unique values for field-name in the group.');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Minimum;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'minimum'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the minimum value in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = $self->output_field()->getvar();
		my $if = $self->output_field()->input_field()->getvar();
		if ($self->output_field()->pequel_type()->name() eq 'date')
		{
			$c->code("@{[ $self->prinfo() ]} --  date_type: @{[ $self->output_field()->date_type()->name() ]}");
			$c->code("if (!defined($of))");
			$c->open_block();
			$c->code("$of = $if;");
			$c->close_block();
			$c->code("else");
			$c->open_block();
			$c->add($self->output_field()->date_type()->code_cmp_date($if, $of));
			$c->code("if (\$cmp == -1) { $of = $if; }");
			$c->close_block();
		}
		else
		{
			$c->code("$of = $if");
			$c->over();
			$c->code("if");
			$c->open_block("(");
				$c->code("!defined($of)");
				$c->code("||");
				$c->open_block("(");
					$c->code("defined($if)");
					$c->code("&& $if ne ''");
					$c->code("&& $if");
					$c->over();
						$c->code("@{[ $self->output_field()->pequel_type()->name() eq 'string' ? 'lt' : '<' ]} $of");
					$c->back();
					#TODO: what about date comparison ??? -- need to use pequel_type()->lt()
				$c->close_block();
			$c->close_block();
			$c->code(";@{[ $self->prinfo() ]}");
			$c->back();
		}
#?		$c->code("if (!defined($of) || (defined($if) && @{[ $self->output_field()->pequel_type()->lt($if, $of) ]}));");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Maximum;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'maximum'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the maximum value in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = $self->output_field()->getvar();
		my $if = $self->output_field()->input_field()->getvar();
		if ($self->output_field()->pequel_type()->name() eq 'date')
		{
			$c->code("@{[ $self->prinfo() ]} --  date_type: @{[ $self->output_field()->date_type()->name() ]}");
			$c->code("if (!defined($of))");
			$c->open_block();
			$c->code("$of = $if;");
			$c->close_block();
			$c->code("else");
			$c->open_block();
			$c->add($self->output_field()->date_type()->code_cmp_date($if, $of));
			$c->code("if (\$cmp == 1) { $of = $if; }");
			$c->close_block;
		}
		else
		{
			$c->code("$of = $if");
			$c->over();
			$c->code("if");
			$c->open_block("(");
				$c->code("!defined($of)");
				$c->code("||");
				$c->open_block("(");
					$c->code("defined($if)");
					$c->code("&& $if ne ''");
					$c->code("&& $if");
					$c->over();
						$c->code("@{[ $self->output_field()->pequel_type()->name() eq 'string' ? 'gt' : '>' ]} $of");
					$c->back();
					#TODO: what about date comparison ??? -- need to use pequel_type()->gt()
				$c->close_block();
			$c->close_block();
			$c->code(";@{[ $self->prinfo() ]}");
			$c->back();
		}
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Min;
	use base qw(ETL::Pequel3::Type::Aggregates::Minimum);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'min'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the minimum value in the group. Same as minimum.');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Max;
	use base qw(ETL::Pequel3::Type::Aggregates::Maximum);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'max'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the maximum value in the group. Same as maximum.');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Last;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'last'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the last value in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("@{[ $self->output_field()->getvar() ]} = @{[ $self->output_field()->input_field()->getvar() ]};@{[ $self->prinfo() ]}")
			unless ($self->output_field()->getvar() eq $self->output_field()->input_field()->getvar());
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::First;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'first'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the first value in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = $self->output_field()->getvar();
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("$of = $if");
		$c->over();
		$c->code("if (!defined($of));@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Flag;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'require_input_field', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'flag'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => '');
	use Class::STL::ClassMembers::Constructor;
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("@{[ $self->output_field()->getvar() ]} = 1;@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::AvgDistinct;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'avg_distinct'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the average value for unique values for field-name in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%AVG_DISTINCT_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		$c->code("my \%AVG_DISTINCT_KEYS_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%AVG_DISTINCT_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		$c->code("\%AVG_DISTINCT_KEYS_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		$c->code("@{[ $self->output_field()->getvar() ]} =");
		$c->over();
		$c->open_block("(");
			$c->code("\$AVG_DISTINCT_$of\{_COUNT} == 0");
			$c->over();
				$c->code("? 0");
				$c->code(": \$AVG_DISTINCT_$of\{_SUM} / \$AVG_DISTINCT_$of\{_COUNT}");
			$c->back();
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("if");
		$c->open_block("(");
			$c->code("++\$AVG_DISTINCT_KEYS_$of\{qq{$if}}");
			$c->code("== 1");
		$c->close_block();
		$c->open_block();
			$c->code("\$AVG_DISTINCT_$of\{_SUM} += $if;");
			$c->code("\$AVG_DISTINCT_$of\{_COUNT}++;");
		$c->close_block();
		$c->code("@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Avg;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'avg'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the average value in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%@{[ uc($self->name()) ]}_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%@{[ uc($self->name()) ]}_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		$c->code("@{[ $self->output_field()->getvar() ]} =");
		$c->over();
		$c->open_block("(");
			$c->code("!defined(\$@{[ uc($self->name()) ]}_$of\{_COUNT}) || \$@{[ uc($self->name()) ]}_$of\{_COUNT} == 0");
			$c->over();
			$c->code("? 0");
			$c->code(": \$@{[ uc($self->name()) ]}_$of\{_SUM} / \$@{[ uc($self->name()) ]}_$of\{_COUNT}");
			$c->back();
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("\$@{[ uc($self->name()) ]}_$of\{_SUM} += $if;@{[ $self->prinfo() ]}");
		$c->code("\$@{[ uc($self->name()) ]}_$of\{_COUNT}++;@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Mean;
	use base qw(ETL::Pequel3::Type::Aggregates::Avg);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'mean'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the average value in the group. Same as avg.');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Median;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'median'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the middle vale in the group, or, in the case of an even number of values, output the mean of the two middle values in the group.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%MEDIAN_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		$c->code("my @{[ $self->hash() eq '' ? '$' : '%' ]}MEDIAN_COUNT_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%MEDIAN_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		$c->code("\$MEDIAN_COUNT_@{[ $self->output_field()->name() ]} = @{[ $self->hash() eq '' ? 'undef' : '()' ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $v1 = $self->hash() eq '' ? "\$MEDIAN_COUNT_$of / 2 - 1" : "\${\$MEDIAN_COUNT_$of} / 2 - 1";
		my $v2 = $self->hash() eq '' ? "\$MEDIAN_COUNT_$of / 2" : "\${\$MEDIAN_COUNT_$of} / 2";
		my $keys = $self->hash() eq '' ? "keys \%MEDIAN_$of" : "keys \%{\$MEDIAN_$of}";
		$c->newline_off();
		$c->code("@{[ $self->output_field()->getvar() ]} = ");
		$c->newline_on();
		$self->hash() eq ''
			? $c->code("\$MEDIAN_COUNT_$of % 2 == 0")
			: $c->code("\${\$MEDIAN_COUNT_$of} % 2 == 0");
		$c->over();
			$c->code("? \&{sub{(\$_[0] + \$_[1]) / 2}}");
			$c->open_block("(");
				$c->open_block("(");
					$c->code("( sort {\$a <=> \$b} $keys )");
					$c->over();
						$c->code("[$v1, $v2]");
					$c->back();
				$c->close_block();
				$c->code("[0,1]");
			$c->close_block();
			$c->code(": ( sort {\$a <=> \$b} $keys )");
			$c->open_block("[");
				$self->hash() eq ''
					? $c->code("((\$MEDIAN_COUNT_$of+1)/2)-1")
					: $c->code("((\${\$MEDIAN_COUNT_$of}+1)/2)-1");
			$c->close_block();
			$c->code(";@{[ $self->prinfo() ]}"); # hash
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$self->hash() eq ''
			? $c->code("\$MEDIAN_COUNT_$of++ ")
			: $c->code("\${\$MEDIAN_COUNT_$of}++ "); # hash
		$c->over();
		$c->code("if (++\$MEDIAN_$of\{qq{$if}} == 1);@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Variance;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'variance'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output ( sum squares / count ) - ( mean ** 2); sum_squares is each value in the distribution squared (** 2); count is the number of values in the distribution.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%VARIANCE_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%VARIANCE_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		$c->code("@{[ $self->output_field()->getvar() ]} = ");
		$c->over();
		$c->open_block("(");
			$c->open_block("(");
				$c->code("\$VARIANCE_$of\{_SUM_SQUARES}");
				$c->code("/ (\$VARIANCE_$of\{_COUNT} == 0 ? 1 : \$VARIANCE_$of\{_COUNT})");
			$c->close_block();
			$c->code("-");
			$c->open_block("(");
				$c->open_block("(");
					$c->code("\$VARIANCE_$of\{_SUM}");
					$c->code("/ (\$VARIANCE_$of\{_COUNT} == 0 ? 1 : \$VARIANCE_$of\{_COUNT})");
				$c->close_block();
				$c->code("** 2");
			$c->close_block();
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("\$VARIANCE_$of\{_SUM} += $if;@{[ $self->prinfo() ]}");
		$c->code("\$VARIANCE_$of\{_SUM_SQUARES} += $if ** 2;@{[ $self->prinfo() ]}");
		$c->code("\$VARIANCE_$of\{_COUNT}++;@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Stddev;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'stddev'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the square-root of variance.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%STDDEV_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%STDDEV_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		$c->code("@{[ $self->output_field()->getvar() ]} = ");
		$c->over();
		$c->code("sqrt");
		$c->open_block("(");
			$c->open_block("(");
				$c->code("\$STDDEV_$of\{_SUM_SQUARES}");
				$c->code("/ (\$STDDEV_$of\{_COUNT} == 0 ? 1 : \$STDDEV_$of\{_COUNT})");
			$c->close_block();
			$c->code("-");
			$c->open_block("(");
				$c->open_block("(");
					$c->code("\$STDDEV_$of\{_SUM}");
					$c->code("/ (\$STDDEV_$of\{_COUNT} == 0 ? 1 : \$STDDEV_$of\{_COUNT})");
				$c->close_block();
				$c->code("** 2");
			$c->close_block();
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("\$STDDEV_$of\{_SUM} += $if;@{[ $self->prinfo() ]}");
		$c->code("\$STDDEV_$of\{_SUM_SQUARES} += $if ** 2;@{[ $self->prinfo() ]}");
		$c->code("\$STDDEV_$of\{_COUNT}++;@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Range;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'range'),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric|date'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the maximum value minus the minimum value in a distribution.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%RANGE_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%RANGE_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		$c->code("@{[ $self->output_field()->getvar() ]} =");
		$c->over();
			$c->code("\$RANGE_$of\{_MAX}");
			$c->code("- \$RANGE_$of\{_MIN}");
			$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("\$RANGE_$of\{_MIN} = $if");
		$c->over();
		$c->code("if");
		$c->open_block("(");
			$c->code("!defined(\$RANGE_$of\{_MIN})");
			$c->code("||");
			$c->open_block("(");
				$c->code("defined($if)");
				$c->code("&& $if < \$RANGE_$of\{_MIN}");
			$c->close_block();
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();

		$c->code("\$RANGE_$of\{_MAX} = $if");
		$c->over();
		$c->code("if");
		$c->open_block("(");
			$c->code("!defined(\$RANGE_$of\{_MAX})");
			$c->code("||");
			$c->open_block("(");
				$c->code("defined($if)");
				$c->code("&& $if > \$RANGE_$of\{_MAX}");
			$c->close_block();
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Mode;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'mode'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Output the most frequently occuring score or scores (space delimited) in a distribution.');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::CodeStyler::Program::Perl->new();
		$c->code("my \%MODE_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_reset
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\%MODE_@{[ $self->output_field()->name() ]} = ();@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $keys = $self->hash() eq '' ? "keys \%MODE_$of" : "keys \%{\$MODE_$of}";
		$c->code("@{[ $self->output_field()->getvar() ]} = ");
		$c->over();
		$c->code("join");
		$c->open_block("(");
			$c->code("' ',");
			$c->code("&");
			$c->open_block();
				$c->code("sub");
				$c->open_block();
					$c->code("my \@top;");
					$c->code("foreach my \$k");
					$c->open_block("(");
						$c->code("sort");
						$c->open_block("{");
							$c->code("\$MODE_$of\{\$b} <=> \$MODE_$of\{\$a}");
						$c->close_block();
						$c->code("$keys");
					$c->close_block();
					$c->open_block();
						$c->code("last if (\$MODE_$of\{\$k} != \$MODE_$of\{\$_[0]});");
						$c->code("push(\@top, \$k);");
					$c->close_block();
					$c->code("\@top;");
				$c->close_block();
			$c->close_block();
			$c->open_block("(");
				$c->open_block("(");
					$c->code("sort");
					$c->open_block("{");
						$c->code("\$MODE_$of\{\$b} <=> \$MODE_$of\{\$a}");
					$c->close_block();
					$c->code("$keys");
				$c->close_block();
				$c->code("[0]");
			$c->close_block();
		$c->close_block();
		$c->code(";@{[ $self->prinfo() ]}");
		$c->back();
		return $c;
	}
    sub code_output
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $of = "@{[ $self->output_field()->name() ]}@{[ $self->hash() ]}";
		my $if = $self->output_field()->input_field()->getvar();
		$c->code("\$MODE_$of\{qq{$if}}++;@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Serial;
	use base qw(ETL::Pequel3::Type::Aggregates::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'serial'),
		Class::STL::ClassMembers::DataMember->new(name => 'require_input_field', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'valid_types', default => 'amount|numeric'), #TODO: date
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => '');
	use Class::STL::ClassMembers::Constructor;
    sub code_init
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
#>		$self->output_field()->pequel_type()->name() eq 'date'
#>			? $c->code("my \$_SERIAL_@{[ $self->output_field()->name() ]} = @{[ 
#>					defined($self->output_field()->serial_start()) ? $self->output_field()->serial_start() : '<today>' 
#>				]};")
#>			: $c->code("my \$_SERIAL_@{[ $self->output_field()->name() ]} = @{[ 
#>					defined($self->output_field()->serial_start()) ? $self->output_field()->serial_start() : '0' 
#>				]};");
		$c->code("my \$_SERIAL_@{[ $self->output_field()->name() ]} = @{[ $self->output_field()->serial_start() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
    sub code_output_final
	{ 
		my $self = shift; 
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
#>		$self->output_field()->pequel_type()->name() eq 'date'
#>			? $c->code("@{[ $self->output_field()->getvar() ]} = do { \$_SERIAL_@{[ $self->output_field()->name() ]} = &date_next_day(...); ... };");
		$c->code("@{[ $self->output_field()->getvar() ]} = ++\$_SERIAL_@{[ $self->output_field()->name() ]};@{[ $self->prinfo() ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Aggregates::Catalogue;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'aggregates'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::Aggregates::Abstract');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Type::Aggregates::Count->_new(),
			ETL::Pequel3::Type::Aggregates::CountDistinct->_new(),
			ETL::Pequel3::Type::Aggregates::Min->_new(),
			ETL::Pequel3::Type::Aggregates::Max->_new(),
			ETL::Pequel3::Type::Aggregates::Minimum->_new(),
			ETL::Pequel3::Type::Aggregates::Maximum->_new(),
			ETL::Pequel3::Type::Aggregates::Sum->_new(),
			ETL::Pequel3::Type::Aggregates::SumDistinct->_new(),
			ETL::Pequel3::Type::Aggregates::First->_new(),
			ETL::Pequel3::Type::Aggregates::Last->_new(),
			ETL::Pequel3::Type::Aggregates::Flag->_new(),
			ETL::Pequel3::Type::Aggregates::Distinct->_new(),
			ETL::Pequel3::Type::Aggregates::Avg->_new(),
			ETL::Pequel3::Type::Aggregates::AvgDistinct->_new(),
			ETL::Pequel3::Type::Aggregates::Serial->_new(),
			ETL::Pequel3::Type::Aggregates::ValuesAll->_new(),
			ETL::Pequel3::Type::Aggregates::ValuesUniq->_new(),
			ETL::Pequel3::Type::Aggregates::Mean->_new(),
			ETL::Pequel3::Type::Aggregates::Median->_new(),
			ETL::Pequel3::Type::Aggregates::Variance->_new(),
			ETL::Pequel3::Type::Aggregates::Stddev->_new(),
			ETL::Pequel3::Type::Aggregates::Range->_new(),
			ETL::Pequel3::Type::Aggregates::Mode->_new(),
#>			stderr (standard error) = stddev / count-square-root (eg: 28.56 / square-root 14 = 28.56 / 3.74 = 7.6
			# --> stddev(f) /  sqrt(count(*))
#>			bool_and -- true if all input values are true, otherwise false
#>			every -- same as bool_and
#>			bool_or -- true if at least one input value is true, otherwise false
#>			bit_and -- the bitwise AND of all non-null input values, or null if none
#>			bit_or -- the bitwise OR of all non-null input values, or null if none
		);
		return $self;
	}
	sub catalogue
	{
		my $self = shift;
		my $xml_node = shift;
		foreach ($self->to_array()) {
			my $s_xml = $xml_node->createChild("aggregate-type");
			$s_xml->attribute('aggregate-name', $_->name());
				my $m_xml = $s_xml->createChild("property");
				$m_xml->attribute('valid-types', $_->valid_types());
		}
	}
}
# ----------------------------------------------------------------------------------------------------
1;
