# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::CmdOption.pm
#  Created	: 14 March 2007
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::CmdOption;
require 5.005_62;
use strict;
use warnings;
use stl;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Abstract;
	use base qw(Class::STL::Element);
	use Class::STL::ClassMembers 
		qw( 
			long_name 
			short_name
			default 
			description 
			prop_name
		),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 0);
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->data(1) if ($self->flag() && !defined($self->data()))
	}
	sub prep #NOTE: called before produce()
	{
		my $self = shift;
		my $POM = shift;
	}
	sub exec #NOTE: called after produce()
	{
		my $self = shift;
		my $POM = shift;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::UserOption;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use ETL::Pequel3;
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'user_option'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'User defined option');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Version;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use ETL::Pequel3;
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'version'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'v'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Display Pequel version information');
	use Class::STL::ClassMembers::Constructor;
	sub prep
	{
		my $self = shift;
		my $POM = shift;
		print STDERR "Pequel3 version @{[ $POM->properties()->version() ]}\n";
		print STDERR "-->please check http://sourceforge.net/projects/pequel/ for updates.\n";
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Prefix;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'prefix'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'pre'),
		Class::STL::ClassMembers::DataMember->new(name => 'prop_name', default => 'prefix'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => ''),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Directory pathname prefix');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::XmlParser;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'xml_parser'),
		Class::STL::ClassMembers::DataMember->new(name => 'prop_name', default => 'xml_parser'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => 'element'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'XML parser type name');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::ExecDir;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'exec_dir'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'ed'),
		Class::STL::ClassMembers::DataMember->new(name => 'prop_name', default => 'exec_dir'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => './_Pequel'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Temporary directory for generated programs');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::TraceDir;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'trace_dir'),
		Class::STL::ClassMembers::DataMember->new(name => 'prop_name', default => 'td'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => './_Pequel'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Temporary directory for trace log files and generated programs');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::TraceFile;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'trace_file'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'tf'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Trace log filename');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Trace;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'trace'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Save trace information in trace-file during generated program execution');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Diag;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'diagnostics'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'diag'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Display diagnostic information during generated program execution');
	use Class::STL::ClassMembers::Constructor;
	sub prep
	{
		my $self = shift;
		my $POM = shift;
		ETL::Pequel3::Error->new()->diag_level(1);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::DiagLevel;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'diagnostics_level'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'diag_level'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Specify diagnostic detail level');
	use Class::STL::ClassMembers::Constructor;
	sub prep
	{
		my $self = shift;
		my $POM = shift;
		ETL::Pequel3::Error->new()->diag_level($self->data());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Check;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'check'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'c'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Syntax check the generated program');
	use Class::STL::ClassMembers::Constructor;
	sub prep
	{
		my $self = shift;
		my $POM = shift;
		my $getopt = ETL::Pequel3::GetOpt->new();
		$getopt->user()->push_back(ETL::Pequel3::Type::CmdOption::NoExec->new())
			unless ($getopt->user()->exists('noexec'));
	}
	sub exec
	{
		my $self = shift;
		my $POM = shift;
		if ($POM->check() !~ /syntax OK/i)
		{
			print STDERR "Script failed syntax check!\n";
			print STDERR "-->for debugging please check end of generated code dump in '@{[
				$POM->generator()->user_prog()->program_name() ]}.DEBUG' for error messages.\n";
		}
		else
		{
			print STDERR "@{[ $POM->generator()->user_prog()->program_name() ]} syntax OK\n";
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Xsdgen;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'xsdgen'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'xsd'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Produce the Pequel XML Schema XSD');
	use Class::STL::ClassMembers::Constructor;
	sub prep
	{
		my $self = shift;
		my $POM = shift;
		$POM->xsdgen();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Docgen;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'docgen'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Produce the Pequel User Script Reference document pdf');
	use Class::STL::ClassMembers::Constructor;
	sub exec
	{
		my $self = shift;
		my $POM = shift;
		$POM->docgen();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Catgen;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'catgen'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'catalogue'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Produce the Pequel Type Catalogue pdf');
	use Class::STL::ClassMembers::Constructor;
	sub exec
	{
		my $self = shift;
		my $POM = shift;
		$POM->catgen();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::ViewCode;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'view_code'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'viewcode'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Display the generated program');
	use Class::STL::ClassMembers::Constructor;
	sub prep
	{
		my $self = shift;
		my $POM = shift;
		my $getopt = ETL::Pequel3::GetOpt->new();
		$getopt->user()->push_back(ETL::Pequel3::Type::CmdOption::NoExec->new())
			unless ($getopt->user()->exists('noexec'));
	}
	sub exec
	{
		my $self = shift;
		my $POM = shift;
		$POM->display_code();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Silent;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'silent'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 's'),
		Class::STL::ClassMembers::DataMember->new(name => 'prop_name', default => 'silent'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Suppress all process messages');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Verbose;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'verbose'),
		Class::STL::ClassMembers::DataMember->new(name => 'prop_name', default => 'verbose'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => ''),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Display progress information');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::DumpCode;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'dump_code'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'dumpcode'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Save the generated program in external file');
	use Class::STL::ClassMembers::Constructor;
	sub exec
	{
		my $self = shift;
		my $POM = shift;
		$POM->dump_code($self->data());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::DumpHash;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'dump_hash'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Save a hash dump of the Pequel-Object-Model in external file');
	use Class::STL::ClassMembers::Constructor;
	sub exec
	{
		my $self = shift;
		my $POM = shift;
		$POM->to_hash($self->data());
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::ScriptFilename;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'script_filename'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'f'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => ''),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Specify the pequel script xml or dump filename to execute');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::NoExec;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'no_exec'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'noexec'),
		Class::STL::ClassMembers::DataMember->new(name => 'default', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 0),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Do no execute the generated program');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Usage;
	use base qw(ETL::Pequel3::Type::CmdOption::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'long_name', default => 'usage'),
		Class::STL::ClassMembers::DataMember->new(name => 'short_name', default => 'help'),
		Class::STL::ClassMembers::DataMember->new(name => 'flag', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'user_option', default => 1),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Display pequel3 command-line usage information');
	use Class::STL::ClassMembers::Constructor;
	sub prep
	{
		my $self = shift;
		my $POM = shift;
		print STDERR "Pequel3 @{[ $POM->properties()->version() ]} Usage:\n";
		foreach ($POM->catalogue()->options()->to_array())
		{
			print STDERR sprintf("%-30s -- %s",
				"-@{[ $_->long_name() ]}@{[ defined($_->short_name()) ? (' | -' . $_->short_name()) : '' ]}",
				"@{[ $_->description() ]}\n");
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Catalogue::Abstract;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'options'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'long_name|short_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::CmdOption::Abstract');
	use Class::STL::ClassMembers::Constructor;
	use vars qw( $AUTOLOAD );
	sub AUTOLOAD
	{
		my $self = shift;
		(my $option = $AUTOLOAD) =~ s/.*:://;
		my $p;
		return $self->err()->user_error(10601, "Unknown option name '$option'!")
			unless ($p = $self->exists($option));
		return $p->data(@_);
	}
	sub DESTROY { }
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Catalogue;
	use base qw(ETL::Pequel3::Type::CmdOption::Catalogue::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Type::CmdOption::Diag->new(),
			ETL::Pequel3::Type::CmdOption::DiagLevel->new(),
			ETL::Pequel3::Type::CmdOption::TraceDir->new(),
			ETL::Pequel3::Type::CmdOption::TraceFile->new(),
			ETL::Pequel3::Type::CmdOption::Trace->new(),
			ETL::Pequel3::Type::CmdOption::Usage->new(),
			ETL::Pequel3::Type::CmdOption::Version->new(),
			ETL::Pequel3::Type::CmdOption::Prefix->new(),
			ETL::Pequel3::Type::CmdOption::DumpHash->new(),
			ETL::Pequel3::Type::CmdOption::DumpCode->new(),
			ETL::Pequel3::Type::CmdOption::Verbose->new(),
			ETL::Pequel3::Type::CmdOption::Silent->new(),
			ETL::Pequel3::Type::CmdOption::ViewCode->new(),
			ETL::Pequel3::Type::CmdOption::Catgen->new(),
			ETL::Pequel3::Type::CmdOption::Docgen->new(),
			ETL::Pequel3::Type::CmdOption::Xsdgen->new(),
			ETL::Pequel3::Type::CmdOption::Check->new(),
			ETL::Pequel3::Type::CmdOption::XmlParser->new(),
			ETL::Pequel3::Type::CmdOption::ExecDir->new(),
			ETL::Pequel3::Type::CmdOption::ScriptFilename->new(),
			ETL::Pequel3::Type::CmdOption::NoExec->new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::CmdOption::Catalogue::User;
	use base qw(ETL::Pequel3::Type::CmdOption::Catalogue::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
1;
