# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::DataSet.pm
#  Created	: 7 September 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# 16/05/2007	0.94		DB		More efficient split in Input/Ascii;
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::DataSet;
require 5.005_62;
use strict;
use warnings;
use ETL::Pequel3::Type::DataMember;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Abstract;
	use base qw(Class::STL::Element); # --> ETL::Pequel3::Type::Code; -- contains code_ functions
	use Class::STL::ClassMembers 
		qw( 
			err
			catalogue
			properties 
			configuration

			name 
			dataset_spec 
			datasource
			count_vname
			arr_vname
			numflds_vname
			default_datasource_spec
			datasource_type 
			datasource_factory_type
			datasource_factory
			use_list
			pequel_ref
			description
		),
		ETL::Pequel3::Type::DataMember::User->new(name => 'shared', default => 0, validate => '^(0|1)$'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'verify', default => 0, validate => '^(0|skip|abort)$');
	use Class::STL::ClassMembers::Constructor;
	use ETL::Pequel3::CodeStyler;
	sub new_extra
	{
		my $self = shift;
		$self->err(ETL::Pequel3::Error->new());
		use ETL::Pequel3::Catalogue;
		$self->catalogue(ETL::Pequel3::Catalogue->new());
		$self->properties($self->catalogue()->properties());
		$self->configuration($self->pequel_ref()->config());

		$self->datasource_factory(eval("@{[ $self->datasource_factory_type() ]}->new()"));
		$self->datasource($self->datasource_factory()->factory(
				@_, 
				pequel_ref => $self->pequel_ref(),
				dataset => $self,
				shared => $self->shared(), 
				datasource_type => $self->datasource_type()
			)
		);

		$self->arr_vname($self->datasource()->vname()) if (defined($self->datasource()->vname()));
		$self->count_vname("@{[ $self->datasource()->vname() || 'UNDEF' ]}_COUNT");
		$self->numflds_vname("@{[ $self->datasource()->vname() || 'UNDEF' ]}_NUMFLDS");
		$self->use_list(stl::list());
		return $self;
	}
	sub parse
	{
		my $self = shift;
#<		$self->configuration($self->pequel_ref()->config());
	}
	sub map_output
	{
		my $self = shift;
		my %p = @_;
		$self->datasource()->_map_output(@_);
	}
	sub code_init
	{	
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		return $self->datasource()->_code_init($c);
	}
	sub code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		return $self->datasource()->_code_open($c);
	}
	sub code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $self->datasource()->_code_close($c);
	}
	sub code_lock
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $self->datasource()->_code_lock($c);
	}
	sub code_unlock
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $self->datasource()->_code_unlock($c);
	}
	sub code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $self->datasource()->_code_prepare($c);
	}
	sub code_reset
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::Abstract;
	use base qw(ETL::Pequel3::Type::DataSet::Abstract);
	use Class::STL::ClassMembers qw( input_fields ),
		Class::STL::ClassMembers::DataMember->new(name => 'default_datasource_spec', default => 'stdin'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_factory_type', default => 'ETL::Pequel3::Type::DataSource::Input::Factory::Local'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'rmctrlm', default => 0),
		ETL::Pequel3::Type::DataMember::User->new(name => 'discard_header', default => 0),
		ETL::Pequel3::Type::DataMember::User->new(name => 'chomp', default => 0);
	use Class::STL::ClassMembers::Constructor;
	sub code_init
	{	
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		$c->code('my @' . $self->arr_vname() . ';');
		$c->over();
			$c->code("# @{[ defined($self->dataset_spec()) ? $self->dataset_spec() : 'UNDEF' ]}");
			$c->code("# @{[ ref($self) ]}");
		$c->back;
		$c->code("my \$@{[ $self->count_vname() ]}=0;");
		return $self->datasource()->_code_init($c);
	}
	sub code_decode
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $numflds = shift;
		return $c;
	}
	sub code_read 
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		my $numflds = int(@fields);
		return $c;
	}
	sub code_after_read #--> code_decode #TODO: not here -- not applicable to DBI?
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		if ($self->chomp()) {
			$c->code("chomp(\$" . $self->arr_vname() . "[\$#" . $self->arr_vname() . "])");
			$c->over();
				$c->code("if (defined(\$" . $self->arr_vname() . "[\$#" . $self->arr_vname() . "]));");
			$c->back();
		}
		$c->code("chop;") if ($self->rmctrlm());
		return $c;
	}
	sub code_reset
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		$c->code('undef(@' . $self->arr_vname() . ');');
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::Abstract;
	use base qw(ETL::Pequel3::Type::DataSet::Abstract);
	use Class::STL::ClassMembers qw( output_fields ),
		Class::STL::ClassMembers::DataMember->new(name => 'default_datasource_spec', default => 'stdout'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_factory_type', default => 'ETL::Pequel3::Type::DataSource::Output::Factory::Local'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'print_header', default => 0);
	use Class::STL::ClassMembers::Constructor;
	sub code_init
	{	
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		$c->code('my ' . ($self->configuration()->hash() ? '%' : '@') . $self->arr_vname() . ';'); 
		$c->over();
			$c->code("# @{[ defined($self->dataset_spec()) ? $self->dataset_spec() : 'UNDEF' ]}");
			$c->code("# @{[ ref($self) ]}");
		$c->back();
		return $self->datasource()->_code_init($c);
	}
	sub code_write
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		return $c;
	}
	sub code_write_header
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		return $c;
	}
	sub code_reset
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		$c->code('undef(' . ($self->configuration()->hash() ? '%' : '@') . $self->arr_vname() . ');');
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::AsciiDelimited;
	use base qw(ETL::Pequel3::Type::DataSet::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'This dataset consists of I<ascii delimited text> with the delimiter as specified by the F<field_delimiter> option.'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'ascii'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'chomp', default => 1),
		ETL::Pequel3::Type::DataMember::User->new(name => 'rmctrlm', default => 0),
		ETL::Pequel3::Type::DataMember::User->new(name => 'field_delimiter', default => '|');
	use Class::STL::ClassMembers::Constructor;
	sub code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		$self->datasource()->_code_open($c);
		if ($self->discard_header()) {
			$self->code_lock($c);
			$c->code("my \$_discard_@{[ $self->datasource()->vname() ]} = <@{[ $self->datasource()->fdname() ]}>;");
			$self->code_unlock($c);
		}
		return $c;
	}
	sub code_decode
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $numflds = shift;
        my $fdelim_len = length($self->field_delimiter());
        my $fdelim     = $self->field_delimiter();
        ## if we pass in a single character as a delimiter, it seems to be about 7%
        ## faster if we do not enclose it inside of [] in the regular expression
        ##
        ## I tried implementing split as an Inline::C function and it did not perform
        ## as well as I had hoped
        ##
        ## - David Bartle
		## This patch will use a different regular expression syntax if the delimiter 
		## is only a single character (which is most of the time).  On my system at lest, 
		## it showed about a 7% performance increase which is only really noticeable on 
		## very huge input sources.  I don't know if your benchmarks would show a marked 
		## improvement but I thought I would  send the patch along none-the-less.
        if( length($fdelim) == 1 || (length($fdelim) == 2 && $fdelim =~ /^\\/)) {
            my $fdelim_copy = $fdelim;
            $fdelim_copy =~ s/"/\"/g;
            $c->code("split(\"\\Q$fdelim_copy\\E\", \$_, -1)");
        }
        else {
            $c->code("split(\"[@{[ $self->field_delimiter() ]}]\", \$_, -1)");
        }
		return $c;
	}
	sub code_read
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		my $numflds = int(@fields);
		$c->newline_off();
		#TODO: this may not work with some input pipe type (eg sqlplus)
		$c->code("defined(\$_=<@{[ $self->datasource()->fdname() ]}>) && (\@@{[ $self->arr_vname() ]} = ");
		$self->code_decode($c, $numflds);
		$c->code(")");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::AsciiDelimited;
	use base qw(ETL::Pequel3::Type::DataSet::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'This dataset consists of I<ascii delimited text> with the delimiter as specified by the F<field_delimiter> option.'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'ascii'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'field_delimiter', default => '|');
	use Class::STL::ClassMembers::Constructor;
	sub code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("local(\$,)=\"@{[ $self->field_delimiter() ]}\";");
		return $self->datasource()->_code_prepare($c);
	}
	sub code_write_header
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		$c->open_block();
			$self->code_prepare($c);
			$c->code("print @{[ $self->datasource()->fdname() ]}");
			$c->over();
				map($c->code("'@{[ $_->name() ]}',"), @fields);
				$c->code(";");
			$c->back();
		$c->close_block();
		return $c;
	}
	sub code_write
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
							# need field pointers so we can determine type, eg for CSV to quote string fields
		$c->open_block();
			$self->code_prepare($c);
			$c->code("print @{[ $self->datasource()->fdname() ]}");
			$c->over();
				map($c->code("@{[ $_->pequel_type()->getvar($_) ]},"), @fields);
#<				map($c->code("@{[ $_->getvar() ]},"), @fields);
				$c->code(";");
			$c->back();
		$c->close_block();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::Xml;
	use base qw(ETL::Pequel3::Type::DataSet::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'This dataset type will output the data in F<XML> format.'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'xml'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'print_header', default => 0),
		ETL::Pequel3::Type::DataMember::User->new(name => 'field_delimiter', default => '\n'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'xml_version_str', default => 'xml version="1.0"'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'qualify_field_names', default => 1),
		ETL::Pequel3::Type::DataMember::User->new(name => 'rec_element_name', default => 'record'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'root_element_name', default => 'file');
	use Class::STL::ClassMembers::Constructor;
	sub code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("local(\$,)=\"@{[ $self->field_delimiter() ]}\";");
		return $self->datasource()->_code_prepare($c);
	}
	sub code_open
	{
		my $self = shift;
		my $c = $self->datasource()->_code_open(@_);
		$c->open_block();
			$self->code_prepare($c);
			$c->code("print @{[ $self->datasource()->fdname() ]}");
			$c->over();
				$c->code("'<?@{[ $self->xml_version_str() ]}?>',");
				$c->code("'<@{[ 
					$self->root_element_name() 
				]} pequel_script=\"@{[ 
					$self->pequel_ref()->script_filename()
				]}\" pequel_version=\"@{[ 
					$self->pequel_ref()->properties()->version()
				]}\">'");
				$c->code(";");
			$c->back();
		$c->close_block();
		return $c;
	}
	sub code_write
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		#print OUT
		#	"  <record att1=\"@{[ $_O_VAL[att1] ]}\" att2=\"@{[ $_O_VAL[att2] ]}\">",
		#	"    <F1>@{[ $_O_VAL[F1] ]}</F1>",
		#	"    <F2>@{[ $_O_VAL[F2] ]}</F2>",
		#	"    <F3><![CDATA[ @{[ $_O_VAL[F3] ]} ]]>",
		#	"  </record>"
		#	;
		$c->open_block();
			$self->code_prepare($c);
			$c->code("print @{[ $self->datasource()->fdname() ]}");
			$c->over();
			my $attrs =
				join
				(
					' ', 
					$self->rec_element_name(),
					map
					(
						qq#@{[ $self->qualify_field_names() ? $_->tname() : $_->name() 
							]}=\\"\@{[ @{[ $_->pequel_type()->getvar($_) ]} ]}\\"#, 
						grep($_->xml_type() eq 'attr', @fields)
					)
				);
			$c->code("\"  <$attrs>\",");
			map
			(
				do
				{
					$c->code("(\"    <@{[ $self->qualify_field_names() ? $_->tname() : $_->name() 
							]}>@{[ $_->xml_type() eq 'CDATA' ? '<![CDATA[' : '']}\"");
					$c->over();
					$c->code(". @{[ $_->pequel_type()->getvar($_) ]}");
					$c->code(". \"@{[ $_->xml_type() eq 'CDATA' ? ']]>' : '' ]}</@{[ 
							$self->qualify_field_names() ? $_->tname() : $_->name() ]}>\"),");
					$c->back();
				},
				grep($_->xml_type() ne 'attr', @fields)
			);
			$c->code("\"  </@{[ $self->rec_element_name() ]}>\"");
			$c->code(";");
			$c->back();
		$c->close_block();
		return $c;
	}
	sub code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->open_block();
			$self->code_prepare($c);
			$c->code("print @{[ $self->datasource()->fdname()
				]} '</@{[ $self->root_element_name() ]}>';");
		$c->close_block();
		return $self->datasource()->_code_close($c);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::Packed;
	use base qw(ETL::Pequel3::Type::DataSet::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Specify F<packed> data format. The format should be specified with the F<pack_format> option. The default option is B<[A3/Z*]>. Enclose the format in square brackets to specify a repeating format; that is, the same format is used for each field.'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'packed'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'pack_fmt', default => '[A3/Z*]'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'field_delimiter', default => ','); # TODO: remove
	use Class::STL::ClassMembers::Constructor;
	sub code_decode
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $numflds = shift;
		$c->code("unpack('"
			. ($self->pack_fmt()  =~ /^\[(.*)\]$/ ? $1 x $numflds : $self->pack_fmt())
			. "', \$_)");
		return $c;
	}
	sub code_read
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		my $numflds = int(@fields);
		$c->newline_off();
		$c->code("defined(\$_=<@{[ $self->datasource()->fdname() ]}>) && (\@@{[ $self->arr_vname() ]} = ");
		$self->code_decode($c, $numflds);
		$c->code(")");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::Packed;
	use base qw(ETL::Pequel3::Type::DataSet::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Specify F<packed> data format. The format should be specified with the F<pack_format> option. The default option is B<[A3/Z*]>. Enclose the format in square brackets to specify a repeating format; that is, the same format is used for each field.'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'packed'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'pack_fmt', default => '[A3/Z*]');
	use Class::STL::ClassMembers::Constructor;
	sub code_write_header
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		$c->open_block();
			$self->code_prepare($c);
			$c->code("print @{[ $self->datasource()->fdname() ]} pack '"
				. ($self->pack_fmt()  =~ /^\[(.*)\]$/ ? $1 x int(@fields) : $self->pack_fmt())
				. "', ");
			$c->over();
				map($c->code("'@{[ $_->name() ]}',"), @fields);
				$c->code(";");
			$c->back();
		$c->close_block();
		return $c;
	}
	sub code_write
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		$c->open_block();
			$self->code_prepare($c);
			$c->code("print @{[ $self->datasource()->fdname() ]} pack '"
				. ($self->pack_fmt()  =~ /^\[(.*)\]$/ ? $1 x int(@fields) : $self->pack_fmt())
				. "', ");
			$c->over();
				map($c->code("@{[ $_->getvar() ]},"), @fields);
				$c->code(";");
			$c->back();
		$c->close_block();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
#>	package ETL::Pequel3::Type::DataSource::Input::Dbi;
#>	package ETL::Pequel3::Type::DataSource::Input::SqlPlus;
#>	package ETL::Pequel3::Type::DataSet::Input::Sql;
	package ETL::Pequel3::Type::DataSet::Input::Dbi;
	use base qw(ETL::Pequel3::Type::DataSet::Input::Abstract);
	use Class::STL::ClassMembers 
		qw( 
			sth_vname
			dbh_vname

			_dsn_parsed
			_user_parsed
			_password_parsed
			_table_name_parsed
			_select_override_parsed
			_select_where_condition_parsed
			_select_fields_parsed
			_select_order_by_parsed
			_select_group_by_parsed
		),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'This dataset is for DBI connectivity.'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'dbi'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_type', default => 'dbi'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'dsn'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'user'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'password'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'table_name'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'select_override'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'select_where_condition'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'select_fields'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'select_order_by'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'select_group_by'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'chomp', default => 1);
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->table_name($self->datasource()->target_name()) if (defined($self->datasource()->target_name()));
		$self->sth_vname("_STH@{[ $self->datasource()->vname() ]}") if (defined($self->datasource()->vname()));
		$self->dbh_vname("_DBH@{[ $self->datasource()->vname() ]}") if (defined($self->datasource()->vname()));
		return $self;
	}
	sub map_output
	{
		my $self = shift;
		my %p = @_;
		$self->input_fields()->map_fields($p{field_map}, $self) 
			if (exists($p{field_map}) && defined($p{field_map}));

		# Preserve derived fields:
		use stl qw(deque);
		my $derived = deque();
		while 
		(
			$self->input_fields()->size() 
			&& $self->input_fields()->back()->can('calc_exp') 
			&& defined($self->input_fields()->back()->calc_exp())
		) 
		{
			$derived->push_front($self->input_fields()->back());
			$self->input_fields()->pop_back();
		}

		if (!$self->input_fields()->size()) {
			foreach ($self->get_fields()) {
				$self->input_fields()->add(
					name => $_, 
					ds_column => $self->input_fields()->size()+1,
					dataset => $self,
#?					target_ref => $p{target_ref},
					configuration => $self->configuration(),
				);
			}
		}
		my @flds = map($_->name(), $self->input_fields()->to_array());
		map($_ =~ s/^_//, @flds);
		$self->select_fields(join(', ', @flds));
		$self->select_order_by($p{key_field}) 
			if (exists($p{key_field}) && defined($p{key_field}));

		foreach ($derived->to_array()) {
			$self->input_fields()->push_back($_);
			$self->input_fields()->back()->field_number($self->input_fields()->size());
		}
	}
	sub get_fields
	{
		my $self = shift;
		use DBI;
		my $dbh = DBI->connect($self->dsn(), $self->user(), $self->password(), { RaiseError => 1, AutoCommit => 0 })
			|| die("Unable to connect to data source table '@{[ $self->dsn() ]}' for '@{[ $self->table_name() ]}': $DBI::errstr");
		my $sth = $dbh->prepare("select @{[ defined($self->select_override()) ? $self->select_override() : '*' ]} from @{[ $self->table_name() ]}");
		my @flds = (@{$sth->{NAME}});
		$sth->finish();
		$dbh->disconnect();
		return (@flds);
	}
	sub code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("use DBI;");
		$c->code("my \$@{[ $self->datasource()->vname() ]}_dsn = '@{[ 
				defined($self->_dsn_parsed()) ? $self->_dsn_parsed()->raw() : $self->dsn()
			]}';");
		$c->code("my \$@{[ $self->datasource()->vname() ]}_user = '@{[ 
				defined($self->_user_parsed()) ? $self->_user_parsed()->raw() : $self->user()
			]}';");
		$c->code("my \$@{[ $self->datasource()->vname() ]}_password = '@{[ 
				defined($self->_password_parsed()) ? $self->_password_parsed()->raw() : $self->password()
			]}';");
		$c->code(
			"my \$@{[ 
				$self->dbh_vname() 
			]} = DBI->connect(\$@{[ 
				$self->datasource()->vname() 
			]}_dsn, \$@{[ 
				$self->datasource()->vname() 
			]}_user, \$@{[ 
				$self->datasource()->vname() 
			]}_password, { RaiseError => 1, AutoCommit => 0 })"
		);
		$c->over();
		$c->code("|| die(\"Unable to connect to data source table '@{[ 
				$self->datasource()->target_name() 
			]}' for '@{[ 
				$self->datasource()->vname() 
			]}': \$DBI::errstr\");");
		$c->back();
		$c->newline_off();
		$c->code("my \$@{[ $self->sth_vname() ]} = \$@{[ $self->dbh_vname() ]}->prepare(\"");
		if (defined($self->select_override())) {
			$c->code(defined($self->_select_override_parsed()) 
						? $self->_select_override_parsed()->raw() 
						: $self->select_override()
				);
		}
		else
		{
			$c->code("select ");
			$c->code(
					defined($self->_select_fields_parsed())
						? $self->_select_fields_parsed()->raw()
						: $self->select_fields()
					);
			$c->code(" from @{[ 
					defined($self->_table_name_parsed()) 
						? $self->_table_name_parsed()->raw() 
						: $self->table_name()
				]}");
			$c->code(" where @{[ 
					defined($self->_select_where_condition_parsed()) 
						? $self->_select_where_condition_parsed()->raw() 
						: $self->select_where_condition()
				]}") 
				if (defined($self->select_where_condition()));
			$c->code(" order by @{[ 
					defined($self->_select_order_by_parsed()) 
						? $self->_select_order_by_parsed()->raw() 
						: $self->select_order_by()
				]}") 
				if (defined($self->select_order_by()));
			$c->code(" group by @{[ 
					defined($self->_select_group_by_parsed()) 
						? $self->_select_group_by_parsed()->raw() 
						: $self->select_group_by()
				]}") 
				if (defined($self->select_group_by()));
		}
		$c->newline_on();
		$c->code("\");");
		$c->code("\$@{[ $self->sth_vname() ]}->execute() || die(\"Can't execute statement: \$DBI::errstr\");");
		return $c;
	}
	sub code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("\$@{[ $self->sth_vname() ]}->finish();");
		$c->code("\$@{[ $self->dbh_vname() ]}->disconnect();");
		return $c;
	}
	sub code_decode
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $numflds = shift;
		$c->code("\$@{[ $self->sth_vname() ]}->fetchrow_array()");
		return $c;
	}
	sub code_read
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my @fields = @_; # ETL::Pequel3::Type::Field::Abstract object list;
		my $numflds = int(@fields);
		$c->newline_off();
		$c->code("\@@{[ $self->arr_vname() ]} = ");
		$self->code_decode($c, $numflds);
		return $c;
	}
	sub parse
	{
		my $self = shift;
		$self->_dsn_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->dsn()))
			if (defined($self->dsn()));
		$self->_user_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->user()))
			if (defined($self->user()));
		$self->_password_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->password()))
			if (defined($self->password()));
		$self->_table_name_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->table_name()))
			if (defined($self->table_name()));
		$self->_select_override_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->select_override()))
			if (defined($self->select_override()));
		$self->_select_where_condition_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->select_where_condition()))
			if (defined($self->select_where_condition()));
		$self->_select_fields_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->select_fields()))
			if (defined($self->select_fields()));
		$self->_select_order_by_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->select_order_by()))
			if (defined($self->select_order_by()));
		$self->_select_group_by_parsed($self->pequel_ref()->parser()->used_by($self)->translate($self->select_group_by()))
			if (defined($self->select_group_by()));
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::Sampler;
	use base qw(ETL::Pequel3::Type::DataSet::Input::AsciiDelimited);
	use Class::STL::ClassMembers qw( recs_vname _field_init_values ),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'sampler'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_type', default => 'void'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'records', default => 1);
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->recs_vname("@{[ $self->datasource()->vname() || 'UNDEF' ]}_RECS");
	}
	sub code_init
	{	
		my $self = shift;
		my $c = $self->SUPER::code_init(@_);
		$c->code("my \$@{[ $self->recs_vname() ]}=@{[ $self->records() ]};");
		return $c;
	}
	sub code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub code_decode
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $numflds = shift;
		return $c;
	}
	sub code_read
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("--\$@{[ $self->recs_vname() ]} >= 0");
		return $c;
	}
	sub code_after_read
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->newline_off();
		$c->code("\@@{[ $self->arr_vname() ]} = (");
		$c->add(
			defined($self->_field_init_values())
				? $self->_field_init_values()
				: join(', ', map(q{''}, grep(!defined($_->calc_exp()), $self->input_fields()->to_array())))
		);
		$c->newline_on();
		$c->code(");");
		return $c;
	}
	sub parse
	{
		my $self = shift;
		$self->_field_init_values(ETL::Pequel3::CodeStyler::Program::Perl->new());
		$self->_field_init_values()->code(
			join
			(
				', ', 
				map
				(
					$_->pequel_type()->name() eq 'date' 
						? $self->pequel_ref()->parser()->used_by($_)->translate(
								'(&to_date(&today()<YYYYMMDD>, ' . $_->date_type()->name() . '))'
							)->raw()
						: $_->pequel_type()->name() =~ /^(numeric|amount)$/ ? '0' : q{''}, 
					grep(!defined($_->calc_exp()), $self->input_fields()->to_array())
				)
			)
		);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::User::Find; # ???
	use base qw(Class::STL::Utilities::FunctionObject::UnaryFunction);
	use Class::STL::ClassMembers qw( dataset_spec );
	use Class::STL::ClassMembers::Constructor;
	sub function_operator
	{
		my $self = shift;
		my $arg = shift; # element object
		return $arg->dataset_spec() eq $self->dataset_spec() ? $arg : 0;
	}
}
# ----------------------------------------------------------------------------------------------------
# CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Catalogue::Abstract;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::DataSet::Abstract');
	use Class::STL::ClassMembers::Constructor;
	sub catalogue # --> to_xml
	{
		my $self = shift;
		my $xml_node = shift;
		foreach ($self->to_array()) {
			my $s_xml = $xml_node->createChild("dataset-type");
			$s_xml->attribute('dataset-name', $_->name());
			foreach my $k (keys %{$_->members()}) {
				my $attr = ${$_->members()}{$k};
				next unless UNIVERSAL::isa(${$attr}[2], 'ETL::Pequel3::Type::Properties::Abstract');
				my $m_xml = $s_xml->createChild("property-type");
				$m_xml->attribute('name', $k);
				$m_xml->attribute('default', ${$attr}[0]);
				$m_xml->attribute('validate', ${$attr}[1]);
			}
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::Catalogue;
	use base qw(ETL::Pequel3::Type::DataSet::Catalogue::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'input_datasets');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Type::DataSet::Input::AsciiDelimited->_new(),
			ETL::Pequel3::Type::DataSet::Input::Packed->_new(),
			ETL::Pequel3::Type::DataSet::Input::Dbi->_new(),
			ETL::Pequel3::Type::DataSet::Input::Sampler->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::Catalogue;
	use base qw(ETL::Pequel3::Type::DataSet::Catalogue::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'output_datasets');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Type::DataSet::Output::AsciiDelimited->_new(),
			ETL::Pequel3::Type::DataSet::Output::Packed->_new(),
			ETL::Pequel3::Type::DataSet::Output::Xml->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::UserShared::Global;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'shared_datasets'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'dataset_spec'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::DataSet::Abstract');
	use Class::STL::ClassMembers::SingletonConstructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::User::Local;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'local_datasets'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'dataset_spec'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::DataSet::Abstract');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
# FACTORY
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Factory::Abstract;
	use ETL::Pequel3::Type::Factory;
	use base qw(ETL::Pequel3::Type::Factory::Abstract); # --> Type::Factory
	use Class::STL::ClassMembers qw( default_dataset_spec shared_datasets ), # datasets --> catalogue
		Class::STL::ClassMembers::DataMember->new(name => 'factory_name', default => 'dataset'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::DataSet::Abstract');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->shared_datasets(ETL::Pequel3::Type::DataSet::UserShared::Global->new()); # Singleton
	}
	sub factory
	{
		my $self = shift;
		my %p = @_;
		$p{dataset_spec} = $self->default_dataset_spec()
			if (!defined($p{dataset_spec}) && defined($self->default_dataset_spec()));

		use Carp qw(confess);
		confess "Undefined dataset_spec parameter!" unless (defined($p{dataset_spec}));
		confess "Undefined pequel_ref parameter!" unless (defined($p{pequel_ref}));

		# 1) type:ds
		# 2) type(args):ds
		# 3) type(args):ds(args)
		# 4) type:ds(args)
		# 5) type --> type:default
		# 6) ds --> ascii:ds
		# 7) type(args) --> type:default
		# 8) ds(args) --> ascii:ds
		# 9) type:
		# 10) type(args):
		# 11) :ds
		# 12) :ds(args)
		my $type;
		my @args;
		if ($p{dataset_spec} =~ /\s*(\w+)\s*\((.*?)\)\s*:\s*(.*)$/) # 2, 3, 10
		{
			$type = $1; # must be dataset-type
			@args=('datasource_spec', $3,
				ETL::Pequel3::Parser->new(pequel_ref => $p{pequel_ref})->extract_args($2));
		}
		elsif ($p{dataset_spec} =~ /\s*(\w+)\s*:\s*(.*)$/) # 1, 4, 9
		{
			$type = $1; # must be dataset-type
			@args=('datasource_spec', $2);
		}
		elsif ($p{dataset_spec} =~ /\s*:?\s*(\w+)\s*\((.*?)\)\s*$/) # 7, 8, 12
		{
			$type = $1; # dataset or datasource type
			my $args = $2;
			if ($self->catalogue()->exists($type)) { # 7
				@args=('datasource_spec', $self->catalogue()->exists($type)->default_datasource_spec(), 
					ETL::Pequel3::Parser->new(pequel_ref => $p{pequel_ref})->extract_args($args));
			}
			else # 8
			{
				$type = 'ascii';
				@args=('datasource_spec', $p{dataset_spec});
			}
		}
		else # 5, 6, 11
		{
			$type = $p{dataset_spec}; # dataset or datasource type
			$type =~ s/^\s*://;
			if ($self->catalogue()->exists($type)) { # 5
				@args=('datasource_spec', $self->catalogue()->exists($type)->default_datasource_spec());
			}
			else # 6, 11
			{
				$type = 'ascii';
				@args=('datasource_spec', $p{dataset_spec});
			}
		}
		confess "\nUnknown dataset type $type -- [$p{dataset_spec}]!" unless ($self->catalogue()->exists($type));
		my $ds = $self->catalogue()->exists($type)->new(@args, %p); # explicit datasource_spec in @_ will take preference;
		$self->err()->trace_msg(10, "@{[ $ds->pequel_ref()->pequel_name() ]} -->"
			. "Dataset Type:" . $ds->name()
			. "; target_name:" . $ds->datasource()->target_name()
			. "; fdname:" . $ds->datasource()->fdname()
			. "; Datasource Type:" . $ds->datasource()->datasource_name()
			. "; dataset_spec:" . $ds->dataset_spec()
			. "; io_name:" . $ds->datasource()->io_name()
		);
		$self->shared_datasets()->push_back($ds) 
			if ($ds->shared() && !$self->shared_datasets()->exists($ds->dataset_spec()));
		$p{pequel_ref}->user_datasets()->push_back($ds)
			unless ($p{pequel_ref}->user_datasets()->exists($ds->dataset_spec()));
		return $ds;
	}
	sub add
	{
		my $self = shift;
		$self->push_back($self->factory(@_));
		return $self->back();
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::Factory::Abstract;
	use base qw(ETL::Pequel3::Type::DataSet::Factory::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_type', default => 'input_datasets'),
		Class::STL::ClassMembers::DataMember->new(name => 'default_dataset_spec', default => 'stdin');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::Factory::Abstract;
	use base qw(ETL::Pequel3::Type::DataSet::Factory::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_type', default => 'output_datasets'),
		Class::STL::ClassMembers::DataMember->new(name => 'default_dataset_spec', default => 'stdout');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::Factory::Local;
	use base qw(ETL::Pequel3::Type::DataSet::Input::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::Factory::Local;
	use base qw(ETL::Pequel3::Type::DataSet::Output::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Input::Factory::Global;
	use base qw(ETL::Pequel3::Type::DataSet::Input::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::SingletonConstructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSet::Output::Factory::Global;
	use base qw(ETL::Pequel3::Type::DataSet::Output::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::SingletonConstructor;
}
# ----------------------------------------------------------------------------------------------------
1;
