# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::DataSource.pm
#  Created	: 7 September 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::DataSource;
require 5.005_62;
use strict;
use warnings;
use ETL::Pequel3::Type::DataMember;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Abstract;
	use base qw(Class::STL::Element);
	use Class::STL::ClassMembers 
		qw( 
			err 
			properties 
			datasource_name 
			fdname 
			io_name
			vname
			sorter 
			shared 
			ds_vname 
			target_name
			pequel_ref
			dataset
			description
		), 
		ETL::Pequel3::Type::DataMember::User->new(name => 'record_delimiter', default => '\n');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		use ETL::Pequel3::Error;
		$self->err(ETL::Pequel3::Error->new());
		use ETL::Pequel3::Type::Properties;
		$self->properties(ETL::Pequel3::Type::Properties->new());
		$self->target_name($self->fdname()) if (defined($self->fdname()));
		$self->vname($self->fdname()) if (defined($self->fdname()));
		$self->vname
		(
			'_' 
			. uc(substr($self->io_name(),0,1)) 
			. '_' 
			. do{ my $f=uc($self->target_name());$f=~s/\W+/_/g;$f}
		) unless (defined($self->vname()));
		$self->fdname(($self->shared() ? '::' : '') . $self->vname()) unless (defined($self->fdname()));
		$self->ds_vname('_DS' . $self->vname());
		return $self;
	}
	sub _map_output
	{
		my $self = shift;
		my %p = @_;
	}
	sub _code_init
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_lock
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("use Fcntl ':flock';");
		$c->code("flock(@{[ $self->fdname() ]}, LOCK_EX);");
		return $c;
	}
	sub _code_unlock
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("flock(@{[ $self->fdname() ]}, LOCK_UN);");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Abstract; 
	use base qw(ETL::Pequel3::Type::DataSource::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'io_name', default => 'input');
	use Class::STL::ClassMembers::Constructor;
	sub _code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		defined($self->record_delimiter()) && $self->record_delimiter() ne ''
			? $c->code("local(\$/)=\"@{[ $self->record_delimiter() ]}\";")
			: $c->code("local(\$/)=undef;");
		return $c;
	}
	sub _map_output
	{
		my $self = shift;
		my %p = @_;
#<		$p{dataset}->input_fields()->map_fields($p{field_map}, $p{dataset}) #TODO: key_field ???
		$self->dataset()->input_fields()->map_fields($p{field_map}, $self->dataset()) #TODO: key_field ???
			if (exists($p{field_map}) && defined($p{field_map}));
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Abstract;
	use base qw(ETL::Pequel3::Type::DataSource::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'io_name', default => 'output');
	use Class::STL::ClassMembers::Constructor;
	sub _code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		defined($self->record_delimiter()) && $self->record_delimiter() ne ''
			? $c->code("local(\$\\)=\"@{[ $self->record_delimiter() ]}\";")
			: $c->code("local(\$\\)=undef;");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Pipe;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'The input data stream for this datasource type is I<piped> via F<STDIN>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'stdin'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'filter_cmd'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'target_name', default => 'STDIN'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'fdname', default => 'STDIN');
	use Class::STL::ClassMembers::Constructor;
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->newline_off();
		$c->code("open(@{[ $self->fdname() ]}, ");
		$c->code("'-|', ") if (defined($self->filter_cmd()) || defined($self->sorter()));
		$c->code('"');
		if (defined($self->filter_cmd())) 
		{
			$c->code("@{[ $self->filter_cmd() ]} 2>/dev/null"); # gzcat, etc
			$c->code(" | @{[ $self->sorter()->code_open()->raw() ]} 2>/dev/null")
				if (defined($self->sorter()));
		}
		elsif (defined($self->sorter()))
		{
			$c->code("@{[ $self->sorter()->code_open()->raw() ]} 2>/dev/null")
		}
		else
		{
			$c->code("-");
		}
		$c->newline_on();
		$c->code('")');
		$c->code(" || die(\"Unable to open '@{[ $self->fdname() ]}'\");");
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("close(@{[ $self->fdname() ]});");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Pipe;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'The output data stream for this datasource type is I<piped> via F<STDOUT>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'stdout'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'filter_cmd'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'target_name', default => 'STDOUT'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'fdname', default => 'STDOUT');
	use Class::STL::ClassMembers::Constructor;
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c unless (defined($self->sorter()) || defined($self->filter_cmd()));
		$c->newline_off();
		$c->code("open(@{[ $self->fdname() ]}, ");
		$c->code("'|-', ");
		$c->code('"');
		$c->code("@{[ $self->sorter()->code_open()->raw() ]} 2>/dev/null") if (defined($self->sorter()));
		$c->code(" | ") if (defined($self->sorter()) && defined($self->filter_cmd()));
		$c->code("@{[ $self->filter_cmd() ]}") if (defined($self->filter_cmd())); # gzcat, etc
		$c->newline_on();
		$c->code('")');
		$c->over();
		$c->code(" || die(\"Unable to open '@{[ $self->fdname() ]}'\");");
		$c->back();
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("close(@{[ $self->fdname() ]});");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::File;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers qw( path filepath ),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Specify an external disk file for the datasource.'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'filter_cmd'),
#?		ETL::Pequel3::Type::DataMember::User->new(name => 'target_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'file');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		my $path_delim = '/';
#>		if (index($self->target_name(), $self->os_util()->path_delim()) == -1) {
		if (defined($self->target_name())) {
		   if (index($self->target_name(), $path_delim) != -1) {
				$self->path(substr($self->target_name(), 0, rindex($self->target_name(), $path_delim)));
				$self->target_name(substr($self->target_name(), rindex($self->target_name(), $path_delim)+1));
			}
			$self->filepath((defined($self->path()) ? $self->path() . '/' : '') . $self->target_name());
		}
#<		$self->ds_vname("_DS@{[ $self->fdname() ]}") if (defined($self->fdname()));
		return $self;
	}
	sub _code_init 
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();

		$self->filepath((defined($self->path()) ? $self->path() . '/' : '') . $self->target_name()); # path may have changed;
		$c->newline_off();
		$c->code("my \$@{[ $self->ds_vname() ]} = ");
		$c->newline_on();
		($self->filepath() =~ /^\.?\//)
			? $c->code("'@{[ $self->filepath() ]}';")
			: $c->code("\${\$g_USER_OPTION\{prefix}}\[0] . '/@{[ $self->filepath() ]}';");
#<		($self->filepath() =~ /^\.?\//)
#<			? $c->code("'@{[ $self->filepath() ]}';")
#<			: $self->properties()->prefix() ne ''
#<				? $c->code("'@{[ $self->properties()->prefix() ]}/@{[ $self->filepath() ]}';")
#<				: $c->code("(\"\@{[ defined(\$ENV{PEQUEL_PREFIX}) ? \$ENV{PEQUEL_PREFIX} . '/' : '' ]}\") . '@{[ $self->filepath() ]}';");
		return $c;
	}
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("die(\"Data source file '\$@{[ $self->ds_vname() ]}'\"");
		$c->over();
			$c->code(". \" for '@{[ $self->fdname() ]}' does not exist or is unreadable!\")");
			$c->code("unless (-e \$@{[ $self->ds_vname() ]} && -r \$@{[ $self->ds_vname() ]});");
		$c->back();
		$c->newline_off();
		$c->code("open(@{[ $self->fdname() ]}, ");
		$c->code(defined($self->sorter()) || defined($self->filter_cmd()) ? "'-|', " : "'<', ");
		$c->code('"');
		if (defined($self->filter_cmd())) 
		{
			$c->code("@{[ $self->filter_cmd() ]} \$@{[ $self->ds_vname() ]} 2>/dev/null");
			$c->code(" | @{[ $self->sorter()->code_open()->raw() ]} 2>/dev/null")
				if (defined($self->sorter()));
		}
		elsif (defined($self->sorter()))
		{
			$c->code("@{[ $self->sorter()->code_open()->raw() ]} \$@{[ $self->ds_vname() ]} 2>/dev/null")
		}
		else
		{
			$c->code("\$@{[ $self->ds_vname() ]}");
		}
		$c->newline_on();
		$c->code('")');
		$c->over();
		$c->code(" || die(\"Unable to open data source file '\$@{[ $self->ds_vname() ]}'\"");
		$c->over();
		$c->code(". \" for '@{[ $self->fdname() ]}'\");");
		$c->back();
		$c->back();
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("close(@{[ $self->fdname() ]});");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::File;
	use base qw(ETL::Pequel3::Type::DataSource::Input::File);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Specify an external disk file for the datasource.'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'file'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'append', default => 0), #TODO: test
#?		ETL::Pequel3::Type::DataMember::User->new(name => 'target_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'io_name', default => 'output');
	use Class::STL::ClassMembers::Constructor;
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->newline_off();
		$c->code("open(@{[ $self->fdname() ]}, ");
		$c->code(defined($self->sorter()) || defined($self->filter_cmd()) 
			? "'|-', " 
			: "'>@{[ $self->append() ? q/>/ : q// ]}', ");
		$c->code('"');
		if (defined($self->sorter()))
		{
			$c->code("@{[ $self->sorter()->code_open()->raw() ]} 2>/dev/null");
		   	$c->code(defined($self->filter_cmd()) ? " | @{[ $self->filter_cmd() ]} 2>/dev/null >@{[ $self->append() ? q/>/ : q// ]} " : ' -o ');
		}
		elsif (defined($self->filter_cmd()))
		{
			$c->code("@{[ $self->filter_cmd() ]} 2>/dev/null >@{[ $self->append() ? q/>/ : q// ]} ")
		}
		$c->code("\$@{[ $self->ds_vname() ]}");
		$c->newline_on();
		$c->code('")');
		$c->over();
		$c->code(" || die(\"Unable to open data source file '\$@{[ $self->ds_vname() ]}'\"");
		$c->over();
		$c->code(". \" for '@{[ $self->fdname() ]}'\");");
		$c->back();
		$c->back();
		return $c;
	}
	sub _code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		defined($self->record_delimiter()) && $self->record_delimiter() ne ''
			? $c->code("local(\$\\)=\"@{[ $self->record_delimiter() ]}\";")
			: $c->code("local(\$\\)=undef;");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Dbi;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'dbi');
	use Class::STL::ClassMembers::Constructor;
	sub _code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::SqlPlus;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers qw( filter_cmd ),
		qw( 
			dsn 
			user 
			password 
			table_name 
			select_override 
			select_where_condition 
			select_fields
			select_order_by
			select_group_by
		),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'sqlplus');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::External;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Specify an external command to send (pipe) the incoming datastream.'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'filter_cmd'),
		ETL::Pequel3::Type::DataMember::User->new(name => 'target_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'external');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		my $path_delim = '/';
#<		$self->ds_vname("_DS@{[ $self->fdname() ]}") if (defined($self->fdname()));
		return $self;
	}
	sub _code_init 
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("my \$@{[ $self->ds_vname() ]} = '@{[ $self->target_name() ]}';");
		return $c;
	}
	# target_name contains the external command...
	# my _DS_EXT1 = "perl script.pl arg arg arg"
	# open(FD, "$_DS_EXT1 2>/dev/null |");
	# open(FD, "$_DS_EXT1 2>/dev/null | filter_cmd 2>/dev/null |");
	# open(FD, "$_DS_EXT1 2>/dev/null | filter_cmd 2>/dev/null | sort arg arg 2>/dev/null |");
	# open(FD, "$_DS_EXT1 2>/dev/null | sort arg arg 2>/dev/null |");
	#
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->newline_off();
		$c->code("open(@{[ $self->fdname() ]}, \"@{[ $self->ds_vname() ]} 2>/dev/null |");
		$c->code(" @{[ $self->filter_cmd() ]} 2>/dev/null |") if (defined($self->filter_cmd()));
		$c->code(" @{[ $self->sorter()->code_open()->raw() ]} 2>/dev/null |") if (defined($self->sorter()));
		$c->newline_on();
		$c->code('")');
		$c->over();
		$c->code(" || die(\"Unable to open external datasource '\$@{[ $self->ds_vname() ]}'\"");
		$c->over();
		$c->code(". \" for '@{[ $self->fdname() ]}'\");");
		$c->back();
		$c->back();
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("close(@{[ $self->fdname() ]});");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::External;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'external');
	use Class::STL::ClassMembers::Constructor;
	#TODO
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Socket;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'socket');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Socket;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'socket');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Fifo;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'fifo');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Fifo;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'fifo');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::MsgQueue;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'msg_queue');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::MsgQueue;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'msg_queue');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Void;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'void'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_name', default => 'VOID');
	use Class::STL::ClassMembers::Constructor;
	sub _code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Void;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'datasource_name', default => 'void'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_name', default => 'VOID');
	use Class::STL::ClassMembers::Constructor;
	sub _code_prepare
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_close
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
	sub _code_open
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
# CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Catalogue::Abstract;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'datasource_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::DataSource::Abstract');
	use Class::STL::ClassMembers::Constructor;
	sub catalogue # --> to_xml
	{
		my $self = shift;
		my $xml_node = shift;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Catalogue;
	use base qw(ETL::Pequel3::Type::DataSource::Catalogue::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'input_datasources');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		use ETL::Pequel3::Type::DataSource::Pequel;
		$self->push_back(
			ETL::Pequel3::Type::DataSource::Input::File->_new(),
			ETL::Pequel3::Type::DataSource::Input::Pipe->_new(),
			ETL::Pequel3::Type::DataSource::Input::Pequel->_new(), # in DataSource/Pequel.pm
			ETL::Pequel3::Type::DataSource::Input::External->_new(),
			ETL::Pequel3::Type::DataSource::Input::Dbi->_new(),
#>			ETL::Pequel3::Type::DataSource::Input::Socket->_new(),
			ETL::Pequel3::Type::DataSource::Input::Void->_new(),
#>			ETL::Pequel3::Type::DataSource::Input::Fifo->_new(),
#>			ETL::Pequel3::Type::DataSource::Input::MsgQueue->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Catalogue;
	use base qw(ETL::Pequel3::Type::DataSource::Catalogue::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'output_datasources');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		use ETL::Pequel3::Type::DataSource::Pequel;
		$self->push_back(
			ETL::Pequel3::Type::DataSource::Output::File->_new(),
			ETL::Pequel3::Type::DataSource::Output::Pipe->_new(),
			ETL::Pequel3::Type::DataSource::Output::Pequel->_new(), # in DataSource/Pequel.pm
#>			ETL::Pequel3::Type::DataSource::Output::Socket->_new(),
			ETL::Pequel3::Type::DataSource::Output::Void->_new(),
#>			ETL::Pequel3::Type::DataSource::Output::Fifo->_new(),
#>			ETL::Pequel3::Type::DataSource::Output::MsgQueue->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
# FACTORY
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Factory::Abstract;
	use ETL::Pequel3::Type::Factory;
	use base qw(ETL::Pequel3::Type::Factory::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'factory_name', default => 'datasourse'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'datasource_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::DataSource::Abstract');
	use Class::STL::ClassMembers::Constructor;
	sub factory
	{
		my $self = shift;
		use Carp qw(confess);
		my %p = @_;
		# stdin
		# stdout
		# void
		# <target_name>
		# pequel(target_name=...)
		# pequel(<pequel_name>)
		# file(<filename>)
		# socket(args...)

		my $type = exists($p{datasource_spec}) ? $p{datasource_spec} : 'void';
		if ($type =~ /\s*:?\s*(\w+)\s*\((.*?)\)\s*$/) 
		{
			$type = $1; # must be datasource-type
			$self->err()->user_error(10801, "Unknown datasource type '$type' @{[ 
				exists($p{datasource_spec}) ? 'in spec ' . $p{datasource_spec} : '' ]}")
				unless ($self->catalogue()->exists(lc($type)));
			my $args = $2;
			use ETL::Pequel3::Parser;
			my @args = ETL::Pequel3::Parser->new(pequel_ref => $p{pequel_ref})->extract_args($args);
			%p = (int(@args) == 1 ? ('target_name', @args) : @args);
		}
		return $self->catalogue()->exists(lc($type))
			? $self->catalogue()->exists(lc($type))->new(@_, %p)
			: defined($p{datasource_type})
				? $self->catalogue()->exists($p{datasource_type})->new('target_name' => $type, @_, %p)
				: $self->catalogue()->exists('file')->new('target_name' => $type, @_, %p);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Factory::Abstract;
	use base qw(ETL::Pequel3::Type::DataSource::Factory::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_type', default => 'input_datasources');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Factory::Abstract;
	use base qw(ETL::Pequel3::Type::DataSource::Factory::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_type', default => 'output_datasources');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Factory::Local;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Factory::Local;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Input::Factory::Global;
	use base qw(ETL::Pequel3::Type::DataSource::Input::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::SingletonConstructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::DataSource::Output::Factory::Global;
	use base qw(ETL::Pequel3::Type::DataSource::Output::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::SingletonConstructor;
}
# ----------------------------------------------------------------------------------------------------
1;
