# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::Factory.pm
#  Created	: 20 September 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::Factory;
require 5.005_62;
use strict;
use warnings;
use stl;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Factory::FindName;
	use base qw(Class::STL::Utilities::FunctionObject::UnaryFunction);
	use Class::STL::ClassMembers qw( name target_mem_name );
	use Class::STL::ClassMembers::Constructor;
	sub function_operator
	{
		my $self = shift;
		my $arg = shift; # element object
		my $n = $self->target_mem_name();
		return $arg->$n() eq $self->name() ? $arg : 0;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Factory::Abstract;
	use base qw(Class::STL::Containers::Deque);
	use Class::STL::ClassMembers qw( err main_catalogue catalogue catalogue_type pequel_ref factory_name ),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'Class::STL::Element');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		use ETL::Pequel3::Error;
		$self->err(ETL::Pequel3::Error->new());
		use ETL::Pequel3::Catalogue;
		$self->main_catalogue(ETL::Pequel3::Catalogue->new());
#?TODO	$self->err()->user_error(10000, "Unkown Catalogue type '@{[ $self->catalogue_type() ]}'!")
#?			unless ($self->main_catalogue()->exists($self->catalogue_type()));
		$self->catalogue($self->main_catalogue()->exists($self->catalogue_type()));
		return $self;
	}
	sub exists
	{
		my $self = shift;
		my $name = shift;
		my $obj;
		use stl qw( find_if );
		return $obj->p_element()
			if ($obj = find_if($self->begin(), $self->end(), 
				ETL::Pequel3::Type::Factory::FindName->new(name => $name, target_mem_name => $self->target_mem_name())));
		return 0;
	}
	sub factory
	{
		my $self = shift;
		my %params = @_; # named parameters;
		# virtual function
	}
	sub add
	{
		my $self = shift;
		# virtual function
	}
}
# ----------------------------------------------------------------------------------------------------
1;
