# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::General.pm
#  Created	: 8 May 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
require 5.005_62;
use strict;
use warnings;
use vars qw($VERSION $BUILD);
$VERSION = 0.01;
$BUILD = 'Friday September 22 21:19:58 BST 2006';
use ETL::Pequel3::Type::Macros;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::General::Abstract;
	use base qw(ETL::Pequel3::Type::Macros::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'category', default => 'general');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Map; #TODO: test
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'The F<&map()> macro will lookup each element within the array I<expr> in I<table> and setting that element to the looked up value. Returns an array of results. Non-existing key values will be mapped to null.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&map(E<lt> I<table> E<gt>, E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'map');
	use Class::STL::ClassMembers::Constructor;
	sub translate # -- fmt => 'table, input_field | output_field',
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("join");
		$c->open_block('(');
			$c->code("\"@{[ $self->configuration()->default_list_delimiter() ]}\",");
			$c->code("map");
			$c->open_block("(");
				$c->code("@{[ $_[0] =~ /^%/ ? '' : '%' ]}$_[0](\$_),");
				$c->code("split(\/\\s*@{[ $self->configuration()->default_list_delimiter() ]}\\s*\/,$_[1],-1)");
			$c->close_block(")");
		$c->close_block(")");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Unpack;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Unpacks the sequence of bytes in string I<expr> into a list using I<template>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&unpack(E<lt> I<template> E<gt>, E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'unpack');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		my $delim = $self->configuration()->default_list_delimiter();
		$c->code("join('$delim', unpack($_[0], $_[1]))");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Pack;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Packs the values in the array I<expr> into a sequence of bytes, using the specified I<template>. Returns the sequence as a string.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&pack(E<lt> I<template> E<gt>, E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'pack');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("pack($_[0], @{[ join(',', @_[1..$#_]) ]})");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::InputRepeatValue;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the current value for the repeating field number I<expr> as specified in the I<repeat_on_input> section.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&irval(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'irval');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("\$_IRVAL_@{[ defined($_[0]) ? $_[0] : '1' ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::OutputRepeatValue;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the current value for the repeating field number I<expr> as specified in the I<repeat_on_output> section.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&orval(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'orval');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("\$_ORVAL_@{[ defined($_[0]) ? $_[0] : '1' ]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Commify;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the comma formatted number for the value in I<expr>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&commify(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'commify');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("do");
		$c->open_block();
			$c->code("my \$idec = index($_[0], '.');");
			$c->code("my \$dec = \$idec > 0 ? substr($_[0], \$idec) : '';");
			$c->code("my \$txt = reverse(\$idec > 0 ? substr($_[0], 0, \$idec) : $_[0]);");
			$c->code("\$txt =~ s/(\\d\\d\\d)(?=\\d)(?!\\d*\\.)/\$1,/g;");
			$c->code("(scalar reverse \$txt) . \$dec;");
		$c->close_block();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Env; #TODO: test
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the value for the environment variable as specified in I<expr>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&env(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'env');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("\$ENV{$_[0]}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Configuration;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the value for the configuration variable as specified in I<expr>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&configuration(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'configuration');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		#TODO: check exists
		#TODO: remove quote delimiters
		$c->code("q{@{[ $self->configuration()->exists($_[0])->value() ]}}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::MatchAny;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns 1 if any one of the elements in array I<expr> matches the expression I<patt>. Otherwise returns 0.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&match_any(E<lt> I<patt> E<gt>, E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'match_any');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("$_[0] =~ m/^(@{[ join('|', map(do{s/^[\"']?//; s/[\"']?$//; $_; },  @_[1..$#_])) ]})\$/ ? 1 : 0");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Match;
	use base qw(ETL::Pequel3::Type::Macros::MatchAny);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns 1 if any one of the elements in array I<expr> matches the expression I<patt>. Otherwise returns 0. (same as I<&match_any()>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&match(E<lt> I<patt> E<gt>, E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'match');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Select;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Similar to a switch statement. Parameters consist of a list of expression-value pairs, followed by one default value. Each expression is evaluated in turn and the first to evaluate to true will return its associated valued, otherwise the default value is returned.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&select(E<lt> I<expr> E<gt>, E<lt> I<value> E<gt> [, E<lt> I<expr> E<gt>, E<lt> I<value> E<gt> [, ...] ], E<lt> I<default> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'select');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		my $i;
		for ($i=0; $i < $#_; $i+=2)
		{
			$c->over();
			$c->code($_[$i]);
			$c->code(" ? $_[$i+1]");
			$c->newline_off();
			$c->code(" : ");
			$c->newline_on();
		}
		$c->code($_[$i]);
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::ToNumber;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the numeric value for I<expr>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&to_number(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'to_number');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("do{ my \$tmp=$_[0]; \$tmp =~ s/.(?<![0-9])//g; \$tmp; }");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::ExtractNumeric;
	use base qw(ETL::Pequel3::Type::Macros::ToNumber);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the numeric value for I<expr>. Same as I<&to_number()>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&extract_numeric(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'extract_numeric');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::RemoveNonNumeric;
	use base qw(ETL::Pequel3::Type::Macros::ToNumber);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the numeric value for I<expr>. Same as I<&to_number()>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&remove_non_numeric(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'remove_non_numeric');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::RemoveNumeric;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Removes any numeric characters from the string in I<expr>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&remove_numeric(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'remove_numeric');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("do{ my \$tmp=$_[0]; \$tmp =~ s/\\d+//g; \$tmp; }");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::RemoveSpecial;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Removes any special (non-alphanumeric) characters from the string in I<expr>.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&remove_special(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'remove_special');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->code("do{ my \$tmp=$_[0]; \$tmp =~ s/\\W+//g; \$tmp; }");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Soundex;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The soundex macro returns a character string containing the phonetic representation of I<expr>. This function lets you compare words that are spelled differently, but sound alike in English.\n\nThe phonetic representation is defined in The Art of Computer Programming, Volume 3: Sorting and Searching, by Donald E. Knuth, as follows:\n\nRetain the first letter of the string and remove all other occurrences of the following letters: a, e, h, i, o, u, w, y.\n\nAssign numbers to the remaining letters (after the first) as follows:\n\nb, f, p, v = 1\n\nc, g, j, k, q, s, x, z = 2\n\nd, t = 3\n\nl = 4\n\nm, n = 5\n\nr = 6\n\n\n\nIf two or more letters with the same number were adjacent in the original name (before step 1), or adjacent except for any intervening h and w, then omit all but the first. Return the first four bytes padded with 0."),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&soundex(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'soundex');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		# Retain the first letter of the string and remove all other occurrences of the 
		# following letters: a, e, h, i, o, u, w, y. 
		# Assign numbers to the remaining letters (after the first) as follows: 
		# b, f, p, v = 1
		# c, g, j, k, q, s, x, z = 2
		# d, t = 3
		# l = 4
		# m, n = 5
		# r = 6
		# 
		# If two or more letters with the same number were adjacent in the original 
		# name (before step 1), or adjacent except for any intervening h and w, 
		# then omit all but the first. 
		# Return the first four bytes padded with 0. 

		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		$c->newline_off();
		$c->code("&{sub{");
		$c->code("my \$f = substr($_[0], 0, 1); ");
		$c->code("my \$rest = substr($_[0], 1); ");
		$c->code("\$rest =~ y/bfpvcgjkqsxzdtlmnraehiouwy/111122222222334556        /d; ");
		$c->code("\$rest =~ y///cs; ");	# remove consecutive duplicate chars.
		$c->code("\$rest =~ s/\\s*//g; ");
		$c->code("\$f = substr(\$f . \$rest, 0, 4); ");
		$c->code("\$f . '0' x ( 4 - length( \$f ) );  ");
		$c->newline_on();
		$c->code("}}");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Banding; #TODO: test
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'The F<&banding()> macro will return the band number (starting from 1) for I<expr>, depending on the value of I<expr> in relation to the I<band-divisor>. The I<band-divisor> must be a non zero numeric value. The returned band number is calculated as int( ( I<expr> - 1 ) / I<band-divisor> ) + 1.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&banding(E<lt> I<expr> E<gt>, E<lt> I<band-divisor> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'banding');
	use Class::STL::ClassMembers::Constructor;
	sub translate
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
#		$self->PARAM->error->fatalError("[4012] Invalid band divisor -- zero not allowed") 
#			if (!$_[1] || $_[1] == 0);
		$c->code("$_[0] eq '' || $_[0] =~ /^\\s+\$/ ? ''");
		$c->over;
		$c->code(": $_[0] == 0 ? 'U'");
		$c->over;
		$c->code(": int(($_[0] - 1) / $_[1]) + 1");
		$c->back;
		$c->back;
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::UserOption;
	use base qw(ETL::Pequel3::Type::Macros::General::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the value for the user-option as specified in I<expr>. Alternate name for the F<&option()> macro.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&user_option(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'user_option');
	use Class::STL::ClassMembers::Constructor;
	sub translate # args: user-option-name
	{
		my $self = shift;
		my $c = $self->SUPER::translate(@_);
		# TODO: unknown option...can check at generation time
		$c->code("\${\$g_USER_OPTION{$_[0]}}[0]");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::Option;
	use base qw(ETL::Pequel3::Type::Macros::UserOption);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'Returns the value for the user-option as specified in I<expr>. Alternate name for the F<&user_option()> macro.'),
		Class::STL::ClassMembers::DataMember->new(name => 'usage', default => '&option(E<lt> I<expr> E<gt>)'),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'option');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Macros::General;
	use ETL::Pequel3::Type::Extension;
	use base qw(ETL::Pequel3::Type::Extension::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->catalogue()->macros()->push_front(
        	ETL::Pequel3::Type::Macros::Map->new(),
        	ETL::Pequel3::Type::Macros::Unpack->new(),
        	ETL::Pequel3::Type::Macros::Pack->new(),
        	ETL::Pequel3::Type::Macros::Commify->new(),
        	ETL::Pequel3::Type::Macros::Env->new(),
        	ETL::Pequel3::Type::Macros::Configuration->new(),
        	ETL::Pequel3::Type::Macros::MatchAny->new(),
        	ETL::Pequel3::Type::Macros::Match->new(),
        	ETL::Pequel3::Type::Macros::Select->new(),
        	ETL::Pequel3::Type::Macros::ToNumber->new(),
        	ETL::Pequel3::Type::Macros::ExtractNumeric->new(),
        	ETL::Pequel3::Type::Macros::RemoveNonNumeric->new(),
        	ETL::Pequel3::Type::Macros::RemoveNumeric->new(),
        	ETL::Pequel3::Type::Macros::RemoveSpecial->new(),
        	ETL::Pequel3::Type::Macros::Soundex->new(),
        	ETL::Pequel3::Type::Macros::Banding->new(),
        	ETL::Pequel3::Type::Macros::UserOption->new(),
        	ETL::Pequel3::Type::Macros::Option->new(),
        	ETL::Pequel3::Type::Macros::InputRepeatValue->new(), # &irval(<n>) -- return value for repeat #n;
        	ETL::Pequel3::Type::Macros::OutputRepeatValue->new(), # &orval(<n>) -- return value for repeat #n;
#>			greatest(value, ...) -- return greatest value from list
#>			least(value, ...)
#>COALESCE(value [, ...])The COALESCE function returns the first of its arguments that is not null. Null is returned only if all arguments are null. It is often used to substitute a default value for null values when data is retrieved for display, for example:
#>
#>SELECT COALESCE(description, short_description, '(none)') ...
#>
#>NULLIF(value1, value2)The NULLIF function returns a null value if value1 and value2 are equal; otherwise it returns value1. This can be used to perform the inverse operation of the COALESCE example given above:
#>
#>SELECT NULLIF(value, '(none)') ...
#>If value1 is (none), return a null, otherwise return value1. 

#>md5(string text) text Calculates the MD5 hash of string, returning the result in hexadecimal  md5('abc') 900150983cd24fb0 d6963f7d28e17f72 
#>repeat(string text, number int) text Repeat string the specified number of times repeat('Pg', 4) PgPgPgPg 

		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
1;
