# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::Section::FieldType.pm
#  Created	: 22 June 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::Section::FieldType;
require 5.005_62;
use strict;
use warnings;
use ETL::Pequel3::Type::Section;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Section::FieldType;
	use base qw(ETL::Pequel3::Type::Section::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'section_name'),
		Class::STL::ClassMembers::DataMember->new(name => 'section_name', default => 'field_type'),
		Class::STL::ClassMembers::DataMember->new(name => 'item_type', default => 'ETL::Pequel3::Type::Section::FieldType::Item'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => "@{[ __PACKAGE__ ]}"),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => 'This section is used to create additional data types. This information is relevant to the dataset F<verify> property -- the generated verification code will use the regular expression in F<validate> to validate the date against.');
	use Class::STL::ClassMembers::Constructor;
	sub prepare
	{
		my $self = shift;
		foreach my $i ($self->items()->to_array()) {
			$self->pequel_ref()->catalogue()->pequel_types()->push_front(
				$self->pequel_ref()->catalogue()->pequel_types()->factory(
					name => $i->field_type_name(), validate => $i->validate())
			);
			$self->err()->trace_msg(10, "@{[ $self->pequel_ref()->pequel_name() ]} -->"
				. "Field Type:"
				. $i->field_type_name()
				. "; validate:" 
				. $i->validate()
			);
		}
	}
	package ETL::Pequel3::Type::Section::FieldType::Item; # extend pequel_type
	use base qw(ETL::Pequel3::Type::Section::Item::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->attributes()->push_back(
			$self->attributes()->factory(name => 'field_type_name', required => 1),
			$self->attributes()->factory(name => 'validate', required => 1),
#?			$self->attributes()->factory(name => 'default', required => 0),
#?			$self->attributes()->factory(name => 'min_len', required => 0),
#?			$self->attributes()->factory(name => 'max_len', required => 0),
#?			$self->attributes()->factory(name => 'random_min', required => 0), # for sample dataset
#?			$self->attributes()->factory(name => 'random_max', required => 0), # for sample dataset
		);
	}
}
# ----------------------------------------------------------------------------------------------------
1;
