# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::Sorters.pm
#  Created	: 16 June 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::Sorters;
require 5.005_62;
use strict;
use warnings;
use ETL::Pequel3::Type::Field;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Sorters::Abstract;
	use base qw(Class::STL::Element);
	use Class::STL::ClassMembers 
		qw( 
			err
			catalogue
			fields 
			properties 
			pequel_types 
			sorter_name 
			field_delimiter
			uniq
			ignore_case
			pequel_type
			pequel_ref
		);
	use Class::STL::ClassMembers::Constructor;
	use stl qw(list find_if);
	sub new_extra
	{
		my $self = shift;
		$self->err(ETL::Pequel3::Error->new());
		$self->catalogue(ETL::Pequel3::Catalogue->new());
		$self->properties(ETL::Pequel3::Type::Properties->new());
		$self->fields($self->catalogue()->fieldsets()->sort_fields()->new(pequel_ref => $self->pequel_ref()));
		$self->pequel_types(ETL::Pequel3::Type::PequelTypes::Catalogue->new());
		$self->pequel_type($self->pequel_types()->exists('string')) unless (defined($self->pequel_type()));
		return $self;
	}
	sub add
	{
		my $self = shift;
		$self->fields()->add(@_);
	}
	sub exists
	{
		my $self = shift;
		my $name = shift;
		my $s;
		return $s->p_element()
			if ($s = find_if($self->fields()->begin(), $self->fields()->end(), 
				ETL::Pequel3::Type::DataSets::FindName->new(name => $name)));
		return 0;
	}
	sub code_open
	{
		# virtual must override
		# return Class::CodeStyler::Program object;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Sorters::Unix;
	use base qw( ETL::Pequel3::Type::Sorters::Abstract );
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'field_delimiter', default => '|'), 
		Class::STL::ClassMembers::DataMember->new(name => 'uniq', default => 0), 
		Class::STL::ClassMembers::DataMember->new(name => 'ignore_case', default => 0), 
		Class::STL::ClassMembers::DataMember->new(name => 'sorter_name', default => 'unix');
	use Class::STL::ClassMembers::Constructor;
	sub code_open
	{
		my $self = shift;
		my $c = Class::CodeStyler::Program::Perl->new();
		$c->newline_off();
		$c->code($self->properties()->exists('sort_cmd')->value());
		$c->code(' ' . $self->properties()->exists('sort_args')->value())
			if ($self->properties()->exists('sort_args')->value() ne '');
		$c->code(" -u") if ($self->uniq());
		$c->code(" -t'@{[ $self->field_delimiter() ]}'") 
			if (defined($self->field_delimiter()) && $self->field_delimiter() ne '');
		map($c->code($_->getvar()), $self->fields()->to_array());
		$c->code(' -T ' . $self->properties()->exists('sort_tmp_dir')->value())
			if ($self->properties()->exists('sort_tmp_dir')->value() ne '');
#?		$c->code(" 2>/dev/null");
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
# CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Sorters::Catalogue;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'sorters'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'sorter_name');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Type::Sorters::Unix->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
1;
