# vim:ts=4 sw=4
# ----------------------------------------------------------------------------------------------------
#  Name		: ETL::Pequel3::Type::Table.pm
#  Created	: 2 June 2006
#  Author	: Mario Gaffiero (gaffie)
#
# Copyright 1999-2007 Mario Gaffiero.
# 
# This file is part of Pequel(TM).
# 
# Pequel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# Pequel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Pequel; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# ----------------------------------------------------------------------------------------------------
# Modification History
# When          Version     Who     What
# ----------------------------------------------------------------------------------------------------
package ETL::Pequel3::Type::Table;
require 5.005_62;
use strict;
use warnings;
use ETL::Pequel3::CodeStyler;
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::Abstract;
	use base qw(Class::STL::Element);
	use Class::STL::ClassMembers 
		qw( 
			catalogue
			err 
			properties 
			pequel_types 
			sorters
			table_number 
			fields 
			tname 
			field_map
			key_field 
			verify
			configuration
			used_by_list 
			use_list
			package_name
			load_function_name
			load_name 
			pequel_ref

			name 
			description
		), 
		Class::STL::ClassMembers::DataMember->new(name => 'xs_type', default => 'NMTOKEN'),
		Class::STL::ClassMembers::DataMember->new(name => 'fieldset_type', default => 'table_fields'),
		Class::STL::ClassMembers::DataMember->new(name => 'key_type', default => 'string'); #??? not used ???
	use Class::STL::ClassMembers::Constructor;
	use ETL::Pequel3::Error;
	sub new_extra
	{
		my $self = shift;
		use ETL::Pequel3::Catalogue;
		$self->catalogue(ETL::Pequel3::Catalogue->new());
		$self->err(ETL::Pequel3::Error->new()); 
		$self->properties(ETL::Pequel3::Type::Properties->new());
		defined($self->pequel_ref()->config())
			? $self->configuration($self->pequel_ref()->config())
			: $self->configuration(ETL::Pequel3::Type::Properties::User->new());
		$self->pequel_types(ETL::Pequel3::Type::PequelTypes::Catalogue->new());
		$self->sorters(ETL::Pequel3::Type::Sorters::Catalogue->new());

		$self->tname('_T_' . $self->name());
		use ETL::Pequel3::Type::FieldSet;
		$self->fields($self->catalogue()->fieldsets()->exists($self->fieldset_type())
			->new(table => $self, pequel_ref => $self->pequel_ref()));
		$self->used_by_list(stl::list());
		$self->use_list(stl::list());
		$self->package_name(ref($self));
		$self->load_function_name("load@{[ $self->tname() ]}");
		$self->load_name($self->package_name() . '::' . $self->load_function_name());
		return $self;
	}
	sub factory
	{
		my $self = shift;
		my %p = @_; # table_type, static, input_dataset, dataset_spec, ...
		return $self->new(%p);
	}
	sub prepare
	{
		my $self = shift;
		my $items = shift;
	}
	sub code_load_table
	{
		# virtual function
	}
	sub code_start
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		$c->code("our \$@{[ $self->tname() ]} =");
		$c->over();
			$c->code("\&@{[ $self->load_name() ]}();");
			$c->code("# ref to \%@{[ $self->tname() ]} hash");
		$c->back();
		return $c;
	}
	sub translate
	{
		my $self = shift;
		my $key_exp = shift; 
		my $field = shift; # || undef;
		my $show_synonyms = shift;
		my $target_obj = shift; #TODO: not best solution?
		my $c = ETL::Pequel3::CodeStyler::Program::Perl->new();
		if (defined($field) && !ref($field)) 
		{
			$field =~ s/^@{[ $self->name() ]}->//;
			my $f = $self->fields()->exists($field); 
			$self->err()->user_error(10302, "Unknown field '$field' for table '@{[ $self->name() ]}'")
				unless (defined($f) && ref($f));
			$field = $f;
		}
		if (defined($field)) 
		{
#>			$c->code($field->getvar($key_exp, $show_synonyms, $target_obj)); # update ref's to key-fields
			$c->code($field->getvar($key_exp, $show_synonyms));
			$target_obj->xref()->references($field) if (defined($target_obj) && $target_obj->can('xref'));
			$field->xref()->referenced_by($target_obj) if (defined($target_obj) && $field->can('xref'));
		} 
		else 
		{
			# !! this should never happen -- a field name is required !!
			# keyfld can be an expression eg: another macro
			my $key_eval = $key_exp =~ /[&\(|\[|,]/ ? "\@{[ $key_exp ]}" : $key_exp;
			$c->code('$$' . $self->tname() . '{qq{' . $key_eval . '}}');
		}
		return $c;
	}
	sub code_display_info
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		my $extra = shift;
		$c->comment("++++++ Table @{[ $self->name() ]} --> Type @{[ ref($self) ]} ++++++");
		$c->comment("++++++ $extra") if (defined($extra));
		return $c;
	}
	sub attach_sorter #--> External::Abstract ?? --> move to dataset
	{
		my $self = shift;
		my $key_type = shift;
		# TODO:Need to verify key_field exists!
		$self->input_dataset()->datasource()->sorter(
			$self->sorters()->exists($self->properties()->sorter())->new(
				uniq => 1,
				field_delimiter => $self->input_dataset()->can('field_delimiter')
					? $self->input_dataset()->field_delimiter() : undef,
				pequel_ref => $self->pequel_ref()
			)
		);
#<		# TODO: dataset-type to determine sorter: eg packed type no sorter...
		$self->input_dataset()->datasource()->sorter()->add(
			name => $self->key_field(),
			column => $self->fields()->exists($self->key_field())->ds_column(),
			pequel_type => $key_type,
			pequel_ref => $self->pequel_ref(),
		);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::Local::Data;
	use base qw(Class::STL::Element);
	use Class::STL::ClassMembers qw( key value );
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->value([]) unless defined($self->value());
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::Local;
	use base qw(ETL::Pequel3::Type::Table::Abstract);
	use Class::STL::ClassMembers qw( data ),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The F<local> table type can be used to create a table with the data declared within the I<pequel script>. The table data is declared within the I<items> with the F<key>, F<value> attribute pairs. The F<value> attribute consists of string of space delimited data values"),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'local');
	use Class::STL::ClassMembers::Constructor;
	sub new_extra
	{
		my $self = shift;
		$self->data(stl::list(element_type => 'ETL::Pequel3::Type::Table::Local::Data'));
		return $self;
	}
	sub factory
	{
		my $self = shift;
		my %p = @_; # table_type, static, input_dataset, dataset_spec, ...
		return $p{table_type} eq 'local' ? $self->new(%p) : 0;
	}
	sub code_load_table
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program::Perl->new();
		return if ($c->jump("load@{[ $self->tname() ]}")); # ???
		$c->divider();
		$self->code_display_info($c, "@{[ __PACKAGE__ ]}");
		$c->open_block("{");
			$c->code("package " . $self->package_name() . ";");
			$c->code("sub @{[ $self->load_function_name() ]}");
			$c->open_block("{");
				$c->code("my \%@{[ $self->tname() ]} =");
				$c->code("(");
				$c->over();
					my $count=0;
					foreach my $d (sort { $a->key() cmp $b->key() } $self->data()->to_array())
					{
						$c->newline_on();
						$c->indent_off();
						$c->code(',') if (++$count > 1);
						$c->indent_on();
						$c->newline_off();
						$c->code("'@{[ $d->key() ]}' => ");
						($self->fields()->size() == 0)
							? $c->code("'@{[ $d->key() ]}'")
							: ($self->fields()->size() == 1)
								? $c->code("'@{[ ${$d->value()}[0] ]}'")
								: $c->code("['@{[ join(q{', '}, map($_, map(defined($_) ? $_ : '', @{$d->value()}))) ]}']");
					}
					$c->newline_on();
					$c->code("");
				$c->back();
				$c->code(");");
				$c->code('return \%' . $self->tname() . ';');
			$c->close_block();
		$c->close_block();
		return $c;
	}
	sub prepare
	{
		my $self = shift;
		my $items = shift;

		$self->fields()->map_fields($self->field_map());
		# Copy the data:
		foreach (grep(defined($_->key_value()), $items->to_array())) {
			$self->data()->push_back($self->data()->factory(
				key => $_->key_value(),
				value => $_->data_values()
			));
		}
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::External::Abstract;
	use base qw(ETL::Pequel3::Type::Table::Local);
	use Class::STL::ClassMembers qw( input_dataset ),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external I<dataset>. This table type can be loaded I<dynamically> or I<statically>.\n\n\nThe advantage of the F<static> type over the F<dynamic> type is (1) there is no wait, at the start of execution, for the table data to load; (2) the table-data is combined with the generated program so will execute without any dependence on other external files. The disadvantages with F<static> is (1) the program size, especially with large data volumes, and (2) the program will need to be re-generated whenever the data needs to be refreshed.\n\n\nF<Dynanic Table>\n\nThe data for a dynamic type table is loaded at run-time, that is, when the I<generated> program executes.\n\n\nF<Static Table>\n\nWith static tables the data is pre-loaded during the code-generation resulting with the table-data embedded within the generated program (as a I<hash> variable)."),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'external');
	use Class::STL::ClassMembers::Constructor;
	sub factory
	{
		my $self = shift;
		my %p = @_; # table_type, static, dataset_spec, ...
		return 
		(
			defined($p{input_dataset}) 
			&& 
			(
				$p{table_type} eq $self->name()
				|| 
				(
					$p{table_type} eq 'external'
					&& 
					( 
						($self->isa('ETL::Pequel3::Type::Table::ExternalStatic') && $p{static})
						|| 
						(!$self->isa('ETL::Pequel3::Type::Table::ExternalStatic') && !$p{static})
					)
				)
			)
		) 
		? $self->new(%p) : 0;
	}
	sub prepare
	{
		my $self = shift;
		my $items = shift;

		my $key_type = $self->key_type() || 'string';
		$key_type = $self->pequel_types()->exists($key_type)
			|| $self->err()->user_error(10210, "Unkown pequel-type '@{[ $key_type ]}' for table '@{[ $self->name() ]}' key_field");
	
		$self->err()->user_error(10213, "Key field not defined for table '@{[ $self->name() ]}'")
			unless (defined($self->key_field()));

		# Disable discard_header because sort will shift it:
		$self->input_dataset()->discard_header(0) if ($self->input_dataset()->can('discard_header'));

		$self->input_dataset()->map_output(field_map => $self->field_map(), key_field => $self->key_field());
		$self->err()->user_error(10214, "Table @{[ $self->name() ]}->key field '@{[ 
				$self->key_field() ]}' does not exist in source '@{[ 
				$self->input_dataset()->datasource()->target_name()
			]}'")
			unless (grep($self->key_field() eq $_->name(), $self->fields()->to_array()));
		$self->attach_sorter($key_type);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::ExternalStatic; # preload table data into generated code.
	use base qw(ETL::Pequel3::Type::Table::External::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external I<dataset>. With static tables the data is pre-loaded during the code-generation resulting with the table-data embedded within the generated program (as a I<hash> variable)."),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'external_static');
	use Class::STL::ClassMembers::Constructor;
	sub code_load_table
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		our %__loaded;

		my $preload_function = "preload@{[ $self->tname() ]}";

		# Prevent re-evaling preload function when same static table declared more 
		# than once (as in some tests)
		return $self->SUPER::code_load_table($c) if (++$__loaded{$preload_function} > 1);

		my $ctmp = ETL::Pequel3::CodeStyler::Program->new();
		$ctmp->open_block();
		$ctmp->code("package " . $self->package_name() . ";");
		$ctmp->code("sub $preload_function");
		$ctmp->open_block();
		$ctmp->code("my \$self = shift;");
		$ctmp->code("my \$data = shift;");
		$self->input_dataset()->code_init($ctmp);
		$self->catalogue()->code_segments()->verbose_msg()->new(
			target_name => $self->properties()->script_name(),
			user_prog => $ctmp,
			msg_text => "Pre-loading static table @{[ $self->name() ]} from @{[ 
				$self->input_dataset()->datasource()->target_name() ]}..."
		)->prepare() if ($self->configuration()->verbose()); 
		$self->input_dataset()->code_open($ctmp);
		$self->input_dataset()->code_prepare($ctmp);
		$ctmp->newline_off();
		$ctmp->code("while (");
		$self->input_dataset()->code_read($ctmp, $self->fields()->to_array());
		$ctmp->newline_on();
		$ctmp->code(")");
		$ctmp->open_block();
			$self->input_dataset()->code_after_read($ctmp, $self->fields()->to_array());
			$ctmp->newline_off();
			$ctmp->code("\$data->push_back(\$data->factory(");
			$ctmp->code("key => \$@{[ $self->input_dataset()->arr_vname() ]}\[@{[ 
				$self->fields()->exists($self->key_field())->ds_column() -1 ]}],");
			$ctmp->code("value => [ ");
			$self->fields()->size() == 0
				? $ctmp->code("1")
				: $ctmp->code("\@@{[ $self->input_dataset()->arr_vname() ]}\[ 
					@{[ join(',', map($_->ds_column() -1, $self->fields()->to_array())) ]} ] "); # includes key as well
			$ctmp->code("]");
			$ctmp->code(")"); # close factory
			$ctmp->newline_on();
			$ctmp->code(");"); # close push_back
			$self->catalogue()->code_segments()->verbose_msg()->new(
				target_name => $self->properties()->script_name(),
				user_prog => $ctmp,
				msg_text => "Table @{[ $self->name ]} \$. records...", 
				cond => "if (\$. % 100000 == 0)"
			)->prepare() if ($self->configuration()->verbose()); 
		$ctmp->close_block(); # while
		$self->catalogue()->code_segments()->verbose_msg()->new(
			target_name => $self->properties()->script_name(),
			user_prog => $ctmp,
			msg_text => "Table @{[ $self->name() ]} pre-loaded \$. records."
		)->prepare() if ($self->configuration()->verbose()); 
		$self->input_dataset()->code_close($ctmp);
		$ctmp->close_block(); # sub
		$ctmp->close_block(); # package
		if ($self->input_dataset()->datasource()->datasource_name() eq 'pequel') {
			my $pname = "Pequel::@{[ $self->input_dataset()->datasource()->target_name() ]}\::execute";
			undef(&{$pname});
			my $sub_code = ETL::Pequel3::Generator::SubProgram->new(
				main => $self->input_dataset()->datasource()->pequel_ref());
			$sub_code->build();
			$ctmp->add($sub_code->user_prog());
		}
		$ctmp->prepare();
		$self->err()->trace_msg(10, $ctmp->print());
		$ctmp->eval();
		$self->data()->clear();
		$self->$preload_function($self->data());
		return $self->SUPER::code_load_table($c);
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::ExternalDynamic; # load table at runtime.
	use base qw(ETL::Pequel3::Type::Table::External::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external I<dataset>. The data for a dynamic type table is loaded at run-time, that is, when the I<generated> program executes."),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'external_dynamic');
	use Class::STL::ClassMembers::Constructor;
	sub code_load_table
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new(); 
		$c->divider();
		$self->code_display_info($c, "@{[ __PACKAGE__ ]}");
		$c->open_block("{");
		$c->code("package " . $self->package_name() . ";");
		$c->code("sub @{[ $self->load_function_name() ]}");
		$c->open_block("{");
		$self->input_dataset()->code_init($c);
		$self->catalogue()->code_segments()->verbose_msg()->new(
			target_name => $self->properties()->script_name(),
			user_prog => $c,
			msg_text => "Loading table @{[ $self->name() ]} from @{[ 
				$self->input_dataset()->datasource()->target_name() 
			]}..."
		)->prepare() if ($self->configuration()->verbose()); 
		$self->input_dataset()->code_open($c);
		$self->input_dataset()->code_prepare($c);
		$c->code("my \%@{[ $self->tname() ]};");
		$c->newline_off();
		$c->code("while (");
		$self->input_dataset()->code_read($c, $self->fields()->to_array());
		$c->newline_on();
		$c->code(")");
		$c->open_block("{");
			$self->input_dataset()->code_after_read($c, $self->fields()->to_array());
			$c->newline_off();
			$c->code("\$@{[ $self->tname() ]}\{\$@{[ 
					$self->input_dataset()->arr_vname() 
				]}\[@{[ 
					$self->fields()->exists($self->key_field())->ds_column() -1 
				]}]} = ");

			$c->newline_on();
			($self->fields()->size() == 0)
				? $c->code("'1',")
				: $c->code("[ \@@{[ $self->input_dataset()->arr_vname() ]}\[ @{[ 
					join(',', map($_->ds_column() -1, $self->fields()->to_array())) 
				]} ] ];");

			$self->catalogue()->code_segments()->verbose_msg()->new(
				target_name => $self->properties()->script_name(),
				user_prog => $c,
				msg_text => "Table @{[ $self->name ]} \$. records...", 
				cond => "if (\$. % 100000 == 0)"
			)->prepare() if ($self->configuration()->verbose()); 
		$c->close_block();
		$self->catalogue()->code_segments()->verbose_msg()->new(
			target_name => $self->properties()->script_name(),
			user_prog => $c,
			msg_text => "Table @{[ $self->name() ]} loaded \$. records."
		)->prepare() if ($self->configuration()->verbose()); 
		$self->input_dataset()->code_close($c);
		$c->code('return \%' . $self->tname() . ';');
		$c->close_block();
		$c->close_block();
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::Dbi::Abstract;
	use base qw(ETL::Pequel3::Type::Table::External::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'dbi'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external F<dbi> I<dataset>. This table type can be loaded I<dynamically> or I<statically>. Please refer to the description for F<external> for further information regarding F<static> and F<dynamic>."),
		Class::STL::ClassMembers::DataMember->new(name => 'fieldset_type', default => 'table_fields_dbi');
	use Class::STL::ClassMembers::Constructor;
	sub factory
	{
		my $self = shift;
		my %p = @_; # table_type, static, dataset_spec, ...
		return 
		(
			defined($p{input_dataset}) 
			&& 
			(
				$p{table_type} eq $self->name()
				|| 
				(
					$p{table_type} eq 'dbi'
					&& 
					( 
						($self->isa('ETL::Pequel3::Type::Table::ExternalStatic') && $p{static})
						|| 
						(!$self->isa('ETL::Pequel3::Type::Table::ExternalStatic') && !$p{static})
					)
				)
			)
		) 
		? $self->new(%p) : 0;
	}
	sub attach_sorter
	{
		my $self = shift;
		my $key_type = shift;
#>		my $key_field = shift;
#>		$self->input_dataset()->select_order_by($self->key_field());
	}
}
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Type::Table::DbiStatic;
	use base qw(ETL::Pequel3::Type::Table::Dbi::Abstract);
	use base qw(ETL::Pequel3::Type::Table::ExternalStatic);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'dbi_static'),
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external F<dbi> I<dataset>. With static tables the data is pre-loaded during the code-generation resulting with the table-data embedded within the generated program (as a I<hash> variable)."),
		Class::STL::ClassMembers::DataMember->new(name => 'fieldset_type', default => 'table_fields');
	use Class::STL::ClassMembers::Constructor;
	sub code_load_table
	{
		my $self = shift;
		return $self->ETL::Pequel3::Type::Table::ExternalStatic::code_load_table(@_);
	}
}
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Type::Table::DbiDynamic;
	use base qw(ETL::Pequel3::Type::Table::Dbi::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external F<dbi> I<dataset>. The data for a dynamic type table is loaded at run-time, that is, when the I<generated> program executes."),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'dbi_dynamic');
	use Class::STL::ClassMembers::Constructor;
	sub code_load_table
	{
		my $self = shift;
		my $c = shift || ETL::Pequel3::CodeStyler::Program->new();
		$c->divider();
		$self->code_display_info($c, "@{[ __PACKAGE__ ]}");
		$c->open_block("{");
			$c->code("package " . $self->package_name() . ";");
			$c->code("sub @{[ $self->load_function_name() ]}");
			$c->open_block("{");
				$self->input_dataset()->code_init($c);
				$self->catalogue()->code_segments()->verbose_msg()->new(
					target_name => $self->properties()->script_name(),
					user_prog => $c,
					msg_text => "Loading table @{[ $self->name() ]} from @{[ $self->input_dataset()->dsn() ]}..."
				)->prepare() if ($self->configuration()->verbose()); 
				$self->input_dataset()->code_open($c);
				$self->input_dataset()->code_prepare($c);
				$c->newline_off();
				$c->code("my \$@{[ $self->tname() ]} = "); # hash_ref
				$c->newline_on();
				$c->code("\$@{[ $self->input_dataset()->sth_vname() ]}->fetchall_hashref('@{[ $self->key_field() ]}');");
				$self->catalogue()->code_segments()->verbose_msg()->new(
					target_name => $self->properties()->script_name(),
					user_prog => $c,
					msg_text => "Table @{[ $self->name() ]} loaded \@{[ scalar(keys(\%\$@{[ $self->tname() ]})) ]} records."
				)->prepare() if ($self->configuration()->verbose()); 
				$self->input_dataset()->code_close($c);
				$c->code("return \$@{[$self->tname() ]};");
			$c->close_block();
		$c->close_block();
		return $c;
	}
	sub translate
	{
		my $self = shift;
		my $key_exp = shift; 
		my $field = shift; # || undef;
		my $show_synonyms = shift;
		my $target_obj = shift; #TODO: not best solution?
		my $c = ETL::Pequel3::CodeStyler::Program::Perl->new();
		if (defined($field) && !ref($field)) {
			$field =~ s/^.*\-\>//;
			my $f = $self->fields()->exists($field); 
			$self->err()->user_error(10307, "Unknown field '$field' for table '@{[ $self->name() ]}'")
				unless (defined($f) && ref($f));
			$field = $f;
		}
		if (defined($field)) {
			$c->code($field->getvar($key_exp));
			$target_obj->xref()->references($field) if ($target_obj->can('xref'));
			$field->xref()->referenced_by($target_obj) if ($field->can('xref'));
		} else {
			# keyfld can be an expression eg: another macro
			my $key_eval = $key_exp =~ /[&\(|\[|,]/ ? "\@{[ $key_exp ]}" : $key_exp;
			$c->code('$' . $self->tname() . '->{qq{' . $key_eval . '}}');
		}
		return $c;
	}
}
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Type::Table::Pequel::Abstract;
	use base qw(ETL::Pequel3::Type::Table::External::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external F<pequel> I<dataset>. This table type can be loaded I<dynamically> or I<statically>. Please refer to the description for F<external> for further information regarding F<static> and F<dynamic>."),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'pequel');
	use Class::STL::ClassMembers::Constructor;
	sub factory
	{
		my $self = shift;
		my %p = @_; # table_type, static, dataset_spec, ...
		return 
		(
			defined($p{input_dataset}) 
			&& 
			(
				$p{table_type} eq $self->name()
				|| 
				(
#?					$p{table_type} eq 'pequel'
					$p{input_dataset}->datasource()->datasource_name() eq 'pequel'
					&& 
					( 
						($self->isa('ETL::Pequel3::Type::Table::ExternalStatic') && $p{static})
						|| 
						(!$self->isa('ETL::Pequel3::Type::Table::ExternalStatic') && !$p{static})
					)
				)
			)
		) 
		? $self->new(%p) : 0;
	}
	sub prepare
	{
		my $self = shift;
		my $items = shift; # local table data items

		$self->err()->user_error(10213, "Key field not defined for table '@{[ $self->name() ]}'")
			unless (defined($self->key_field()));

		$self->input_dataset()->map_output(main => $self->pequel_ref(), field_map => $self->field_map());

		$self->input_dataset()->datasource()->pequel_ref()->output()->output_dataset()->datasource()->sorter(
			$self->input_dataset()->datasource()->sorter());

		$self->err()->user_error(10214, "Key field '@{[ 
				$self->key_field() ]}' does not exist in pequel '@{[ 
				$self->input_dataset()->datasource()->pequel_ref()->pequel_name() 
			]}'")
			unless (grep($self->key_field() eq $_->name(), 
					$self->input_dataset()->datasource()->pequel_ref()->output_interface()));

		$self->key_type($self->input_dataset()->datasource()->pequel_ref()->output()->output_fields()->exists($self->key_field())->pequel_type()->name())
			unless (defined($self->key_type()));
		my $key_type = $self->key_type();
		$key_type = $self->pequel_types()->exists($key_type)
			|| $self->err()->user_error(10210, "Unkown pequel-type '@{[ $key_type ]}' for table '@{[ $self->name() ]}' key_field");
		$self->attach_sorter($key_type);
	}
}
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Type::Table::PequelStatic;
	use base qw(ETL::Pequel3::Type::Table::Pequel::Abstract);
	use base qw(ETL::Pequel3::Type::Table::ExternalStatic);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external F<pequel> I<dataset>. With static tables the data is pre-loaded during the code-generation resulting with the table-data embedded within the generated program (as a I<hash> variable)."),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'pequel_static');
	use Class::STL::ClassMembers::Constructor;
	sub code_load_table
	{
		my $self = shift;
		return $self->ETL::Pequel3::Type::Table::ExternalStatic::code_load_table(@_);
	}
}
# ----------------------------------------------------------------------------------------------------
{	
	package ETL::Pequel3::Type::Table::PequelDynamic;
	use base qw(ETL::Pequel3::Type::Table::Pequel::Abstract);
	use base qw(ETL::Pequel3::Type::Table::ExternalDynamic);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'description', default => "The data for this table type is loaded from an external F<pequel> I<dataset>. The data for a dynamic type table is loaded at run-time, that is, when the I<generated> program executes."),
		Class::STL::ClassMembers::DataMember->new(name => 'name', default => 'pequel_dynamic');
	use Class::STL::ClassMembers::Constructor;
	sub code_load_table
	{
		my $self = shift;
		return $self->ETL::Pequel3::Type::Table::ExternalDynamic::code_load_table(@_);
	}
}
# ----------------------------------------------------------------------------------------------------
# CATALOGUE
# ----------------------------------------------------------------------------------------------------
{
#?	package ETL::Pequel3::Catalogue::Tables::Abstract;
	package ETL::Pequel3::Type::Table::Catalogue::Abstract;
	use base qw(ETL::Pequel3::Type::Catalogue);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::Table::Abstract');
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
#?	package ETL::Pequel3::Catalogue::Tables;
	package ETL::Pequel3::Type::Table::Catalogue;
	use base qw(ETL::Pequel3::Type::Table::Catalogue::Abstract);
	use Class::STL::ClassMembers
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_name', default => 'table_types');
	use Class::STL::ClassMembers::SingletonConstructor;
	sub new_extra
	{
		my $self = shift;
		$self->push_back(
			ETL::Pequel3::Type::Table::Local->_new(),
			ETL::Pequel3::Type::Table::PequelStatic->_new(),
			ETL::Pequel3::Type::Table::PequelDynamic->_new(),
			ETL::Pequel3::Type::Table::Pequel::Abstract->_new(),
			ETL::Pequel3::Type::Table::DbiStatic->_new(),
			ETL::Pequel3::Type::Table::DbiDynamic->_new(),
			ETL::Pequel3::Type::Table::Dbi::Abstract->_new(),
			ETL::Pequel3::Type::Table::ExternalStatic->_new(),
			ETL::Pequel3::Type::Table::ExternalDynamic->_new(),
			ETL::Pequel3::Type::Table::External::Abstract->_new(),
		);
		return $self;
	}
}
# ----------------------------------------------------------------------------------------------------
# FACTORY
# ----------------------------------------------------------------------------------------------------
{
#?	package ETL::Pequel3::Factory::Tables::Abstract;
	package ETL::Pequel3::Type::Table::Factory::Abstract;
	use ETL::Pequel3::Type::Factory;
	use base qw(ETL::Pequel3::Type::Factory::Abstract);
	use Class::STL::ClassMembers 
		Class::STL::ClassMembers::DataMember->new(name => 'factory_name', default => 'tables'),
		Class::STL::ClassMembers::DataMember->new(name => 'target_mem_name', default => 'name'),
		Class::STL::ClassMembers::DataMember->new(name => 'element_type', default => 'ETL::Pequel3::Type::Table::Abstract'),
		Class::STL::ClassMembers::DataMember->new(name => 'catalogue_type', default => 'table_types');
	use Class::STL::ClassMembers::Constructor;
	sub factory
	{
		my $self = shift;
		my %p = @_; # table_type, static, dataset_spec, ...
		$self->err()->user_error(10209, "Undefined pequel_ref parameter!") 
		unless (defined($p{pequel_ref}));
		$self->err()->user_error(10210, "Undefined table_type parameter!") 
		unless (defined($p{table_type}));

		$p{input_dataset} = ETL::Pequel3::Type::DataSet::Input::Factory::Global->new()->factory(
			dataset_spec => $p{dataset_spec},
			pequel_ref => $p{pequel_ref}
		) 
			if (defined($p{dataset_spec}));

		foreach ($self->catalogue()->to_array()) {
			if ((my $tobj = $_->factory(%p, pequel_ref => $p{pequel_ref})) != 0) {
				if (defined($p{input_dataset})) {
					$p{input_dataset}->datasource()->fdname($tobj->tname());
					$p{input_dataset}->input_fields($tobj->fields());
#?					$tobj->fields()->dataset($p{input_dataset});
				}
				$self->err()->trace_msg(10, "@{[ $tobj->pequel_ref()->pequel_name() ]} -->"
					. "Table:@{[ $tobj->name() ]} [@{[ ref($self) ]}]");
				return $tobj;
			}
		}
		$self->err()->user_error(10208, "No match for table type @{[ 
				$p{table_type} 
				. ($p{static} ? q/(static)/ : q/(dynamic)/) 
				. q/ [dataset_spec => /
				. (defined($p{input_dataset}) ? $p{dataset_spec} : q/NULL/)
				. q/]/
			]} for table @{[ $p{name} ]}!");
		return 0;
	}	
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::Factory::Local;
	use base qw(ETL::Pequel3::Type::Table::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::Constructor;
}
# ----------------------------------------------------------------------------------------------------
{
	package ETL::Pequel3::Type::Table::Factory::Global;
	use base qw(ETL::Pequel3::Type::Table::Factory::Abstract);
	use Class::STL::ClassMembers;
	use Class::STL::ClassMembers::SingletonConstructor;
}
# ----------------------------------------------------------------------------------------------------
1;
